//  -*- Mode: Java -*-
//
// Description.java

/*--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1997-2003              |
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------*/

package edu.isi.powerloom.logic;

import edu.isi.stella.javalib.Native;
import edu.isi.stella.javalib.StellaSpecialVariable;
import edu.isi.stella.*;

public class Description extends LogicObject {
    public Proposition proposition;
    public Vector ioVariables;
    /** The set of internally quantified variables besides the
     * descriptions's IO variables.  This set includes any top-level implicit
     * existential variables plus any variables introduced by some nested quantifier.
     * IMPORTANT: therefore, a description's internal variables are not necessarily
     * equivalent to its top-level implicit existential variables.
     */
    public Vector internalVariables;
    public KeyValueList queryPatterns;

  public static Description newDescription() {
    { Description self = null;

      self = new Description();
      self.dependentPropositionsIndex = null;
      self.dynamicSlots = KeyValueList.newKeyValueList();
      self.homeContext = ((Context)(Stella.$CONTEXT$.get()));
      self.surrogateValueInverse = null;
      self.variableValueInverse = null;
      self.queryPatterns = null;
      self.internalVariables = Logic.ZERO_VARIABLES_VECTOR;
      self.ioVariables = null;
      self.proposition = null;
      LogicObject.logLogicObject(self);
      return (self);
    }
  }

  /** Classify 'self' and return three values, its direct
   * supers, direct subs, and a list of equivalent descriptions.
   * Setting 'supersOnly?' may speed up the computation (perhaps by a lot).
   * If 'description' is nameless and has no dependent propositions, then
   * it is automatically removed from the hierarchy after classification.
   * @param self
   * @param onlysupersP
   * @param MV_returnarray
   * @return Cons
   */
  public static Cons findDirectSupersAndSubs(Description self, boolean onlysupersP, Object [] MV_returnarray) {
    Logic.finalizeObjects();
    if (self == null) {
      { Cons _return_temp = Stella.NIL;

        MV_returnarray[0] = Stella.NIL;
        MV_returnarray[1] = Stella.NIL;
        return (_return_temp);
      }
    }
    { Object old$Findsupersandsubsdescription$000 = Logic.$FINDSUPERSANDSUBSDESCRIPTION$.get();

      try {
        Native.setSpecial(Logic.$FINDSUPERSANDSUBSDESCRIPTION$, self);
        { Object old$Classificationsession$000 = Logic.$CLASSIFICATIONSESSION$.get();
          Object old$Context$000 = Stella.$CONTEXT$.get();

          try {
            Native.setSpecial(Logic.$CLASSIFICATIONSESSION$, Logic.getClassificationSession(Logic.KWD_DESCRIPTION));
            Native.setSpecial(Stella.$CONTEXT$, ((ClassificationSession)(Logic.$CLASSIFICATIONSESSION$.get())).classificationWorld);
            { boolean virginP = self.virginP();
              int oldcurrenttime = Logic.currentClassificationTime();
              Cons supers = Stella.NIL;
              Cons subs = Stella.NIL;
              Cons equivalents = Stella.NIL;

              if (virginP) {
                Description.introduceIntoClassificationSession(self);
              }
              LogicObject.upclassifyOneDescription(self);
              { NamedDescription renamed_Super = null;
                Cons iter000 = Description.allNamedDirectSuperdescriptions(self, Stella.FALSE);

                while (!(iter000 == Stella.NIL)) {
                  {
                    renamed_Super = ((NamedDescription)(iter000.value));
                    iter000 = iter000.rest;
                  }
                  LogicObject.upclassifyOneDescription(renamed_Super);
                }
              }
              supers = Description.allNamedDirectSuperdescriptions(self, Stella.TRUE);
              if (!onlysupersP) {
                LogicObject.downclassifyOneDescription(self);
                subs = LogicObject.allDirectSubcollections(self, Stella.TRUE).consify();
              }
              { LogicObject e = null;
                Cons iter001 = LogicObject.allEquivalentCollections(self, Stella.FALSE);
                Cons collect000 = null;

                while (!(iter001 == Stella.NIL)) {
                  {
                    e = ((LogicObject)(iter001.value));
                    iter001 = iter001.rest;
                  }
                  if (Description.namedDescriptionP(((Description)(e)))) {
                    if (collect000 == null) {
                      {
                        collect000 = Stella_Object.cons(e, Stella.NIL);
                        if (equivalents == Stella.NIL) {
                          equivalents = collect000;
                        }
                        else {
                          Cons.addConsToEndOfConsList(equivalents, collect000);
                        }
                      }
                    }
                    else {
                      {
                        collect000.rest = Stella_Object.cons(e, Stella.NIL);
                        collect000 = collect000.rest;
                      }
                    }
                  }
                }
              }
              if (virginP) {
                Description.evaporateVirgin(self);
                ((ClassificationSession)(Logic.$CLASSIFICATIONSESSION$.get())).introductionTimestampStack.pop();
              }
              { Cons _return_temp = supers;

                MV_returnarray[0] = subs;
                MV_returnarray[1] = equivalents;
                return (_return_temp);
              }
            }

          } finally {
            Stella.$CONTEXT$.set(old$Context$000);
            Logic.$CLASSIFICATIONSESSION$.set(old$Classificationsession$000);
          }
        }

      } finally {
        Logic.$FINDSUPERSANDSUBSDESCRIPTION$.set(old$Findsupersandsubsdescription$000);
      }
    }
  }

  public static void evaporateVirgin(Description self) {
    { List parentimplies = List.newList();
      List childimplies = List.newList();
      Description parent = null;
      Description child = null;
      Proposition bridgeprop = null;

      { Proposition p = null;
        Iterator iter000 = Logic.unfilteredDependentPropositions(self, null).allocateIterator();

        while (iter000.nextP()) {
          p = ((Proposition)(iter000.value));
          if ((p.arguments.theArray)[0] == self) {
            parentimplies.push(p);
          }
          else {
            childimplies.push(p);
          }
        }
      }
      { Proposition cp = null;
        Cons iter001 = childimplies.theConsList;

        while (!(iter001 == Stella.NIL)) {
          {
            cp = ((Proposition)(iter001.value));
            iter001 = iter001.rest;
          }
          child = ((Description)((cp.arguments.theArray)[0]));
          { Proposition pp = null;
            Cons iter002 = parentimplies.theConsList;

            while (!(iter002 == Stella.NIL)) {
              {
                pp = ((Proposition)(iter002.value));
                iter002 = iter002.rest;
              }
              parent = ((Description)((pp.arguments.theArray)[1]));
              bridgeprop = Description.assertDescriptionImpliesDescription(child, parent, Stella.TRUE);
              if (((BooleanWrapper)(KeyValueList.dynamicSlotValue(cp.dynamicSlots, Logic.SYM_LOGIC_SUBRELATION_LINKp, Stella.FALSE_WRAPPER))).wrapperValue &&
                  ((BooleanWrapper)(KeyValueList.dynamicSlotValue(pp.dynamicSlots, Logic.SYM_LOGIC_SUBRELATION_LINKp, Stella.FALSE_WRAPPER))).wrapperValue) {
                KeyValueList.setDynamicSlotValue(bridgeprop.dynamicSlots, Logic.SYM_LOGIC_SUBRELATION_LINKp, (Stella.TRUE ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER), Stella.FALSE_WRAPPER);
              }
              Proposition.updatePropositionTruthValue(bridgeprop, Logic.KWD_ASSERT_TRUE);
            }
          }
        }
      }
      { Proposition p = null;
        Iterator iter003 = Logic.unfilteredDependentPropositions(self, null).allocateIterator();

        while (iter003.nextP()) {
          p = ((Proposition)(iter003.value));
          Proposition.destroyProposition(p);
        }
      }
    }
  }

  public boolean virginP() {
    { Description self = this;

      return (Logic.unfilteredDependentPropositions(self, null).emptyP() &&
          (!LogicObject.namedCollectionP(self)));
    }
  }

  public static void markUnmarkedSuperrelations(Description description, MarkerTable selfisbelowtable) {
    selfisbelowtable.setMarker(description);
    { NamedDescription renamed_Super = null;
      Cons iter000 = Description.allNamedDirectSuperdescriptions(description, Stella.TRUE);

      while (!(iter000 == Stella.NIL)) {
        {
          renamed_Super = ((NamedDescription)(iter000.value));
          iter000 = iter000.rest;
        }
        if (!selfisbelowtable.testMarkerP(description)) {
          Description.markUnmarkedSuperrelations(renamed_Super, selfisbelowtable);
        }
      }
    }
  }

  public static void introduceIntoClassificationSession(Description self) {
    { World world = ((World)(KeyValueList.dynamicSlotValue(((Module)(Stella.$MODULE$.get())).dynamicSlots, Logic.SYM_LOGIC_META_INFERENCE_CACHE, null)));

      if ((world == null) ||
          (((ClassificationSession)(KeyValueList.dynamicSlotValue(world.dynamicSlots, Logic.SYM_LOGIC_CLASSIFICATION_SESSION, null))) == null)) {
        return;
      }
      { int newtimestamp = Logic.bumpClassificationTimeclock();

        Logic.getClassificationCache(self).introductionTimestamp = newtimestamp;
        ((ClassificationSession)(Logic.$CLASSIFICATIONSESSION$.get())).introductionTimestampStack.push(Stella.wrapInteger(newtimestamp));
      }
    }
  }

  public static void addSubsumptionLink(Description subdescription, Description superdescription) {
    { Proposition impliesprop = Description.assertDescriptionImpliesDescription(subdescription, superdescription, Stella.FALSE);

      KeyValueList.setDynamicSlotValue(impliesprop.dynamicSlots, Logic.SYM_LOGIC_SUBSUMPTION_LINKp, (Stella.TRUE ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER), Stella.FALSE_WRAPPER);
      if (Stella_Object.traceKeywordP(Logic.KWD_CLASSIFIER_INFERENCES)) {
        { Object old$Indentcounter$000 = Logic.$INDENTCOUNTER$.get();
          Object old$PrintreadablyP$000 = Stella.$PRINTREADABLYp$.get();

          try {
            Native.setIntSpecial(Logic.$INDENTCOUNTER$, 4);
            Native.setBooleanSpecial(Stella.$PRINTREADABLYp$, Stella.TRUE);
            if (Logic.$EMIT_THINKING_DOTSp$) {
              System.out.println();
            }
            {
              System.out.println("Recognized subsumption link:");
              System.out.println("    " + impliesprop);
            }
;

          } finally {
            Stella.$PRINTREADABLYp$.set(old$PrintreadablyP$000);
            Logic.$INDENTCOUNTER$.set(old$Indentcounter$000);
          }
        }
      }
    }
    LogicObject.allInferableDirectSubcollections(superdescription).push(subdescription);
    { NamedDescription p = null;
      Cons iter000 = Description.allNamedDirectSuperdescriptions(superdescription, Stella.FALSE);

      while (!(iter000 == Stella.NIL)) {
        {
          p = ((NamedDescription)(iter000.value));
          iter000 = iter000.rest;
        }
        { List subcollections = LogicObject.allInferableDirectSubcollections(p);

          subcollections.theConsList = ((Cons)(Logic.mostGeneralCollections(subcollections.theConsList)));
        }
      }
    }
  }

  public static Cons allNamedDirectSubdescriptions(Description self) {
    { Cons result = Stella.NIL;

      { LogicObject d = null;
        Iterator iter000 = LogicObject.allDirectSubcollections(self, Stella.FALSE);

        while (iter000.nextP()) {
          d = ((LogicObject)(iter000.value));
          if (Stella_Object.isaP(d, Logic.SGT_LOGIC_NAMED_DESCRIPTION)) {
            result = Stella_Object.cons(d, result);
          }
        }
      }
      return (result);
    }
  }

  public static Cons allNamedDirectSuperdescriptions(Description self, boolean removeequivalentsP) {
    { Cons result = Stella.NIL;

      { LogicObject d = null;
        Iterator iter000 = LogicObject.allDirectSupercollections(self, removeequivalentsP);

        while (iter000.nextP()) {
          d = ((LogicObject)(iter000.value));
          if (Stella_Object.isaP(d, Logic.SGT_LOGIC_NAMED_DESCRIPTION)) {
            result = Stella_Object.cons(d, result);
          }
        }
      }
      return (result);
    }
  }

  public static boolean nonInferableP(Description self) {
    return (!Description.inferableP(self));
  }

  public static boolean inferableP(Description self) {
    { Object old$Context$000 = Stella.$CONTEXT$.get();
      Object old$Module$000 = Stella.$MODULE$.get();

      try {
        Native.setSpecial(Stella.$CONTEXT$, Logic.getInferableTestContext());
        Native.setSpecial(Stella.$MODULE$, ((Context)(Stella.$CONTEXT$.get())).baseModule);
        return (Description.inferableWithCycleCheckP(self, Stella.NIL));

      } finally {
        Stella.$MODULE$.set(old$Module$000);
        Stella.$CONTEXT$.set(old$Context$000);
      }
    }
  }

  public static boolean inferableWithCycleCheckP(Description self, Cons alreadyvisitedlist) {
    if (alreadyvisitedlist.membP(self)) {
      return (Stella.FALSE);
    }
    { boolean inferableP = Stella.FALSE;

      if (Surrogate.subtypeOfP(Stella_Object.safePrimaryType(self), Logic.SGT_LOGIC_NAMED_DESCRIPTION)) {
        { NamedDescription self000 = ((NamedDescription)(self));

          { MemoizationTable memoTable000 = ((MemoizationTable)(Logic.SGT_LOGIC_INFERABLE_WITH_CYCLE_CHECKp_MEMO_TABLE_000.surrogateValue));
            Cons memoizedEntry000 = null;
            Stella_Object memoizedValue000 = null;

            if (memoTable000 == null) {
              Surrogate.initializeMemoizationTable(Logic.SGT_LOGIC_INFERABLE_WITH_CYCLE_CHECKp_MEMO_TABLE_000, "(:MAX-VALUES 500 :TIMESTAMPS (:IMPLIES-PROPOSITION-UPDATE))");
              memoTable000 = ((MemoizationTable)(Logic.SGT_LOGIC_INFERABLE_WITH_CYCLE_CHECKp_MEMO_TABLE_000.surrogateValue));
            }
            if (Stella.$MEMOIZATION_ENABLEDp$) {
              memoizedEntry000 = MruMemoizationTable.lookupMruMemoizedValue(((MruMemoizationTable)(memoTable000)), self000, ((Context)(Stella.$CONTEXT$.get())), Stella.MEMOIZED_NULL_VALUE, null, -1);
              memoizedValue000 = memoizedEntry000.value;
            }
            if (memoizedValue000 != null) {
              if (memoizedValue000 == Stella.MEMOIZED_NULL_VALUE) {
                memoizedValue000 = null;
              }
            }
            else {
              { boolean testValue000 = Stella.FALSE;

                testValue000 = !(self000.surrogateValueInverse == Logic.SGT_STELLA_THING);
                if (testValue000) {
                  testValue000 = (alreadyvisitedlist = Stella_Object.cons(self000, alreadyvisitedlist)) != null;
                  if (testValue000) {
                    { boolean foundP000 = Stella.FALSE;

                      { Proposition p = null;
                        Cons iter000 = LogicObject.applicableRulesOfDescription(self000, Logic.KWD_BACKWARD, Stella.FALSE);

                        loop000 : while (!(iter000 == Stella.NIL)) {
                          {
                            p = ((Proposition)(iter000.value));
                            iter000 = iter000.rest;
                          }
                          if ((!((BooleanWrapper)(KeyValueList.dynamicSlotValue(p.dynamicSlots, Logic.SYM_LOGIC_FORWARD_ONLYp, Stella.FALSE_WRAPPER))).wrapperValue) &&
                              (Proposition.trueP(p) &&
                               NamedDescription.inferableThroughBySomeDescendantP(self000, ((LogicObject)((p.arguments.theArray)[0])), alreadyvisitedlist))) {
                            foundP000 = Stella.TRUE;
                            break loop000;
                          }
                        }
                      }
                      testValue000 = foundP000;
                    }
                  }
                }
                memoizedValue000 = (testValue000 ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER);
              }
              if (Stella.$MEMOIZATION_ENABLEDp$) {
                memoizedEntry000.value = ((memoizedValue000 == null) ? Stella.MEMOIZED_NULL_VALUE : memoizedValue000);
              }
            }
            inferableP = BooleanWrapper.coerceWrappedBooleanToBoolean(((BooleanWrapper)(memoizedValue000)));
          }
        }
      }
      else {
        inferableP = Stella.TRUE;
      }
      return (inferableP);
    }
  }

  public static List allIntensionalParents(Description self) {
    { List collection = List.newList();

      Description.helpCollectIntensionalParents(self, self.proposition, collection);
      return (collection);
    }
  }

  public static void helpCollectIntensionalParents(Description description, Proposition proposition, List collection) {
    { Keyword testValue000 = proposition.kind;

      if ((testValue000 == Logic.KWD_PREDICATE) ||
          ((testValue000 == Logic.KWD_FUNCTION) ||
           (testValue000 == Logic.KWD_ISA))) {
        { boolean testValue001 = Stella.FALSE;

          { boolean alwaysP000 = Stella.TRUE;

            { PatternVariable arg2 = null;
              Stella_Object arg1 = null;
              Vector vector000 = proposition.arguments;
              int index000 = 0;
              int length000 = vector000.length();
              Vector vector001 = description.ioVariables;
              int index001 = 0;
              int length001 = vector001.length();

              loop000 : while ((index000 < length000) &&
                  (index001 < length001)) {
                {
                  arg1 = (vector000.theArray)[index000];
                  index000 = index000 + 1;
                }
                {
                  arg2 = ((PatternVariable)((vector001.theArray)[index001]));
                  index001 = index001 + 1;
                }
                if (!(arg1 == arg2)) {
                  alwaysP000 = Stella.FALSE;
                  break loop000;
                }
              }
            }
            testValue001 = alwaysP000;
          }
          if (testValue001) {
            testValue001 = proposition.arguments.length() == description.ioVariables.length();
          }
          if (testValue001) {
            collection.push(Logic.getDescription(((Surrogate)(proposition.operator))));
          }
        }
      }
      else if (testValue000 == Logic.KWD_AND) {
        { Stella_Object arg = null;
          Vector vector002 = proposition.arguments;
          int index002 = 0;
          int length002 = vector002.length();

          while (index002 < length002) {
            {
              arg = (vector002.theArray)[index002];
              index002 = index002 + 1;
            }
            Description.helpCollectIntensionalParents(description, ((Proposition)(arg)), collection);
          }
        }
      }
      else if (testValue000 == Logic.KWD_EXISTS) {
        Description.helpCollectIntensionalParents(description, ((Proposition)((proposition.arguments.theArray)[0])), collection);
      }
      else {
      }
    }
  }

  public static Description copyDescription(Description self, KeyValueList parentmapping, boolean addbacklinksP) {
    { Object old$Context$000 = Stella.$CONTEXT$.get();
      Object old$Module$000 = Stella.$MODULE$.get();

      try {
        Native.setSpecial(Stella.$CONTEXT$, self.homeContext);
        Native.setSpecial(Stella.$MODULE$, ((Context)(Stella.$CONTEXT$.get())).baseModule);
        { Description copy = Logic.createDescription(Stella.NULL_INTEGER, Stella.FALSE);
          KeyValueList mapping = KeyValueList.newKeyValueList();

          if (parentmapping != null) {
            { Stella_Object v = null;
              Stella_Object k = null;
              KvCons iter000 = parentmapping.theKvList;

              while (iter000 != null) {
                k = iter000.key;
                {
                  v = iter000.value;
                  iter000 = iter000.rest;
                }
                if (Stella_Object.isaP(k, Logic.SGT_LOGIC_PATTERN_VARIABLE)) {
                  mapping.insertAt(k, v);
                }
              }
            }
          }
          copy.ioVariables = Logic.copyVariablesVector(self.ioVariables, mapping);
          copy.internalVariables = Logic.copyVariablesVector(self.internalVariables, mapping);
          if (((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))) != null) {
            { Vector externalvarscopy = Logic.copyVariablesVector(((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))), mapping);
              Cons realexternalvars = Stella.NIL;

              { boolean foundP000 = Stella.FALSE;

                { PatternVariable copy000 = null;
                  Vector vector000 = externalvarscopy;
                  int index000 = 0;
                  int length000 = vector000.length();

                  loop001 : while (index000 < length000) {
                    {
                      copy000 = ((PatternVariable)((vector000.theArray)[index000]));
                      index000 = index000 + 1;
                    }
                    if (!Logic.variableP(copy000)) {
                      foundP000 = Stella.TRUE;
                      break loop001;
                    }
                  }
                }
                if (foundP000) {
                  { PatternVariable copy000 = null;
                    Vector vector001 = externalvarscopy;
                    int index001 = 0;
                    int length001 = vector001.length();
                    Cons collect000 = null;

                    while (index001 < length001) {
                      {
                        copy000 = ((PatternVariable)((vector001.theArray)[index001]));
                        index001 = index001 + 1;
                      }
                      if (Logic.variableP(copy000)) {
                        if (collect000 == null) {
                          {
                            collect000 = Stella_Object.cons(copy000, Stella.NIL);
                            if (realexternalvars == Stella.NIL) {
                              realexternalvars = collect000;
                            }
                            else {
                              Cons.addConsToEndOfConsList(realexternalvars, collect000);
                            }
                          }
                        }
                        else {
                          {
                            collect000.rest = Stella_Object.cons(copy000, Stella.NIL);
                            collect000 = collect000.rest;
                          }
                        }
                      }
                    }
                  }
                  if (!(realexternalvars == Stella.NIL)) {
                    externalvarscopy = Logic.copyConsListToVariablesVector(realexternalvars);
                  }
                  else {
                    externalvarscopy = null;
                  }
                }
              }
              KeyValueList.setDynamicSlotValue(copy.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, externalvarscopy, null);
            }
          }
          if (self.proposition != null) {
            { Object old$Boundtooffsetcounter$000 = Logic.$BOUNDTOOFFSETCOUNTER$.get();
              Object old$Evaluationmode$000 = Logic.$EVALUATIONMODE$.get();

              try {
                Native.setIntSpecial(Logic.$BOUNDTOOFFSETCOUNTER$, Stella.NULL_INTEGER);
                Native.setSpecial(Logic.$EVALUATIONMODE$, Logic.KWD_DESCRIPTION);
                copy.proposition = (addbacklinksP ? Proposition.recursivelyFastenDownPropositions(Proposition.copyProposition(self.proposition, mapping), Stella.FALSE) : Proposition.copyProposition(self.proposition, mapping));

              } finally {
                Logic.$EVALUATIONMODE$.set(old$Evaluationmode$000);
                Logic.$BOUNDTOOFFSETCOUNTER$.set(old$Boundtooffsetcounter$000);
              }
            }
          }
          if (((BooleanWrapper)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_DONT_OPTIMIZEp, Stella.FALSE_WRAPPER))).wrapperValue) {
            KeyValueList.setDynamicSlotValue(copy.dynamicSlots, Logic.SYM_LOGIC_DONT_OPTIMIZEp, (Stella.TRUE ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER), Stella.FALSE_WRAPPER);
          }
          mapping.free();
          return (copy);
        }

      } finally {
        Stella.$MODULE$.set(old$Module$000);
        Stella.$CONTEXT$.set(old$Context$000);
      }
    }
  }

  public static void simplifyDescription(Description description, boolean postoptimizationP) {
    if (description.proposition == null) {
      return;
    }
    { Object old$PostoptimizationP$000 = Logic.$POSTOPTIMIZATIONp$.get();

      try {
        Native.setBooleanSpecial(Logic.$POSTOPTIMIZATIONp$, postoptimizationP);
        Proposition.simplifyProposition(description.proposition);

      } finally {
        Logic.$POSTOPTIMIZATIONp$.set(old$PostoptimizationP$000);
      }
    }
  }

  public static Description selectOptimalQueryPattern(Description description, BooleanVector booleanvector, ControlFrame parentframe) {
    { Description pattern = null;
      KeyValueList querypatterns = description.queryPatterns;

      if (querypatterns != null) {
        pattern = ((Description)(querypatterns.lookup(booleanvector)));
      }
      if (Stella_Object.traceKeywordP(Logic.KWD_OPTIMIZER)) {
        pattern = null;
      }
      if (pattern != null) {
        return (pattern);
      }
      if (querypatterns == null) {
        description.queryPatterns = KeyValueList.newKeyValueList();
      }
      { Object old$Boundtooffsetcounter$000 = Logic.$BOUNDTOOFFSETCOUNTER$.get();

        try {
          Native.setIntSpecial(Logic.$BOUNDTOOFFSETCOUNTER$, 0);
          pattern = Description.copyDescription(description, null, Stella.TRUE);

        } finally {
          Logic.$BOUNDTOOFFSETCOUNTER$.set(old$Boundtooffsetcounter$000);
        }
      }
      description.queryPatterns.insertAt(booleanvector, pattern);
      if (!((!((!(Logic.$QUERY_OPTIMIZATION_STRATEGY$ == Logic.KWD_NONE)) &&
          (!((Logic.$QUERY_OPTIMIZATION_STRATEGY$ == Logic.KWD_DYNAMIC) ||
          (Logic.$QUERY_OPTIMIZATION_STRATEGY$ == Logic.KWD_DYNAMIC_WITH_CLUSTERING))))) ||
          (((BooleanWrapper)(KeyValueList.dynamicSlotValue(description.dynamicSlots, Logic.SYM_LOGIC_DONT_OPTIMIZEp, Stella.FALSE_WRAPPER))).wrapperValue ||
           (Proposition.containsOperatorP(description.proposition, Logic.SGT_PL_KERNEL_KB_CUT) ||
            Proposition.containsOperatorP(description.proposition, Logic.SGT_PL_KERNEL_KB_BOUND_VARIABLES))))) {
        Description.optimizeQuery(pattern, booleanvector, parentframe);
      }
      Description.simplifyDescription(pattern, Stella.TRUE);
      return (pattern);
    }
  }

  public static void optimizeQuery(Description description, BooleanVector booleanvector, ControlFrame parentframe) {
    { ControlFrame frame = ControlFrame.createDownFrame(parentframe, null);
      int variablescount = description.ioVariables.length() + description.internalVariables.length() + (((((Vector)(KeyValueList.dynamicSlotValue(description.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))) != null) ? ((Vector)(KeyValueList.dynamicSlotValue(description.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))).length() : 0));

      Logic.initializeOptimizerMemory(variablescount);
      ControlFrame.createPatternRecord(frame, null, null);
      PatternRecord.activatePatternRecord(frame.patternRecord, variablescount);
      { BooleanWrapper initiallyboundP = null;
        PatternVariable iovar = null;
        Vector vector000 = description.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();
        BooleanVector vector001 = booleanvector;
        int index001 = 0;
        int length001 = vector001.length();

        while ((index000 < length000) &&
            (index001 < length001)) {
          {
            iovar = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          {
            initiallyboundP = ((BooleanWrapper)((vector001.theArray)[index001]));
            index001 = index001 + 1;
          }
          if (BooleanWrapper.coerceWrappedBooleanToBoolean(initiallyboundP)) {
            Logic.simulateBindVariableToValue(iovar);
          }
        }
      }
      { Object old$Queryoptimizercontrolframe$000 = Logic.$QUERYOPTIMIZERCONTROLFRAME$.get();

        try {
          Native.setSpecial(Logic.$QUERYOPTIMIZERCONTROLFRAME$, frame);
          Proposition.simulateAndOptimizeQuery(description.proposition);

        } finally {
          Logic.$QUERYOPTIMIZERCONTROLFRAME$.set(old$Queryoptimizercontrolframe$000);
        }
      }
      frame.state = Logic.KWD_PATTERN;
      ControlFrame.popDownFrame(parentframe);
    }
  }

  public static int accessObservedCardinality(Description self) {
    { Object old$Context$000 = Stella.$CONTEXT$.get();
      Object old$Module$000 = Stella.$MODULE$.get();

      try {
        Native.setSpecial(Stella.$CONTEXT$, ((Module)(Stella.$MODULE$.get())));
        Native.setSpecial(Stella.$MODULE$, ((Context)(Stella.$CONTEXT$.get())).baseModule);
        { Stella_Object value = Logic.accessBinaryValue(self, Logic.SGT_PL_KERNEL_KB_OBSERVED_CARDINALITY_OF);

          if ((value != null) &&
              Stella_Object.isaP(value, Logic.SGT_STELLA_INTEGER_WRAPPER)) {
            return (((IntegerWrapper)(value)).wrapperValue);
          }
          else {
            return (Stella.NULL_INTEGER);
          }
        }

      } finally {
        Stella.$MODULE$.set(old$Module$000);
        Stella.$CONTEXT$.set(old$Context$000);
      }
    }
  }

  public static boolean computedConstraintP(Description self) {
    return (Logic.testPropertyP(self, Logic.SGT_PL_KERNEL_KB_COMPUTED) &&
        (Logic.accessBinaryValue(self, Logic.SGT_PL_KERNEL_KB_RELATION_CONSTRAINT) != null));
  }

  public static boolean computedTermP(Description self) {
    return (Logic.testPropertyP(self, Logic.SGT_PL_KERNEL_KB_COMPUTED));
  }

  public static void dropTaxonomyImpliesSubsumesLink(Description taildescription, Description headdescription) {
    {
      headdescription = headdescription;
      taildescription = taildescription;
    }
    System.out.println("'drop-taxonomy-implies-subsumes-link':  Write me, I am a dummy!");
    return;
  }

  public static void addTaxonomyImpliesSubsumesLink(Description taildescription, Description headdescription) {
    return;
  }

  public static TaxonomyNode findOrCreateDescriptionImplicationSubsumptionNode(Description description) {
    { TaxonomyNode result = Description.findDescriptionImplicationSubsumptionNode(description);

      if (result != null) {
        return (result);
      }
      return (Description.createDescriptionImplicationSubsumptionNode(description, null));
    }
  }

  public static TaxonomyNode createDescriptionImplicationSubsumptionNode(Description description, TaxonomyNode parentnode) {
    { TaxonomyNode node = TaxonomyGraph.createTaxonomyNode(Logic.$IMPLICATION_SUBSUMPTION_GRAPH$, parentnode, description, parentnode == null);

      TaxonomyGraph.linkTaxonomyNodes(Logic.$IMPLICATION_SUBSUMPTION_GRAPH$, parentnode, node);
      KeyValueList.setDynamicSlotValue(description.dynamicSlots, Logic.SYM_STELLA_TAXONOMY_NODE, node, null);
      return (node);
    }
  }

  public static TaxonomyNode findDescriptionImplicationSubsumptionNode(Description description) {
    return (((TaxonomyNode)(KeyValueList.dynamicSlotValue(description.dynamicSlots, Logic.SYM_STELLA_TAXONOMY_NODE, null))));
  }

  public static Cons generateDescriptionsAsRule(Description head, Description tail, Proposition rule, boolean reversepolarityP) {
    { boolean forwardarrowP = ((BooleanWrapper)(KeyValueList.dynamicSlotValue(rule.dynamicSlots, Logic.SYM_LOGIC_FORWARD_ONLYp, Stella.FALSE_WRAPPER))).wrapperValue &&
          (!reversepolarityP);
      boolean reverseargumentsP = forwardarrowP ||
          reversepolarityP;
      Symbol arrow = Surrogate.typeToSymbol(Proposition.chooseImplicationOperator(rule, forwardarrowP));
      boolean mapheadvariablesP = Description.namedDescriptionP(head);
      Stella_Object headprop = null;
      Stella_Object tailprop = null;
      Cons universals = Stella.NIL;

      if (reverseargumentsP) {
        { Description temp = head;

          head = tail;
          tail = temp;
        }
        mapheadvariablesP = !mapheadvariablesP;
      }
      { PatternVariable var = null;
        Vector vector000 = (mapheadvariablesP ? tail.ioVariables : head.ioVariables);
        int index000 = 0;
        int length000 = vector000.length();
        Cons collect000 = null;

        while (index000 < length000) {
          {
            var = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          if (collect000 == null) {
            {
              collect000 = Stella_Object.cons(PatternVariable.generateOneVariable(var, Stella.TRUE), Stella.NIL);
              if (universals == Stella.NIL) {
                universals = collect000;
              }
              else {
                Cons.addConsToEndOfConsList(universals, collect000);
              }
            }
          }
          else {
            {
              collect000.rest = Stella_Object.cons(PatternVariable.generateOneVariable(var, Stella.TRUE), Stella.NIL);
              collect000 = collect000.rest;
            }
          }
        }
      }
      { Object old$Skolemnamemappingtable$000 = Logic.$SKOLEMNAMEMAPPINGTABLE$.get();

        try {
          Native.setSpecial(Logic.$SKOLEMNAMEMAPPINGTABLE$, (mapheadvariablesP ? ((KeyValueList)(Logic.createSkolemMappingTable(head.ioVariables, tail.ioVariables))) : null));
          headprop = Description.generateDescriptionProposition(head, reversepolarityP);

        } finally {
          Logic.$SKOLEMNAMEMAPPINGTABLE$.set(old$Skolemnamemappingtable$000);
        }
      }
      { Object old$Skolemnamemappingtable$001 = Logic.$SKOLEMNAMEMAPPINGTABLE$.get();

        try {
          Native.setSpecial(Logic.$SKOLEMNAMEMAPPINGTABLE$, ((!mapheadvariablesP) ? ((KeyValueList)(Logic.createSkolemMappingTable(tail.ioVariables, head.ioVariables))) : null));
          tailprop = Description.generateDescriptionProposition(tail, reversepolarityP);

        } finally {
          Logic.$SKOLEMNAMEMAPPINGTABLE$.set(old$Skolemnamemappingtable$001);
        }
      }
      return (Stella.list$(Stella_Object.cons(Logic.SYM_STELLA_FORALL, Stella_Object.cons(universals, Stella_Object.cons(Stella_Object.cons(Stella_Object.cons(arrow, Stella_Object.cons(headprop, Stella_Object.cons(tailprop, Stella.NIL))), Stella.NIL), Stella.NIL)))));
    }
  }

  public static Cons generateDescriptionProposition(Description self, boolean invertP) {
    { Stella_Object prop = Proposition.generateProposition(self.proposition);
      Cons existentals = ((self.internalVariables.arraySize > 0) ? Description.topLevelExistentialVariables(self) : Stella.NIL);

      { int i = Stella.NULL_INTEGER;
        Stella_Object var = null;
        Cons iter000 = existentals;
        int iter001 = 0;

        while (!(iter000 == Stella.NIL)) {
          {
            var = iter000.value;
            iter000 = iter000.rest;
          }
          {
            i = iter001;
            iter001 = iter001 + 1;
          }
          existentals.nthSetter(PatternVariable.generateOneVariable(((PatternVariable)(var)), Stella.TRUE), i);
        }
      }
      if (!(existentals == Stella.NIL)) {
        prop = Stella.list$(Stella_Object.cons(Logic.SYM_STELLA_EXISTS, Stella_Object.cons(existentals, Stella_Object.cons(Stella_Object.cons(prop, Stella.NIL), Stella.NIL))));
      }
      if (invertP) {
        return (Stella.list$(Stella_Object.cons(Logic.SYM_STELLA_NOT, Stella_Object.cons(prop, Stella_Object.cons(Stella.NIL, Stella.NIL)))));
      }
      else {
        return (((Cons)(prop)));
      }
    }
  }

  public static Stella_Object generateDescription(Description self) {
    if (Description.namedDescriptionP(self)) {
      return (Logic.internalStellaOperatorToKif(self.descriptionName()));
    }
    else if (((Description)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null))) != null) {
      return (Stella.list$(Stella_Object.cons(Logic.SYM_STELLA_NOT, Stella_Object.cons(Logic.internalStellaOperatorToKif(((Description)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null))).descriptionName()), Stella_Object.cons(Stella.NIL, Stella.NIL)))));
    }
    else {
      return (Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_KAPPA, Stella_Object.cons(Logic.generateVariables(self.ioVariables, Stella.TRUE), Stella_Object.cons(Stella_Object.cons(Description.generateDescriptionProposition(self, Stella.FALSE), Stella.NIL), Stella.NIL)))));
    }
  }

  public static void printDescription(Description self, OutputStream stream, boolean toplevelP) {
    if (self.deletedP()) {
      stream.nativeStream.print("|dElEtEd-description|");
      if (self.surrogateValueInverse != null) {
        stream.nativeStream.print(self.surrogateValueInverse);
      }
      return;
    }
    if (((Boolean)(Logic.$LOADINGREGENERABLEOBJECTSp$.get())).booleanValue() ||
        (BooleanWrapper.coerceWrappedBooleanToBoolean(self.badP()) ||
         Logic.$DEBUG_MODEp$)) {
      stream.nativeStream.print("|d|");
    }
    else if (toplevelP &&
        (!((Boolean)(Stella.$PRINTREADABLYp$.get())).booleanValue())) {
      if (Logic.classP(self)) {
        stream.nativeStream.print("|c|");
      }
      else if (Logic.functionP(self)) {
        stream.nativeStream.print("|f|");
      }
      else if (Logic.relationP(self)) {
        stream.nativeStream.print("|r|");
      }
      else {
        stream.nativeStream.print("|d|");
      }
    }
    Logic.printLogicalForm(self, stream);
  }

  public static void printDescriptionsAsKifRule(Description head, Description tail, Proposition rule, boolean reversepolarityP) {
    { OutputStream stream = ((OutputStream)(Logic.$PRINTLOGICALFORMSTREAM$.get()));
      boolean forwardarrowP = ((BooleanWrapper)(KeyValueList.dynamicSlotValue(rule.dynamicSlots, Logic.SYM_LOGIC_FORWARD_ONLYp, Stella.FALSE_WRAPPER))).wrapperValue &&
          (!reversepolarityP);
      boolean reverseargumentsP = forwardarrowP ||
          reversepolarityP;
      boolean mapheadvariablesP = Description.namedDescriptionP(head);
      int currentindentcounter = ((Integer)(Logic.$INDENTCOUNTER$.get())).intValue();
      Surrogate operatorprefix = Proposition.chooseImplicationOperator(rule, forwardarrowP);
      int operatorprefixindent = 2 + (operatorprefix.symbolName).length();

      if (head.deletedP() ||
          tail.deletedP()) {
        stream.nativeStream.print("(=> <DeLeTeD ArGuMeNt(S)>)");
        return;
      }
      if (reverseargumentsP) {
        { Description temp = head;

          head = tail;
          tail = temp;
        }
        mapheadvariablesP = !mapheadvariablesP;
      }
      { Object old$Indentcounter$000 = Logic.$INDENTCOUNTER$.get();

        try {
          Native.setIntSpecial(Logic.$INDENTCOUNTER$, currentindentcounter);
          stream.nativeStream.print("(" + Logic.stringifiedSurrogate(Logic.SGT_PL_KERNEL_KB_FORALL) + " ");
          Logic.printKifQuantifiedVariables((mapheadvariablesP ? tail.ioVariables : head.ioVariables), Stella.FALSE);
          stream.nativeStream.println();
          Logic.increaseIndent(Stella.NULL_INTEGER);
          Logic.printIndent(stream, Stella.NULL_INTEGER);
          stream.nativeStream.print("(" + Logic.stringifiedSurrogate(operatorprefix) + " ");
          Logic.increaseIndent(operatorprefixindent);
          { Object old$Skolemnamemappingtable$000 = Logic.$SKOLEMNAMEMAPPINGTABLE$.get();

            try {
              Native.setSpecial(Logic.$SKOLEMNAMEMAPPINGTABLE$, (mapheadvariablesP ? ((KeyValueList)(Logic.createSkolemMappingTable(head.ioVariables, tail.ioVariables))) : null));
              Description.printKifDescriptionProposition(head, reversepolarityP);

            } finally {
              Logic.$SKOLEMNAMEMAPPINGTABLE$.set(old$Skolemnamemappingtable$000);
            }
          }
          stream.nativeStream.println();
          Logic.printIndent(stream, Stella.NULL_INTEGER);
          { Object old$Skolemnamemappingtable$001 = Logic.$SKOLEMNAMEMAPPINGTABLE$.get();

            try {
              Native.setSpecial(Logic.$SKOLEMNAMEMAPPINGTABLE$, ((!mapheadvariablesP) ? ((KeyValueList)(Logic.createSkolemMappingTable(tail.ioVariables, head.ioVariables))) : null));
              Description.printKifDescriptionProposition(tail, reversepolarityP);

            } finally {
              Logic.$SKOLEMNAMEMAPPINGTABLE$.set(old$Skolemnamemappingtable$001);
            }
          }
          stream.nativeStream.print("))");
          Logic.decreaseIndent(operatorprefixindent);
          Logic.decreaseIndent(Stella.NULL_INTEGER);

        } finally {
          Logic.$INDENTCOUNTER$.set(old$Indentcounter$000);
        }
      }
    }
  }

  public static void printKifDescriptionProposition(Description self, boolean invertP) {
    { OutputStream stream = ((OutputStream)(Logic.$PRINTLOGICALFORMSTREAM$.get()));
      Vector existentials = ((self.internalVariables.arraySize > 0) ? Logic.copyConsListToVariablesVector(Description.topLevelExistentialVariables(self)) : Logic.ZERO_VARIABLES_VECTOR);

      if (invertP) {
        stream.nativeStream.print("(" + Logic.stringifiedSurrogate(Logic.SGT_PL_KERNEL_KB_NOT) + " ");
        Logic.increaseIndent(4);
      }
      if (existentials.arraySize > 0) {
        stream.nativeStream.print("(" + Logic.stringifiedSurrogate(Logic.SGT_PL_KERNEL_KB_EXISTS) + " ");
        Logic.printKifQuantifiedVariables(existentials, Stella.FALSE);
        stream.nativeStream.println();
        Logic.increaseIndent(Stella.NULL_INTEGER);
        Logic.printIndent(stream, Stella.NULL_INTEGER);
      }
      Logic.printAsKifInternal(self.proposition);
      if (existentials.arraySize > 0) {
        stream.nativeStream.print(")");
        Logic.decreaseIndent(Stella.NULL_INTEGER);
      }
      if (invertP) {
        stream.nativeStream.print(")");
        Logic.decreaseIndent(4);
      }
    }
  }

  public static void printKifDescription(Description self) {
    { Object old$Queryiterator$000 = Logic.$QUERYITERATOR$.get();

      try {
        Native.setSpecial(Logic.$QUERYITERATOR$, null);
        { OutputStream stream = ((OutputStream)(Logic.$PRINTLOGICALFORMSTREAM$.get()));
          Description complement = ((Description)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null)));

          if (self.surrogateValueInverse != null) {
            stream.nativeStream.print(Logic.stringifiedSurrogate(self.surrogateValueInverse));
          }
          else if ((complement != null) &&
              ((complement.surrogateValueInverse != null) &&
               (!((((Keyword)(Logic.$PRINTMODE$.get())) == Logic.KWD_REALISTIC) ||
              (((Keyword)(Logic.$PRINTMODE$.get())) == Logic.KWD_ORIGINAL))))) {
            stream.nativeStream.print("not-" + Logic.stringifiedSurrogate(complement.surrogateValueInverse));
          }
          else {
            stream.nativeStream.print("(" + Logic.stringifiedSurrogate(Logic.SGT_PL_KERNEL_KB_KAPPA) + " ");
            Logic.printKifQuantifiedVariables(self.ioVariables, Stella.TRUE);
            stream.nativeStream.print(" ");
            Description.printKifDescriptionProposition(self, Stella.FALSE);
            stream.nativeStream.print(")");
          }
        }

      } finally {
        Logic.$QUERYITERATOR$.set(old$Queryiterator$000);
      }
    }
  }

  public static Iterator allocateAntecedentsIterator(Description goaldescription, Vector externalarguments, boolean reversepolarityP) {
    { Cons antecedents = Stella.NIL;
      LogicObject antecedent = null;
      boolean incompatibleargumentsP = Stella.FALSE;

      { Proposition prop = null;
        Cons iter000 = LogicObject.applicableRulesOfDescription(goaldescription, (reversepolarityP ? Logic.KWD_FORWARD : Logic.KWD_BACKWARD), Stella.FALSE);

        loop000 : while (!(iter000 == Stella.NIL)) {
          {
            prop = ((Proposition)(iter000.value));
            iter000 = iter000.rest;
          }
          if (((reversepolarityP ? (!((BooleanWrapper)(KeyValueList.dynamicSlotValue(prop.dynamicSlots, Logic.SYM_LOGIC_BACKWARD_ONLYp, Stella.FALSE_WRAPPER))).wrapperValue) : (!((BooleanWrapper)(KeyValueList.dynamicSlotValue(prop.dynamicSlots, Logic.SYM_LOGIC_FORWARD_ONLYp, Stella.FALSE_WRAPPER))).wrapperValue))) &&
              ((!reversepolarityP) ||
               (Description.atomicGoalDescriptionP(((Description)((prop.arguments.theArray)[1]))) ||
                Logic.closedTermP((prop.arguments.theArray)[1])))) {
            antecedent = (reversepolarityP ? ((LogicObject)((prop.arguments.theArray)[1])) : ((LogicObject)((prop.arguments.theArray)[0])));
            if ((!Stella_Object.isaP(antecedent, Logic.SGT_LOGIC_DESCRIPTION)) ||
                (!(externalarguments.length() == ((Description)(antecedent)).ioVariables.length()))) {
              continue loop000;
            }
            incompatibleargumentsP = Stella.FALSE;
            { Stella_Object arg = null;
              PatternVariable v = null;
              Vector vector000 = ((Description)(antecedent)).ioVariables;
              int index000 = 0;
              int length000 = vector000.length();
              Vector vector001 = externalarguments;
              int index001 = 0;
              int length001 = vector001.length();

              loop001 : while ((index000 < length000) &&
                  (index001 < length001)) {
                {
                  v = ((PatternVariable)((vector000.theArray)[index000]));
                  index000 = index000 + 1;
                }
                {
                  arg = (vector001.theArray)[index001];
                  index001 = index001 + 1;
                }
                { Stella_Object argvalue = Logic.argumentBoundTo(arg);
                  Stella_Object varvalue = Logic.groundValueOf(v);

                  if ((argvalue != null) &&
                      (((varvalue != null) &&
                      (!Stella_Object.eqlP(argvalue, varvalue))) ||
                       ((varvalue == null) &&
                        PatternVariable.failsAntecedentTypeCheckP(v, argvalue)))) {
                    incompatibleargumentsP = Stella.TRUE;
                    break loop001;
                  }
                }
              }
            }
            if (!incompatibleargumentsP) {
              antecedents = Stella_Object.cons(prop, antecedents);
            }
          }
        }
      }
      if (antecedents == Stella.NIL) {
        return (null);
      }
      else {
        return (Logic.optimizeOrderOfApplicableRules(antecedents, !reversepolarityP).allocateIterator());
      }
    }
  }

  /** Apply (inherit) the description 'description'
   * to members of the vector 'vector'.  Return TRUE if no clash was detected.
   * Constraint propagation happens only if it is enabled prior to calling
   * 'apply-kappa?'.
   * @param description
   * @param vector
   * @return boolean
   */
  public static boolean applyKappaP(Description description, Vector vector) {
    try {
      Logic.inheritUnnamedDescription(vector, description, Stella.FALSE);
    } catch (Clash e) {
      Stella.STANDARD_ERROR.nativeStream.println(Stella.exceptionMessage(e));
      return (Stella.FALSE);
    }
    return (Stella.TRUE);
  }

  public static TruthValue descriptionSpecializesDescriptionP(Description sub, Description renamed_Super) {
    if (((Boolean)(Logic.$REVERSEPOLARITYp$.get())).booleanValue()) {
      return (Logic.UNKNOWN_TRUTH_VALUE);
    }
    if (LogicObject.collectionImpliesCollectionP(sub, renamed_Super)) {
      return (Logic.TRUE_TRUTH_VALUE);
    }
    if (renamed_Super.arity() == 1) {
      return (Description.unaryDescriptionSpecializesDescriptionP(sub, renamed_Super));
    }
    { Object old$PrintreadablyP$000 = Stella.$PRINTREADABLYp$.get();

      try {
        Native.setBooleanSpecial(Stella.$PRINTREADABLYp$, Stella.TRUE);
        if ((Stella.$TRACED_KEYWORDS$ != null) &&
            Stella.$TRACED_KEYWORDS$.membP(Logic.KWD_GOAL_TREE)) {
          {
            System.out.println();
            System.out.println("Subsumption Test:");
            System.out.println("   sub=" + sub);
            System.out.println("   super=" + renamed_Super);
          }
;
        }

      } finally {
        Stella.$PRINTREADABLYp$.set(old$PrintreadablyP$000);
      }
    }
    { Vector skolemsvector = Vector.newVector(sub.ioVariables.length());
      TruthValue result = Logic.UNKNOWN_TRUTH_VALUE;

      Logic.pushMonotonicWorld();
      Logic.initializeInferenceWorld(((World)(((Context)(Stella.$CONTEXT$.get())))));
      { int i = Stella.NULL_INTEGER;
        int iter000 = 0;
        int upperBound000 = skolemsvector.length() - 1;

        while (iter000 <= upperBound000) {
          {
            i = iter000;
            iter000 = iter000 + 1;
          }
          (skolemsvector.theArray)[i] = (Logic.createHypothesizedSkolem("ANONYMOUS"));
        }
      }
      try {
        Logic.inheritDescription(skolemsvector, sub, Stella.FALSE);
        result = Logic.callAsk(Description.createQueryIterator(renamed_Super, skolemsvector));
      } catch (LogicException e) {
        Stella.STANDARD_ERROR.nativeStream.print(Stella.exceptionMessage(e));
      }
      Stella.popWorld();
      if ((Stella.$TRACED_KEYWORDS$ != null) &&
          Stella.$TRACED_KEYWORDS$.membP(Logic.KWD_GOAL_TREE)) {
        System.out.println();
      }
      return (result);
    }
  }

  public static TruthValue unaryDescriptionSpecializesDescriptionP(Description sub, Description renamed_Super) {
    if (((Boolean)(Logic.$REVERSEPOLARITYp$.get())).booleanValue()) {
      return (Logic.UNKNOWN_TRUTH_VALUE);
    }
    if (LogicObject.collectionImpliesCollectionP(sub, renamed_Super)) {
      return (Logic.TRUE_TRUTH_VALUE);
    }
    { Object old$PrintreadablyP$000 = Stella.$PRINTREADABLYp$.get();

      try {
        Native.setBooleanSpecial(Stella.$PRINTREADABLYp$, Stella.TRUE);
        if ((Stella.$TRACED_KEYWORDS$ != null) &&
            Stella.$TRACED_KEYWORDS$.membP(Logic.KWD_GOAL_TREE)) {
          {
            System.out.println();
            System.out.println("Prototype Subsumption Test:");
            System.out.println("   sub=" + sub);
            System.out.println("   super=" + renamed_Super);
          }
;
        }

      } finally {
        Stella.$PRINTREADABLYp$.set(old$PrintreadablyP$000);
      }
    }
    { TruthValue result = Logic.UNKNOWN_TRUTH_VALUE;

      try {
        { Object old$Context$000 = Stella.$CONTEXT$.get();

          try {
            Native.setSpecial(Stella.$CONTEXT$, Logic.getInferenceCache(((Module)(Stella.$MODULE$.get())), Logic.KWD_META));
            result = Logic.callAsk(Description.createQueryIterator(renamed_Super, Stella.vector(Stella_Object.cons(Description.getPrototype(sub), Stella.NIL))));

          } finally {
            Stella.$CONTEXT$.set(old$Context$000);
          }
        }
      } catch (LogicException e) {
        Stella.STANDARD_ERROR.nativeStream.print(Stella.exceptionMessage(e));
      }
      if ((Stella.$TRACED_KEYWORDS$ != null) &&
          Stella.$TRACED_KEYWORDS$.membP(Logic.KWD_GOAL_TREE)) {
        System.out.println();
      }
      return (result);
    }
  }

  public static LogicObject getPrototype(Description description) {
    if (description == null) {
      return (null);
    }
    { Stella_Object prototype = Logic.accessBinaryValue(description, Logic.SGT_PL_KERNEL_KB_CONCEPT_PROTOTYPE);

      if (prototype == null) {
        prototype = Logic.applyCachedRetrieve(Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_pC, Stella_Object.cons(Logic.SYM_LOGIC_pP, Stella_Object.cons(Stella.NIL, Stella.NIL)))), Stella.list$(Stella_Object.cons(Logic.SYM_PL_KERNEL_KB_CONCEPT_PROTOTYPE, Stella_Object.cons(Logic.SYM_LOGIC_pC, Stella_Object.cons(Logic.SYM_LOGIC_pP, Stella_Object.cons(Stella.NIL, Stella.NIL))))), Stella.consList(Stella_Object.cons(description, Stella_Object.cons(null, Stella.NIL))), Stella.consList(Stella.NIL), Logic.SYM_LOGIC_QUERY_053, new Object[1]).value;
      }
      { Object old$Evaluationmode$000 = Logic.$EVALUATIONMODE$.get();
        Object old$Queryiterator$000 = Logic.$QUERYITERATOR$.get();

        try {
          Native.setSpecial(Logic.$EVALUATIONMODE$, Logic.KWD_EXTENSIONAL_ASSERTION);
          Native.setSpecial(Logic.$QUERYITERATOR$, null);
          Logic.elaborateMetaInstance(prototype);
          return (((LogicObject)(prototype)));

        } finally {
          Logic.$QUERYITERATOR$.set(old$Queryiterator$000);
          Logic.$EVALUATIONMODE$.set(old$Evaluationmode$000);
        }
      }
    }
  }

  public static QueryIterator createQueryIterator(Description description, Vector outsidebindings) {
    return (QueryIterator.initializeQueryIterator(Logic.allocateQueryIterator(), description, outsidebindings));
  }

  public static boolean expensiveDisjointTermsP(Description d1, Description d2) {
    return (Logic.applyCachedAsk(Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_pD1, Stella_Object.cons(Logic.SYM_LOGIC_pD2, Stella_Object.cons(Stella.NIL, Stella.NIL)))), Stella.list$(Stella_Object.cons(Logic.SYM_STELLA_OR, Stella_Object.cons(Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_DISJOINT, Stella_Object.cons(Logic.SYM_LOGIC_pD1, Stella_Object.cons(Logic.SYM_LOGIC_pD2, Stella_Object.cons(Stella.NIL, Stella.NIL))))), Stella_Object.cons(Stella.list$(Stella_Object.cons(Logic.SYM_STELLA_AND, Stella_Object.cons(Stella.list$(Stella_Object.cons(Logic.SYM_PL_KERNEL_KB_CLASS, Stella_Object.cons(Logic.SYM_LOGIC_pD1, Stella_Object.cons(Stella.NIL, Stella.NIL)))), Stella_Object.cons(Stella.list$(Stella_Object.cons(Logic.SYM_PL_KERNEL_KB_CLASS, Stella_Object.cons(Logic.SYM_LOGIC_pD2, Stella_Object.cons(Stella.NIL, Stella.NIL)))), Stella_Object.cons(Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_REFUTATION_DISJOINT, Stella_Object.cons(Logic.SYM_LOGIC_pD1, Stella_Object.cons(Logic.SYM_LOGIC_pD2, Stella_Object.cons(Stella.NIL, Stella.NIL))))), Stella_Object.cons(Stella.NIL, Stella.NIL)))))), Stella_Object.cons(Stella.NIL, Stella.NIL))))), Stella.consList(Stella_Object.cons(d1, Stella_Object.cons(d2, Stella.NIL))), Stella.consList(Stella.NIL), Logic.SYM_LOGIC_QUERY_051, new Object[1]));
  }

  /** Return TRUE if 'd1' and 'd2' belong to disjoint partitions.
   * @param d1
   * @param d2
   * @return boolean
   */
  public static boolean disjointTermsP(Description d1, Description d2) {
    if (Logic.$CYC_KLUDGES_ENABLEDp$) {
      if (Stella_Object.isaP(d1, Logic.SGT_LOGIC_NAMED_DESCRIPTION) &&
          Stella_Object.isaP(d2, Logic.SGT_LOGIC_NAMED_DESCRIPTION)) {
        return (TruthValue.trueTruthValueP(Logic.evaluateSelectionPattern(Logic.makeRelationPattern2(Logic.SGT_PL_KERNEL_KB_DISJOINT, d1, d2))));
      }
    }
    return (Logic.applyCachedAsk(Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_pD1, Stella_Object.cons(Logic.SYM_LOGIC_pD2, Stella_Object.cons(Stella.NIL, Stella.NIL)))), Stella.list$(Stella_Object.cons(Logic.SYM_LOGIC_DISJOINT, Stella_Object.cons(Logic.SYM_LOGIC_pD1, Stella_Object.cons(Logic.SYM_LOGIC_pD2, Stella_Object.cons(Stella.NIL, Stella.NIL))))), Stella.consList(Stella_Object.cons(d1, Stella_Object.cons(d2, Stella.NIL))), Stella.consList(Stella.NIL), Logic.SYM_LOGIC_QUERY_049, new Object[1]));
  }

  public static boolean checkForSingleValuedGoalP(Description pattern, Cons iobindings) {
    { Cons iovariables = Stella.NIL;
      Cons boundvariables = Stella.NIL;

      { PatternVariable vbl = null;
        Vector vector000 = pattern.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();
        Cons collect000 = null;

        while (index000 < length000) {
          {
            vbl = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          if (collect000 == null) {
            {
              collect000 = Stella_Object.cons(vbl, Stella.NIL);
              if (iovariables == Stella.NIL) {
                iovariables = collect000;
              }
              else {
                Cons.addConsToEndOfConsList(iovariables, collect000);
              }
            }
          }
          else {
            {
              collect000.rest = Stella_Object.cons(vbl, Stella.NIL);
              collect000 = collect000.rest;
            }
          }
        }
      }
      { Stella_Object v2 = null;
        Stella_Object v1 = null;
        Cons iter000 = iovariables;
        Cons iter001 = iobindings;

        while ((!(iter000 == Stella.NIL)) &&
            (!(iter001 == Stella.NIL))) {
          {
            v1 = iter000.value;
            iter000 = iter000.rest;
          }
          {
            v2 = iter001.value;
            iter001 = iter001.rest;
          }
          if (v2 != null) {
            boundvariables = Stella_Object.cons(v1, boundvariables);
          }
        }
      }
      return (Proposition.singleValuedGoalP(pattern.proposition, iovariables, boundvariables));
    }
  }

  public static Description instantiateExternalVariables(Description self, KeyValueList bindings) {
    if (!(bindings.emptyP())) {
      { Object old$Evaluationmode$000 = Logic.$EVALUATIONMODE$.get();
        Object old$Queryiterator$000 = Logic.$QUERYITERATOR$.get();

        try {
          Native.setSpecial(Logic.$EVALUATIONMODE$, Logic.KWD_DESCRIPTION);
          Native.setSpecial(Logic.$QUERYITERATOR$, null);
          { Description description = Description.newDescription();

            description.ioVariables = self.ioVariables.copy();
            { PatternVariable var = null;
              Vector vector000 = self.ioVariables;
              int index000 = 0;
              int length000 = vector000.length();

              while (index000 < length000) {
                {
                  var = ((PatternVariable)((vector000.theArray)[index000]));
                  index000 = index000 + 1;
                }
                bindings.insertAt(var, var);
              }
            }
            { PatternVariable var = null;
              Vector vector001 = ((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null)));
              int index001 = 0;
              int length001 = vector001.length();

              while (index001 < length001) {
                {
                  var = ((PatternVariable)((vector001.theArray)[index001]));
                  index001 = index001 + 1;
                }
                if (bindings.lookup(var) == null) {
                  bindings.insertAt(var, var);
                }
              }
            }
            { PatternVariable var = null;
              Vector vector002 = self.internalVariables;
              int index002 = 0;
              int length002 = vector002.length();

              while (index002 < length002) {
                {
                  var = ((PatternVariable)((vector002.theArray)[index002]));
                  index002 = index002 + 1;
                }
                bindings.insertAt(var, var);
              }
            }
            description.proposition = Proposition.inheritProposition(self.proposition, bindings);
            return (Description.finishBuildingDescription(description, Stella.TRUE));
          }

        } finally {
          Logic.$QUERYITERATOR$.set(old$Queryiterator$000);
          Logic.$EVALUATIONMODE$.set(old$Evaluationmode$000);
        }
      }
    }
    return (self);
  }

  public static PatternVariable variableFromUnbindingOffset(Description description, int uboffset) {
    { PatternRecord patternrecord = ((QueryIterator)(Logic.$QUERYITERATOR$.get())).currentPatternRecord;
      int offset = ((IntegerWrapper)((patternrecord.unbindingStack.theArray)[uboffset])).wrapperValue;

      { PatternVariable vbl = null;
        Vector vector000 = description.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();

        while (index000 < length000) {
          {
            vbl = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          if (vbl.boundToOffset == offset) {
            return (vbl);
          }
        }
      }
      { PatternVariable vbl = null;
        Vector vector001 = description.internalVariables;
        int index001 = 0;
        int length001 = vector001.length();

        while (index001 < length001) {
          {
            vbl = ((PatternVariable)((vector001.theArray)[index001]));
            index001 = index001 + 1;
          }
          if (vbl.boundToOffset == offset) {
            return (vbl);
          }
        }
      }
      return (null);
    }
  }

  public static Cons helpGetForwardChainingRules(Description description) {
    { Cons rules = Stella.NIL;

      { ForwardChainingIndex idx = null;
        Cons iter000 = description.forwardChainingIndices().theConsList;
        Cons collect000 = null;

        while (!(iter000 == Stella.NIL)) {
          {
            idx = ((ForwardChainingIndex)(iter000.value));
            iter000 = iter000.rest;
          }
          if (!idx.masterRule.deletedP()) {
            if (collect000 == null) {
              {
                collect000 = Stella_Object.cons(idx.masterRule, Stella.NIL);
                if (rules == Stella.NIL) {
                  rules = collect000;
                }
                else {
                  Cons.addConsToEndOfConsList(rules, collect000);
                }
              }
            }
            else {
              {
                collect000.rest = Stella_Object.cons(idx.masterRule, Stella.NIL);
                collect000 = collect000.rest;
              }
            }
          }
        }
      }
      return (rules);
    }
  }

  public static Description getInferableComplementDescription(Description self) {
    { Description complement = ((Description)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null)));

      if (Logic.$LAZY_SATELLITE_RULESp$) {
        if ((complement == null) &&
            BooleanWrapper.coerceWrappedBooleanToBoolean(self.deferredContrapositivesP())) {
          complement = NamedDescription.getComplementOfGoalDescription(((NamedDescription)(self)));
        }
        if (BooleanWrapper.coerceWrappedBooleanToBoolean(self.deferredContrapositivesP())) {
          Description.deriveDeferredSatelliteRules(self);
          KeyValueList.setDynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_DEFERRED_CONTRAPOSITIVESp, null, null);
        }
      }
      return (complement);
    }
  }

  public static void deriveDeferredContrapositiveSatelliteRules(Description self) {
    if (Logic.$LAZY_SATELLITE_RULESp$) {
      if (BooleanWrapper.coerceWrappedBooleanToBoolean(self.deferredContrapositivesP())) {
        { Proposition masterproposition = null;
          List queuedrules = List.newList();

          { Proposition prop = null;
            Iterator iter000 = Logic.unfilteredDependentImpliesPropositions(self).allocateIterator();

            while (iter000.nextP()) {
              prop = ((Proposition)(iter000.value));
              if (prop.kind == Logic.KWD_IMPLIES) {
                masterproposition = ((Proposition)(KeyValueList.dynamicSlotValue(prop.dynamicSlots, Logic.SYM_LOGIC_MASTER_PROPOSITION, null)));
                if ((masterproposition != null) &&
                    (!masterproposition.deletedP())) {
                  queuedrules.insertNew(masterproposition);
                }
              }
            }
          }
          { Proposition forallprop = null;
            Cons iter001 = queuedrules.theConsList;

            while (!(iter001 == Stella.NIL)) {
              {
                forallprop = ((Proposition)(iter001.value));
                iter001 = iter001.rest;
              }
              Proposition.deriveSatelliteRulesForGoalP(forallprop, self, Logic.KWD_CONTRAPOSITIVE, Stella.FALSE);
            }
          }
        }
      }
    }
  }

  public static void deriveDeferredSatelliteRules(Description self) {
    if (Logic.$LAZY_SATELLITE_RULESp$) {
      { Object old$DeriveDeferredSatelliteRulesInvocations$000 = Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$.get();

        try {
          Native.setSpecial(Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$, ((List)(((((List)(Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$.get())) == null) ? List.newList() : ((List)(Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$.get()))))));
          if (((List)(Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$.get())).membP(self)) {
            return;
          }
          ((List)(Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$.get())).insert(self);
          if ((((SequenceIndex)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_RULES_WITH_DEFERRED_SATELLITES, null))) == null) &&
              (((Description)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null))) == null)) {
            return;
          }
          { SequenceIndex queuedrules = ((SequenceIndex)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_RULES_WITH_DEFERRED_SATELLITES, null)));
            Description complement = ((Description)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null)));

            if (queuedrules != null) {
              { Proposition forallprop = null;
                Iterator iter000 = queuedrules.allocateIterator();

                while (iter000.nextP()) {
                  forallprop = ((Proposition)(iter000.value));
                  Proposition.deriveSatelliteRulesForGoalP(forallprop, self, ((complement == null) ? Logic.KWD_POSITIVE : Logic.KWD_POSITIVE_AND_CONTRAPOSITIVE), Stella.FALSE);
                }
              }
              if (complement != null) {
                KeyValueList.setDynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_DEFERRED_CONTRAPOSITIVESp, null, null);
              }
              KeyValueList.setDynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_RULES_WITH_DEFERRED_SATELLITES, null, null);
            }
            if ((complement != null) &&
                (((SequenceIndex)(KeyValueList.dynamicSlotValue(complement.dynamicSlots, Logic.SYM_LOGIC_RULES_WITH_DEFERRED_SATELLITES, null))) != null)) {
              Description.deriveDeferredSatelliteRules(complement);
            }
          }

        } finally {
          Logic.$DERIVE_DEFERRED_SATELLITE_RULES_INVOCATIONS$.set(old$DeriveDeferredSatelliteRulesInvocations$000);
        }
      }
    }
  }

  public static boolean atomicGoalDescriptionP(Description description) {
    if (Description.namedDescriptionP(description)) {
      return (Stella.TRUE);
    }
    { Description complement = ((Description)(KeyValueList.dynamicSlotValue(description.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null)));

      if ((complement != null) &&
          Description.namedDescriptionP(complement)) {
        return (Stella.TRUE);
      }
      else {
        return (Proposition.atomicGoalPropositionP(description.proposition));
      }
    }
  }

  public static Proposition extractProposition(Description self, KeyValueList mapping) {
    { Proposition proposition = self.proposition;
      Cons existentials = ((self.internalVariables.arraySize > 0) ? Description.topLevelExistentialVariables(self) : Stella.NIL);
      Proposition existsproposition = ((!(existentials == Stella.NIL)) ? Logic.createProposition(Logic.SYM_STELLA_EXISTS, 1) : ((Proposition)(null)));

      if (existsproposition != null) {
        KeyValueList.setDynamicSlotValue(existsproposition.dynamicSlots, Logic.SYM_LOGIC_IO_VARIABLES, Logic.copyVariablesVector(Logic.copyConsListToVariablesVector(existentials), mapping), null);
      }
      proposition = Proposition.copyProposition(proposition, mapping);
      mapping.removeAt(self.proposition);
      if (existsproposition != null) {
        (existsproposition.arguments.theArray)[0] = proposition;
        proposition = existsproposition;
      }
      return (proposition);
    }
  }

  public static KeyValueList inferVariableTypesInDescription(Description description) {
    { Cons visiblevariables = Stella.NIL;

      { PatternVariable v = null;
        Vector vector000 = description.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();

        while (index000 < length000) {
          {
            v = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          visiblevariables = Stella_Object.cons(v, visiblevariables);
        }
      }
      { PatternVariable v = null;
        Cons iter000 = Description.topLevelExistentialVariables(description);

        while (!(iter000 == Stella.NIL)) {
          {
            v = ((PatternVariable)(iter000.value));
            iter000 = iter000.rest;
          }
          visiblevariables = Stella_Object.cons(v, visiblevariables);
        }
      }
      return (Proposition.inferVariableTypesInProposition(description.proposition, visiblevariables));
    }
  }

  public static Cons topLevelExistentialVariables(Description self) {
    { List quantifiedvariables = List.newList();
      List beenthere = List.newList();
      List collection = List.newList();

      { PatternVariable v = null;
        Vector vector000 = self.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();

        while (index000 < length000) {
          {
            v = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          quantifiedvariables.insert(v);
        }
      }
      if (((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))) != null) {
        { PatternVariable v = null;
          Vector vector001 = ((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null)));
          int index001 = 0;
          int length001 = vector001.length();

          while (index001 < length001) {
            {
              v = ((PatternVariable)((vector001.theArray)[index001]));
              index001 = index001 + 1;
            }
            quantifiedvariables.insert(v);
          }
        }
      }
      if (self.proposition == null) {
        return (quantifiedvariables.theConsList);
      }
      Logic.collectFreeVariables(self.proposition, collection, quantifiedvariables, beenthere);
      quantifiedvariables.free();
      beenthere.free();
      return (collection.theConsList);
    }
  }

  public static Description finishBuildingDescription(Description description, boolean checkforduplicateP) {
    { Proposition proposition = description.proposition;

      Proposition.normalizeProposition(proposition);
      if (proposition.kind == Logic.KWD_EXISTS) {
        proposition = ((Proposition)((proposition.arguments.theArray)[0]));
        description.proposition = proposition;
      }
      Logic.resolveUnresolvedSlotReferences(description);
      Proposition.recursivelyFastenDownPropositions(proposition, Stella.FALSE);
      Proposition.updateSkolemTypeFromIsaAssertions(proposition);
      try {
        Proposition.equateTopLevelEquivalences(proposition);
      } catch (Clash e) {
        Stella.STANDARD_ERROR.nativeStream.println(Stella.exceptionMessage(e));
      }
    }
    Description.collapseValueOfChains(description);
    Description.simplifyDescription(description, Stella.FALSE);
    Description.computeInternalVariables(description);
    if (checkforduplicateP) {
      { Description duplicatedescription = Description.findDuplicateDescription(description);

        if (duplicatedescription != null) {
          description.free();
          return (duplicatedescription);
        }
      }
    }
    return (description);
  }

  public static void computeInternalVariables(Description self) {
    { List collection = List.newList();
      List beenthere = List.newList();

      Proposition.collectAllVariables(self.proposition, collection, beenthere);
      { PatternVariable vbl = null;
        Vector vector000 = self.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();

        while (index000 < length000) {
          {
            vbl = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          collection.remove(vbl);
        }
      }
      if (((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))) != null) {
        { PatternVariable vbl = null;
          Vector vector001 = ((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null)));
          int index001 = 0;
          int length001 = vector001.length();

          while (index001 < length001) {
            {
              vbl = ((PatternVariable)((vector001.theArray)[index001]));
              index001 = index001 + 1;
            }
            collection.remove(vbl);
          }
        }
      }
      self.internalVariables = Logic.copyConsListToVariablesVector(collection.theConsList);
      collection.free();
      beenthere.free();
    }
  }

  public static void collapseValueOfChains(Description description) {
    Proposition.tightenArgumentBindings(description.proposition, description.ioVariables);
    Logic.collapseValueOfChainsForIoVariables(description.ioVariables);
    { Stella_Object value = null;

      { int i = Stella.NULL_INTEGER;
        PatternVariable vbl = null;
        Vector vector000 = description.internalVariables;
        int index000 = 0;
        int length000 = vector000.length();
        int iter000 = 0;

        while (index000 < length000) {
          {
            vbl = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          {
            i = iter000;
            iter000 = iter000 + 1;
          }
          if ((((Stella_Object)(Stella_Object.accessInContext(vbl.variableValue, vbl.homeContext, Stella.FALSE))) != null) &&
              Logic.variableP(((Stella_Object)(Stella_Object.accessInContext(vbl.variableValue, vbl.homeContext, Stella.FALSE))))) {
            value = ((Stella_Object)(Stella_Object.accessInContext(vbl.variableValue, vbl.homeContext, Stella.FALSE)));
            (description.internalVariables.theArray)[i] = null;
            if (Surrogate.subtypeOfP(Stella_Object.safePrimaryType(value), Logic.SGT_LOGIC_LOGIC_OBJECT)) {
              { LogicObject value000 = ((LogicObject)(value));

                if (!(description.ioVariables.memberP(value000.variableValueInverse()))) {
                  { LogicObject object000 = value000;
                    Cons value001 = Stella.NIL;
                    Stella_Object oldValue003 = object000.variableValueInverse;
                    Stella_Object newValue000 = Stella_Object.updateInContext(oldValue003, value001, object000.homeContext, Stella.FALSE);

                    if (!((oldValue003 != null) &&
                        (oldValue003.primaryType() == Logic.SGT_STELLA_CS_VALUE))) {
                      object000.variableValueInverse = newValue000;
                    }
                  }
                }
              }
            }
            else {
            }
            vbl.free();
          }
        }
      }
      { int i = Stella.NULL_INTEGER;
        PatternVariable vbl = null;
        Vector vector001 = description.internalVariables;
        int index001 = 0;
        int length001 = vector001.length();
        int iter001 = 0;

        while (index001 < length001) {
          {
            vbl = ((PatternVariable)((vector001.theArray)[index001]));
            index001 = index001 + 1;
          }
          {
            i = iter001;
            iter001 = iter001 + 1;
          }
          if (description.ioVariables.memberP(vbl)) {
            value = vbl;
            (description.internalVariables.theArray)[i] = null;
          }
        }
      }
      description.internalVariables = Logic.removeNullsInVariablesVector(description.internalVariables);
    }
  }

  public static void fastenDownDescription(Description self) {
    { LogicObject ref = null;
      Cons iter000 = Description.collectConstantDescriptionReferences(self, null, Stella.FALSE).theConsList;

      while (!(iter000 == Stella.NIL)) {
        {
          ref = ((LogicObject)(iter000.value));
          iter000 = iter000.rest;
        }
        Logic.addDependentDescriptionLink(ref, self);
      }
    }
  }

  public static Description findDuplicateDescription(Description self) {
    { Description duplicate = null;

      if (Description.namedDescriptionP(self)) {
        {
          return (null);
        }
      }
      else {
        { 
          duplicate = ((Description)(self.findDuplicateNamedDescription()));
          if (duplicate != null) {
            {
              return (duplicate);
            }
          }
          else {
            { 
              duplicate = Description.findExactDuplicateDescription(self);
              if (duplicate != null) {
                {
                  return (duplicate);
                }
              }
              else {
                {
                  return (null);
                }
              }
            }
          }
        }
      }
    }
  }

  public static Description findDuplicateDescriptionWithConstantReference(Description self) {
    { LogicObject referencedconstant = ((LogicObject)(Description.collectConstantDescriptionReferences(self, null, Stella.TRUE).first()));

      if (referencedconstant != null) {
        { KeyValueList mapping = null;

          if (((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null))) != null) {
            if (((QueryIterator)(Logic.$QUERYITERATOR$.get())) == null) {
              return (null);
            }
            { PatternVariable v = null;
              Vector vector000 = ((Vector)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_EXTERNAL_VARIABLES, null)));
              int index000 = 0;
              int length000 = vector000.length();

              while (index000 < length000) {
                {
                  v = ((PatternVariable)((vector000.theArray)[index000]));
                  index000 = index000 + 1;
                }
                { Stella_Object binding = Logic.argumentBoundTo(v);

                  if (binding == null) {
                    return (null);
                  }
                  mapping.insertAt(v, binding);
                }
              }
            }
          }
          { Description dep = null;
            Cons iter000 = referencedconstant.dependentDescriptions().theConsList;

            while (!(iter000 == Stella.NIL)) {
              {
                dep = ((Description)(iter000.value));
                iter000 = iter000.rest;
              }
              if ((!dep.deletedP()) &&
                  Description.equivalentDescriptionsP(dep, self, mapping)) {
                return (dep);
              }
            }
          }
        }
      }
      return (null);
    }
  }

  public static List collectConstantDescriptionReferences(Description self, List collection, boolean stopatoneP) {
    if (collection == null) {
      collection = List.newList();
    }
    Logic.helpCollectConstantReferences(self.proposition, collection, stopatoneP);
    return (collection);
  }

  public static Description findExactDuplicateDescription(Description self) {
    Description.uniquifyDescriptionVariables(self);
    { String stringification = Native.stringify(Logic.uniquifyPropositionTree(Proposition.generateProposition(self.proposition)));
      List bucket = ((List)(Logic.$STRINGIFIED_DESCRIPTION_INDEX$.lookup(stringification)));

      Description.restoreDescriptionVariableNames(self);
      if (bucket == null) {
        Logic.$STRINGIFIED_DESCRIPTION_INDEX$.insertAt(stringification, Stella.list(Stella_Object.cons(self, Stella.NIL)));
        return (null);
      }
      bucket.removeIf(Native.find_java_method("edu.isi.powerloom.logic.LogicObject", "deletedLogicObjectP", new java.lang.Class [] {Native.find_java_class("edu.isi.powerloom.logic.LogicObject")}));
      { Description d = null;
        Cons iter000 = bucket.theConsList;

        while (!(iter000 == Stella.NIL)) {
          {
            d = ((Description)(iter000.value));
            iter000 = iter000.rest;
          }
          if (Description.equivalentDescriptionsP(self, d, null) &&
              ((self.homeContext == d.homeContext) ||
               self.homeContext.allSuperContexts.membP(d.homeContext))) {
            return (d);
          }
        }
      }
      bucket.push(self);
      return (null);
    }
  }

  public static void restoreDescriptionVariableNames(Description self) {
    if (self.ioVariables.length() == 0) {
      return;
    }
    if (((Symbol)(KeyValueList.dynamicSlotValue(((PatternVariable)((self.ioVariables.theArray)[0])).dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null))) == null) {
      return;
    }
    { PatternVariable v = null;
      Vector vector000 = self.ioVariables;
      int index000 = 0;
      int length000 = vector000.length();

      while (index000 < length000) {
        {
          v = ((PatternVariable)((vector000.theArray)[index000]));
          index000 = index000 + 1;
        }
        if (((Symbol)(KeyValueList.dynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null))) != null) {
          v.skolemName = ((Symbol)(KeyValueList.dynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null)));
          KeyValueList.setDynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null, null);
        }
      }
    }
    { PatternVariable v = null;
      Vector vector001 = self.internalVariables;
      int index001 = 0;
      int length001 = vector001.length();

      while (index001 < length001) {
        {
          v = ((PatternVariable)((vector001.theArray)[index001]));
          index001 = index001 + 1;
        }
        if (((Symbol)(KeyValueList.dynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null))) != null) {
          v.skolemName = ((Symbol)(KeyValueList.dynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null)));
          KeyValueList.setDynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null, null);
        }
      }
    }
  }

  public static void uniquifyDescriptionVariables(Description self) {
    { Cons variablenames = Logic.SYSTEM_DEFINED_ARGUMENT_NAMES;

      if ((self.ioVariables.length() + self.internalVariables.length()) > variablenames.length()) {
        return;
      }
      { PatternVariable v = null;
        Vector vector000 = self.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();

        while (index000 < length000) {
          {
            v = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          if (((Symbol)(KeyValueList.dynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null))) == null) {
            KeyValueList.setDynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, v.skolemName, null);
            { Symbol head000 = ((Symbol)(variablenames.value));

              variablenames = variablenames.rest;
              v.skolemName = head000;
            }
          }
        }
      }
      { PatternVariable v = null;
        Vector vector001 = self.internalVariables;
        int index001 = 0;
        int length001 = vector001.length();

        while (index001 < length001) {
          {
            v = ((PatternVariable)((vector001.theArray)[index001]));
            index001 = index001 + 1;
          }
          if (((Symbol)(KeyValueList.dynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, null))) == null) {
            KeyValueList.setDynamicSlotValue(v.dynamicSlots, Logic.SYM_LOGIC_ORIGINAL_SKOLEM_NAME, v.skolemName, null);
            { Symbol head001 = ((Symbol)(variablenames.value));

              variablenames = variablenames.rest;
              v.skolemName = head001;
            }
          }
        }
      }
    }
  }

  public LogicObject findDuplicateNamedDescription() {
    { Description self = this;

      if (Description.namedDescriptionP(self)) {
        return (self);
      }
      { Proposition proposition = self.proposition;

        { Keyword testValue000 = proposition.kind;

          if (testValue000 == Logic.KWD_AND) {
            { Proposition onlygoal = null;

              { Stella_Object arg = null;
                Vector vector000 = proposition.arguments;
                int index000 = 0;
                int length000 = vector000.length();

                while (index000 < length000) {
                  {
                    arg = (vector000.theArray)[index000];
                    index000 = index000 + 1;
                  }
                  if (BooleanWrapper.coerceWrappedBooleanToBoolean(((Proposition)(arg)).variableTypeP())) {
                  }
                  else if (onlygoal != null) {
                    return (null);
                  }
                  else {
                    onlygoal = ((Proposition)(arg));
                  }
                }
              }
              if ((onlygoal != null) &&
                  Logic.sameArgumentsP(self.ioVariables, ((onlygoal.kind == Logic.KWD_NOT) ? ((Proposition)((onlygoal.arguments.theArray)[0])).arguments : onlygoal.arguments))) {
                return (Proposition.extractGoalDescription(onlygoal, null));
              }
            }
          }
          else if (testValue000 == Logic.KWD_NOT) {
            if (Logic.sameArgumentsP(self.ioVariables, ((Proposition)((proposition.arguments.theArray)[0])).arguments)) {
              return (Proposition.extractGoalDescription(proposition, null));
            }
          }
          else {
            if (Logic.sameArgumentsP(self.ioVariables, proposition.arguments)) {
              return (Proposition.extractGoalDescription(proposition, null));
            }
          }
        }
        return (null);
      }
    }
  }

  public static boolean namedDescriptionP(Description self) {
    return (self.surrogateValueInverse != null);
  }

  public static boolean equivalentDescriptionsP(Description self, Description other, KeyValueList mapping) {
    if (self == other) {
      return (Stella.TRUE);
    }
    if ((self.descriptionName() != null) &&
        (other.descriptionName() != null)) {
      return (Stella.FALSE);
    }
    if (!(self.arity() == other.arity())) {
      return (Stella.FALSE);
    }
    if (Description.namedDescriptionP(self)) {
      return (Stella.FALSE);
    }
    else {
      if (mapping == null) {
        mapping = KeyValueList.newKeyValueList();
      }
      { PatternVariable v2 = null;
        PatternVariable v1 = null;
        Vector vector000 = self.ioVariables;
        int index000 = 0;
        int length000 = vector000.length();
        Vector vector001 = other.ioVariables;
        int index001 = 0;
        int length001 = vector001.length();

        while ((index000 < length000) &&
            (index001 < length001)) {
          {
            v1 = ((PatternVariable)((vector000.theArray)[index000]));
            index000 = index000 + 1;
          }
          {
            v2 = ((PatternVariable)((vector001.theArray)[index001]));
            index001 = index001 + 1;
          }
          mapping.insertAt(v1, v2);
        }
      }
      { PatternVariable v2 = null;
        PatternVariable v1 = null;
        Vector vector002 = self.internalVariables;
        int index002 = 0;
        int length002 = vector002.length();
        Vector vector003 = other.internalVariables;
        int index003 = 0;
        int length003 = vector003.length();

        while ((index002 < length002) &&
            (index003 < length003)) {
          {
            v1 = ((PatternVariable)((vector002.theArray)[index002]));
            index002 = index002 + 1;
          }
          {
            v2 = ((PatternVariable)((vector003.theArray)[index003]));
            index003 = index003 + 1;
          }
          mapping.insertAt(v1, v2);
        }
      }
      return (Logic.equivalentFormulaeP(self.proposition, other.proposition, mapping));
    }
  }

  public static void ensureDescriptionBody(Description description) {
    if (((description.proposition == null) ||
        (description.proposition == Logic.TRUE_PROPOSITION)) &&
        Stella_Object.isaP(description, Logic.SGT_LOGIC_NAMED_DESCRIPTION)) {
      NamedDescription.materializePrimitiveDescriptionBody(((NamedDescription)(description)));
    }
  }

  public static void createBaseRelationForPolymorphicDescription(Description description) {
    { Surrogate surrogate = Stella.internSurrogateInModule(description.descriptionName().symbolName, ((Module)(description.homeContext)), Stella.TRUE);
      Stella_Object relation = surrogate.surrogateValue;

      if (relation == null) {
        { Object old$Evaluationmode$000 = Logic.$EVALUATIONMODE$.get();
          Object old$FillingconstraintpropagationqueuesP$000 = Logic.$FILLINGCONSTRAINTPROPAGATIONQUEUESp$.get();

          try {
            Native.setSpecial(Logic.$EVALUATIONMODE$, Logic.KWD_EXTENSIONAL_ASSERTION);
            Native.setBooleanSpecial(Logic.$FILLINGCONSTRAINTPROPAGATIONQUEUESp$, Stella.FALSE);
            Logic.assertIsaProposition(Logic.createLogicInstance(surrogate, Logic.SGT_PL_KERNEL_KB_RELATION), Logic.SGT_PL_KERNEL_KB_POLYMORPHIC);

          } finally {
            Logic.$FILLINGCONSTRAINTPROPAGATIONQUEUESp$.set(old$FillingconstraintpropagationqueuesP$000);
            Logic.$EVALUATIONMODE$.set(old$Evaluationmode$000);
          }
        }
      }
    }
  }

  /** Return the name of the description <code>self</code>, if it has one.
   * @return Symbol
   */
  public Symbol descriptionName() {
    { Description self = this;

      return (null);
    }
  }

  public static void applyForwardRulesToVector(Description triggerdescription, Vector arguments, Proposition triggerproposition) {
    { Vector evaluatedargs = arguments;

      { int i = Stella.NULL_INTEGER;
        Stella_Object arg = null;
        Vector vector000 = arguments;
        int index000 = 0;
        int length000 = vector000.length();
        int iter000 = 0;

        while (index000 < length000) {
          {
            arg = (vector000.theArray)[index000];
            index000 = index000 + 1;
          }
          {
            i = iter000;
            iter000 = iter000 + 1;
          }
          if (arg.deletedP()) {
            return;
          }
          else if (!Stella_Object.eqlP(arg, Logic.valueOf(arg))) {
            if (arguments == evaluatedargs) {
              evaluatedargs = arguments.copy();
            }
            (evaluatedargs.theArray)[i] = (Logic.valueOf(arg));
          }
        }
      }
      arguments = evaluatedargs;
    }
    { KeyValueList rules = null;
      KeyValueList indices = null;

      { Object [] caller_MV_returnarray = new Object[1];

        rules = Description.collectForwardChainingRules(triggerdescription, caller_MV_returnarray);
        indices = ((KeyValueList)(caller_MV_returnarray[0]));
      }
      { BooleanWrapper toucheddefaultP = null;
        Proposition r = null;
        KvCons iter001 = rules.theKvList;

        while (iter001 != null) {
          r = ((Proposition)(iter001.key));
          {
            toucheddefaultP = ((BooleanWrapper)(iter001.value));
            iter001 = iter001.rest;
          }
          if (Proposition.applicableForwardRuleP(r, arguments)) {
            Description.applyRuleConsequentToVector(((Description)((r.arguments.theArray)[1])), arguments, r, triggerdescription, triggerproposition, BooleanWrapper.coerceWrappedBooleanToBoolean(toucheddefaultP));
          }
        }
      }
      { BooleanWrapper toucheddefaultP = null;
        ForwardChainingIndex fwdindex = null;
        KvCons iter002 = indices.theKvList;

        loop002 : while (iter002 != null) {
          fwdindex = ((ForwardChainingIndex)(iter002.key));
          {
            toucheddefaultP = ((BooleanWrapper)(iter002.value));
            iter002 = iter002.rest;
          }
          { Cons inputargs = Stella.NIL;

            if (!(arguments.length() == fwdindex.forwardGoal.arguments.length())) {
              continue loop002;
            }
            { Stella_Object goalarg = null;
              Stella_Object arg = null;
              Vector vector001 = arguments;
              int index001 = 0;
              int length001 = vector001.length();
              Vector vector002 = fwdindex.forwardGoal.arguments;
              int index002 = 0;
              int length002 = vector002.length();

              loop003 : while ((index001 < length001) &&
                  (index002 < length002)) {
                {
                  arg = (vector001.theArray)[index001];
                  index001 = index001 + 1;
                }
                {
                  goalarg = (vector002.theArray)[index002];
                  index002 = index002 + 1;
                }
                if (Logic.variableP(goalarg)) {
                  inputargs = Stella_Object.cons(arg, inputargs);
                }
                else if (!Stella_Object.eqlP(arg, Logic.valueOf(goalarg))) {
                  inputargs = null;
                  break loop003;
                }
              }
            }
            if (inputargs == null) {
              continue loop002;
            }
            { Stella_Object arg = null;
              ConsIterator it = fwdindex.inputBindings.allocateIterator();
              Cons iter003 = inputargs.reverse();

              while (it.nextP() &&
                  (!(iter003 == Stella.NIL))) {
                {
                  arg = iter003.value;
                  iter003 = iter003.rest;
                }
                {
                }
                it.valueSetter(arg);
              }
            }
          }
          { Cons outputbindings = null;
            Cons truthvalues = null;

            { Object [] caller_MV_returnarray = new Object[1];

              outputbindings = Logic.applyCachedRetrieve(fwdindex.ioVariables, fwdindex.queryBody, fwdindex.inputBindings, Stella.consList(Stella_Object.cons(Logic.KWD_SINGLETONSp, Stella_Object.cons(Stella.TRUE_WRAPPER, Stella.NIL))), fwdindex.cacheId, caller_MV_returnarray);
              truthvalues = ((Cons)(caller_MV_returnarray[0]));
            }
            { Stella_Object tv = null;
              Stella_Object ob = null;
              Cons iter004 = outputbindings;
              Cons iter005 = truthvalues;

              while ((!(iter004 == Stella.NIL)) &&
                  (!(iter005 == Stella.NIL))) {
                {
                  ob = iter004.value;
                  iter004 = iter004.rest;
                }
                {
                  tv = iter005.value;
                  iter005 = iter005.rest;
                }
                Description.applyRuleConsequentToVector(fwdindex.consequent, Logic.copyListToArgumentsVector(((Cons)(ob)).listify()), fwdindex.masterRule, triggerdescription, triggerproposition, BooleanWrapper.coerceWrappedBooleanToBoolean(toucheddefaultP) ||
                    (tv == Logic.DEFAULT_TRUE_TRUTH_VALUE));
              }
            }
          }
        }
      }
    }
  }

  public static void applyRuleConsequentToVector(Description consequent, Vector arguments, Proposition rule, Description triggerdescription, Proposition triggerproposition, boolean toucheddefaultP) {
    { Object old$Collectforwardpropositions$000 = Logic.$COLLECTFORWARDPROPOSITIONS$.get();

      try {
        Native.setSpecial(Logic.$COLLECTFORWARDPROPOSITIONS$, Stella.NIL);
        { boolean toucheddefaultknowledgeP = Proposition.defaultTrueP(triggerproposition) ||
              (Proposition.defaultTrueP(rule) ||
               toucheddefaultP);

          { Object old$Queryiterator$000 = Logic.$QUERYITERATOR$.get();

            try {
              Native.setSpecial(Logic.$QUERYITERATOR$, null);
              Logic.inheritDescription(arguments, consequent, toucheddefaultknowledgeP);

            } finally {
              Logic.$QUERYITERATOR$.set(old$Queryiterator$000);
            }
          }
        }
        { Proposition p = null;
          Cons iter000 = ((Cons)(Logic.$COLLECTFORWARDPROPOSITIONS$.get()));

          while (!(iter000 == Stella.NIL)) {
            {
              p = ((Proposition)(iter000.value));
              iter000 = iter000.rest;
            }
            if (Stella_Object.traceKeywordP(Logic.KWD_PROPAGATE)) {
              System.out.println("MODUS PONENS:");
              System.out.print("   ANTECEDENT:  (");
              if (Description.namedDescriptionP(triggerdescription)) {
                System.out.print(triggerdescription.descriptionName());
              }
              else {
                System.out.print("not-" + ((Description)(KeyValueList.dynamicSlotValue(triggerdescription.dynamicSlots, Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION, null))).descriptionName());
              }
              { Stella_Object arg = null;
                Vector vector000 = arguments;
                int index000 = 0;
                int length000 = vector000.length();

                while (index000 < length000) {
                  {
                    arg = (vector000.theArray)[index000];
                    index000 = index000 + 1;
                  }
                  System.out.print(" ");
                  Logic.printFormula(arg, 15);
                }
              }
              System.out.println(")");
              System.out.print("   RULE:  ");
              Logic.printFormula(rule, 10);
              System.out.println();
              System.out.print("   CONSEQUENT: ");
              Logic.printFormula(p, 15);
              System.out.println();
            }
            Proposition.recordForwardGoal(rule, arguments, p);
            Proposition.recordForwardJustification(p, rule, Stella.consList(Stella_Object.cons(triggerproposition, Stella.NIL)));
          }
        }

      } finally {
        Logic.$COLLECTFORWARDPROPOSITIONS$.set(old$Collectforwardpropositions$000);
      }
    }
  }

  public static boolean hasForwardChainingRulesP(Description description, Proposition proposition) {
    if (Proposition.falseP(proposition)) {
      description = Description.getInferableComplementDescription(description);
      if (description == null) {
        return (Stella.FALSE);
      }
    }
    { KeyValueList rules = null;
      KeyValueList indices = null;

      { Object [] caller_MV_returnarray = new Object[1];

        rules = Description.collectForwardChainingRules(description, caller_MV_returnarray);
        indices = ((KeyValueList)(caller_MV_returnarray[0]));
      }
      return (rules.nonEmptyP() ||
          indices.nonEmptyP());
    }
  }

  public static KeyValueList collectForwardChainingRules(Description description, Object [] MV_returnarray) {
    { MemoizationTable memoTable000 = ((MemoizationTable)(Logic.SGT_LOGIC_COLLECT_FORWARD_CHAINING_RULES_MEMO_TABLE_000.surrogateValue));
      Cons memoizedEntry000 = null;
      Stella_Object memoizedValue000 = null;

      if (memoTable000 == null) {
        Surrogate.initializeMemoizationTable(Logic.SGT_LOGIC_COLLECT_FORWARD_CHAINING_RULES_MEMO_TABLE_000, "(:MAX-VALUES 500 :TIMESTAMPS (:META-KB-UPDATE))");
        memoTable000 = ((MemoizationTable)(Logic.SGT_LOGIC_COLLECT_FORWARD_CHAINING_RULES_MEMO_TABLE_000.surrogateValue));
      }
      if (Stella.$MEMOIZATION_ENABLEDp$) {
        memoizedEntry000 = MruMemoizationTable.lookupMruMemoizedValue(((MruMemoizationTable)(memoTable000)), description, ((Context)(Stella.$CONTEXT$.get())), Stella.MEMOIZED_NULL_VALUE, null, -1);
        memoizedValue000 = memoizedEntry000.value;
      }
      if (memoizedValue000 != null) {
        if (memoizedValue000 == Stella.MEMOIZED_NULL_VALUE) {
          memoizedValue000 = null;
        }
      }
      else {
        memoizedValue000 = Description.helpCollectForwardRules(description, null, null, Stella.FALSE, null);
        if (Stella.$MEMOIZATION_ENABLEDp$) {
          memoizedEntry000.value = ((memoizedValue000 == null) ? Stella.MEMOIZED_NULL_VALUE : memoizedValue000);
        }
      }
      { Cons rules = ((Cons)(memoizedValue000));

        { KeyValueList _return_temp = ((KeyValueList)(rules.value));

          MV_returnarray[0] = ((KeyValueList)(rules.rest.value));
          return (_return_temp);
        }
      }
    }
  }

  public static Cons helpCollectForwardRules(Description description, KeyValueList rules, KeyValueList indices, boolean toucheddefaultP, List beenthere) {
    if (rules == null) {
      rules = KeyValueList.newKeyValueList();
    }
    if (indices == null) {
      indices = KeyValueList.newKeyValueList();
    }
    if (beenthere == null) {
      beenthere = List.newList();
    }
    beenthere.insert(description);
    { ForwardChainingIndex index = null;
      Cons iter000 = description.forwardChainingIndices().theConsList;

      while (!(iter000 == Stella.NIL)) {
        {
          index = ((ForwardChainingIndex)(iter000.value));
          iter000 = iter000.rest;
        }
        if ((!index.masterRule.deletedP()) &&
            Proposition.trueP(index.masterRule)) {
          indices.insertAt(index, (toucheddefaultP ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER));
        }
      }
    }
    { Proposition p = null;
      Cons iter001 = LogicObject.applicableRulesOfDescription(description, Logic.KWD_FORWARD, Stella.TRUE);

      while (!(iter001 == Stella.NIL)) {
        {
          p = ((Proposition)(iter001.value));
          iter001 = iter001.rest;
        }
        if (!((BooleanWrapper)(KeyValueList.dynamicSlotValue(p.dynamicSlots, Logic.SYM_LOGIC_BACKWARD_ONLYp, Stella.FALSE_WRAPPER))).wrapperValue) {
          { Stella_Object consequent = (p.arguments.theArray)[1];

            { Surrogate testValue000 = Stella_Object.safePrimaryType(consequent);

              if (Surrogate.subtypeOfP(testValue000, Logic.SGT_LOGIC_NAMED_DESCRIPTION)) {
                { NamedDescription consequent000 = ((NamedDescription)(consequent));

                  { Stella_Object renamed_Super = Logic.valueOf(consequent000);

                    if (Proposition.defaultTrueP(p)) {
                      toucheddefaultP = Stella.TRUE;
                    }
                    if (!(Logic.classP(description) == Logic.classP(renamed_Super))) {
                      rules.insertAt(p, (toucheddefaultP ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER));
                    }
                    if (!beenthere.memberP(renamed_Super)) {
                      Description.helpCollectForwardRules(((Description)(renamed_Super)), rules, indices, toucheddefaultP, beenthere);
                    }
                  }
                }
              }
              else if (Surrogate.subtypeOfP(testValue000, Logic.SGT_LOGIC_DESCRIPTION)) {
                { Description consequent000 = ((Description)(consequent));

                  rules.insertAt(p, (toucheddefaultP ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER));
                }
              }
              else {
              }
            }
          }
        }
      }
    }
    return (Stella_Object.cons(rules, Stella_Object.cons(indices, Stella.NIL)));
  }

  public static boolean descriptionImpliesDescriptionP(Description subdesc, Description superdesc) {
    return (LogicObject.collectionImpliesCollectionP(subdesc, superdesc));
  }

  public Surrogate descriptionSurrogate() {
    { Description self = this;

      { Relation nativerelation = self.nativeRelation();

        if (nativerelation != null) {
          { Surrogate testValue000 = Stella_Object.safePrimaryType(nativerelation);

            if (Surrogate.subtypeOfClassP(testValue000)) {
              { Stella_Class nativerelation000 = ((Stella_Class)(nativerelation));

                return (Logic.classLogicalType(nativerelation000));
              }
            }
            else if (Surrogate.subtypeOfP(testValue000, Logic.SGT_STELLA_SLOT)) {
              { Slot nativerelation000 = ((Slot)(nativerelation));

                return (nativerelation000.slotSlotref);
              }
            }
            else {
            }
          }
        }
        return (null);
      }
    }
  }

  public static Proposition assertDescriptionImpliesDescription(Description sub, Description renamed_Super, boolean dontupdateP) {
    { Proposition impliesprop = Logic.createImpliesProposition(sub, renamed_Super);

      impliesprop = Proposition.fastenDownOneProposition(impliesprop, Stella.FALSE);
      if (!(dontupdateP)) {
        Proposition.updatePropositionTruthValue(impliesprop, Logic.KWD_ASSERT_TRUE);
      }
      return (impliesprop);
    }
  }

  public static boolean descriptionExtensionL(Description desc1, Description desc2) {
    return (NamedDescription.getDescriptionExtension(((NamedDescription)(desc1)), Stella.TRUE).estimatedLength() < NamedDescription.getDescriptionExtension(((NamedDescription)(desc2)), Stella.TRUE).estimatedLength());
  }

  public int arity() {
    { Description self = this;

      return (self.ioVariables.length());
    }
  }

  public static boolean variableArityP(Description self) {
    return (Logic.testPropertyP(self, Logic.SGT_PL_KERNEL_KB_VARIABLE_ARITY));
  }

  public static Stella_Object accessDescriptionSlotValue(Description self, Symbol slotname, Stella_Object value, boolean setvalueP) {
    if (slotname == Logic.SYM_LOGIC_PROPOSITION) {
      if (setvalueP) {
        self.proposition = ((Proposition)(value));
      }
      else {
        value = self.proposition;
      }
    }
    else if (slotname == Logic.SYM_LOGIC_IO_VARIABLES) {
      if (setvalueP) {
        self.ioVariables = ((Vector)(value));
      }
      else {
        value = self.ioVariables;
      }
    }
    else if (slotname == Logic.SYM_LOGIC_INTERNAL_VARIABLES) {
      if (setvalueP) {
        self.internalVariables = ((Vector)(value));
      }
      else {
        value = self.internalVariables;
      }
    }
    else if (slotname == Logic.SYM_LOGIC_QUERY_PATTERNS) {
      if (setvalueP) {
        self.queryPatterns = ((KeyValueList)(value));
      }
      else {
        value = self.queryPatterns;
      }
    }
    else if (slotname == Logic.SYM_LOGIC_FORWARD_CHAINING_INDICES) {
      if (setvalueP) {
        KeyValueList.setDynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_FORWARD_CHAINING_INDICES, ((List)(value)), null);
      }
      else {
        value = self.forwardChainingIndices();
      }
    }
    else if (slotname == Logic.SYM_LOGIC_DEFERRED_CONTRAPOSITIVESp) {
      if (setvalueP) {
        KeyValueList.setDynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_DEFERRED_CONTRAPOSITIVESp, ((BooleanWrapper)(value)), null);
      }
      else {
        value = self.deferredContrapositivesP();
      }
    }
    else {
      if (slotname == Logic.SYM_LOGIC_RELATIVE_COMPLEMENT) {
        slotname = Logic.SYM_LOGIC_COMPLEMENT_DESCRIPTION;
      }
      else {
      }
      if (setvalueP) {
        KeyValueList.setDynamicSlotValue(self.dynamicSlots, slotname, value, null);
      }
      else {
        value = self.dynamicSlots.lookup(slotname);
      }
    }
    return (value);
  }

  public BooleanWrapper deferredContrapositivesP() {
    { Description self = this;

      { BooleanWrapper answer = ((BooleanWrapper)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_DEFERRED_CONTRAPOSITIVESp, null)));

        if (answer == null) {
          return ((Stella.FALSE ? Stella.TRUE_WRAPPER : Stella.FALSE_WRAPPER));
        }
        else {
          return (answer);
        }
      }
    }
  }

  public List forwardChainingIndices() {
    { Description self = this;

      { List answer = ((List)(KeyValueList.dynamicSlotValue(self.dynamicSlots, Logic.SYM_LOGIC_FORWARD_CHAINING_INDICES, null)));

        if (answer == null) {
          return (Stella.NIL_LIST);
        }
        else {
          return (answer);
        }
      }
    }
  }

  public Relation nativeRelation() {
    { Description self = this;

      return (null);
    }
  }

  public Surrogate primaryType() {
    { Description self = this;

      return (Logic.SGT_LOGIC_DESCRIPTION);
    }
  }

}

