;;; -*- Mode: Lisp; Package: STELLA; Syntax: COMMON-LISP; Base: 10 -*-

;;; propagate.lisp

#|--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1997-2003              |
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------|#

(CL:IN-PACKAGE "STELLA")

;;; Auxiliary variables:

(CL:DEFVAR SYM-PROPAGATE-LOGIC-DEFERRED-DEFAULT-PROPOSITIONS-INTERNAL NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-EVALUATION-STATE-TABLE NULL)
(CL:DEFVAR KWD-PROPAGATE-POSTED NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE NULL)
(CL:DEFVAR KWD-PROPAGATE-FORWARD NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-BACKWARD-ONLY? NULL)
(CL:DEFVAR SGT-PROPAGATE-LOGIC-NAMED-DESCRIPTION NULL)
(CL:DEFVAR SGT-PROPAGATE-LOGIC-DESCRIPTION NULL)
(CL:DEFVAR SGT-PROPAGATE-LOGIC-COLLECT-FORWARD-CHAINING-RULES-MEMO-TABLE-000 NULL)
(CL:DEFVAR KWD-PROPAGATE-FUNCTION NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE NULL)
(CL:DEFVAR KWD-PROPAGATE-ISA NULL)
(CL:DEFVAR KWD-PROPAGATE-PREDICATE NULL)
(CL:DEFVAR SGT-PROPAGATE-LOGIC-SKOLEM NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-HYPOTHESIZED-INSTANCE? NULL)
(CL:DEFVAR KWD-PROPAGATE-PROPAGATE NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-COMPLEMENT-DESCRIPTION NULL)
(CL:DEFVAR KWD-PROPAGATE-SINGLETONS? NULL)
(CL:DEFVAR SGT-PROPAGATE-LOGIC-PROPOSITION NULL)
(CL:DEFVAR KWD-PROPAGATE-KB-UPDATE NULL)
(CL:DEFVAR KWD-PROPAGATE-META-KB-UPDATE NULL)
(CL:DEFVAR KWD-PROPAGATE-META NULL)
(CL:DEFVAR SGT-PROPAGATE-STELLA-MODULE NULL)
(CL:DEFVAR SGT-PROPAGATE-LOGIC-LOGIC-OBJECT NULL)
(CL:DEFVAR SGT-PROPAGATE-STELLA-SYMBOL NULL)
(CL:DEFVAR KWD-PROPAGATE-EVALUATED NULL)
(CL:DEFVAR KWD-PROPAGATE-AND NULL)
(CL:DEFVAR KWD-PROPAGATE-OR NULL)
(CL:DEFVAR KWD-PROPAGATE-NOT NULL)
(CL:DEFVAR KWD-PROPAGATE-EQUIVALENT NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-ELABORATED-IN-WORLDS NULL)
(CL:DEFVAR KWD-PROPAGATE-ELABORATE NULL)
(CL:DEFVAR KWD-PROPAGATE-EXTENSIONAL-ASSERTION NULL)
(CL:DEFVAR SGT-PROPAGATE-PL-KERNEL-KB-INEQUALITY NULL)
(CL:DEFVAR SGT-PROPAGATE-STELLA-SURROGATE NULL)
(CL:DEFVAR KWD-PROPAGATE-EXISTS NULL)
(CL:DEFVAR KWD-PROPAGATE-FORALL NULL)
(CL:DEFVAR SGT-PROPAGATE-STELLA-LITERAL-WRAPPER NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-DESCRIPTION NULL)
(CL:DEFVAR SGT-PROPAGATE-STELLA-SLOT NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-ALL-FACTS-OF NULL)
(CL:DEFVAR KWD-PROPAGATE-COMMON-LISP NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-CALL-ALL-FACTS-OF NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-RETRACT-FACTS-OF NULL)
(CL:DEFVAR SGT-PROPAGATE-PL-KERNEL-KB-PHRASE NULL)
(CL:DEFVAR KWD-PROPAGATE-IMPLIES NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-ALL-SENTENCES-OF NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-ALL-FACTS-OF-N NULL)
(CL:DEFVAR SYM-PROPAGATE-LOGIC-STARTUP-PROPAGATE NULL)
(CL:DEFVAR SYM-PROPAGATE-STELLA-METHOD-STARTUP-CLASSNAME NULL)

;;; Forward declarations:

(CL:DECLAIM
 (CL:SPECIAL *STARTUP-TIME-PHASE* STANDARD-ERROR NIL-LIST
  DEFAULT-FALSE-TRUTH-VALUE *TRACED-KEYWORDS* *INVISIBLEASSERTION?* *MODULE*
  DEFAULT-TRUE-TRUTH-VALUE NULL-INTEGER STANDARD-OUTPUT EOL
  *COLLECTFORWARDPROPOSITIONS* *MEMOIZATION-ENABLED?* MEMOIZED-NULL-VALUE
  TRUE-WRAPPER FALSE-WRAPPER TRUE *CONTEXT* NIL FALSE))

;;; (DEFSPECIAL *FILLINGCONSTRAINTPROPAGATIONQUEUES?* ...)

(CL:DEFVAR *FILLINGCONSTRAINTPROPAGATIONQUEUES?* FALSE
  "True if we are inside of 'react-to-kb-update'.")

;;; (DEFSPECIAL *DEFERINGDEFAULTFORWARDINFERENCES?* ...)

(CL:DEFVAR *DEFERINGDEFAULTFORWARDINFERENCES?* FALSE
  "True if we are propagating strict inferences, and
posting derived default propositions to temporary queues.")

;;; (DEFUN (DEFERRED-DEFAULT-PROPOSITIONS (CONS OF PROPOSITION)) ...)

(CL:DEFUN DEFERRED-DEFAULT-PROPOSITIONS (SELF)
  (CL:LET*
   ((VALUE
     (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
      SYM-PROPAGATE-LOGIC-DEFERRED-DEFAULT-PROPOSITIONS-INTERNAL NULL)))
   (CL:IF (CL:EQ VALUE NULL) NIL VALUE)))

;;; (DEFUN DEFERRED-DEFAULT-PROPOSITIONS-SETTER ...)

(CL:DEFUN DEFERRED-DEFAULT-PROPOSITIONS-SETTER (SELF NEWVALUE)
  (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
   SYM-PROPAGATE-LOGIC-DEFERRED-DEFAULT-PROPOSITIONS-INTERNAL NEWVALUE NULL))

;;; (DEFUN (EVALUATION-STATE KEYWORD) ...)

(CL:DEFUN EVALUATION-STATE (PROPOSITION)
  (CL:LET*
   ((TABLE
     (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
      SYM-PROPAGATE-LOGIC-EVALUATION-STATE-TABLE NULL)))
   (CL:WHEN (CL:EQ TABLE NULL) (CL:RETURN-FROM EVALUATION-STATE NULL))
   (LOOKUP TABLE PROPOSITION)))

;;; (DEFUN EVALUATION-STATE-SETTER ...)

(CL:DEFUN EVALUATION-STATE-SETTER (PROPOSITION STATE)
  (CL:LET*
   ((TABLE
     (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
      SYM-PROPAGATE-LOGIC-EVALUATION-STATE-TABLE NULL)))
   (CL:WHEN (CL:EQ TABLE NULL) (CL:RETURN-FROM EVALUATION-STATE-SETTER))
   (INSERT-AT TABLE PROPOSITION STATE)))

;;; (DEFUN POST-FOR-EVALUATION ...)

(CL:DEFUN POST-FOR-EVALUATION (SELF)
  (CL:WHEN (CL:NOT *FILLINGCONSTRAINTPROPAGATIONQUEUES?*)
   (EVALUATE-PROPOSITION SELF) (CL:RETURN-FROM POST-FOR-EVALUATION))
  (CL:WHEN (CL:EQ (EVALUATION-STATE SELF) KWD-PROPAGATE-POSTED)
   (CL:RETURN-FROM POST-FOR-EVALUATION))
  (INSERT
   (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
    SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE NULL)
   SELF)
  (EVALUATION-STATE-SETTER SELF KWD-PROPAGATE-POSTED))

;;; (DEFSPECIAL *CURRENTLYEXECUTINGFORWARDCHAININGQUEUE* ...)

(CL:DEFVAR *CURRENTLYEXECUTINGFORWARDCHAININGQUEUE* NULL
  "Points to an active queue of propositions that will trigger
forward inference.  New queue entries are posted to a separate queue
attached to *context*.")

;;; (DEFUN (HELP-COLLECT-FORWARD-RULES (CONS OF KEY-VALUE-LIST)) ...)

(CL:DEFUN HELP-COLLECT-FORWARD-RULES (DESCRIPTION RULES INDICES TOUCHEDDEFAULT? BEENTHERE)
  (CL:WHEN (CL:EQ RULES NULL) (CL:SETQ RULES (NEW-KEY-VALUE-LIST)))
  (CL:WHEN (CL:EQ INDICES NULL) (CL:SETQ INDICES (NEW-KEY-VALUE-LIST)))
  (CL:WHEN (CL:EQ BEENTHERE NULL) (CL:SETQ BEENTHERE (NEW-LIST)))
  (INSERT BEENTHERE DESCRIPTION)
  (CL:LET*
   ((INDEX NULL)
    (ITER-000 (%THE-CONS-LIST (FORWARD-CHAINING-INDICES DESCRIPTION))))
   (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
    (CL:PROGN (CL:SETQ INDEX (%%VALUE ITER-000))
     (CL:SETQ ITER-000 (%%REST ITER-000)))
    (CL:WHEN
     (CL:AND (CL:NOT (DELETED? (%MASTER-RULE INDEX)))
      (TRUE? (%MASTER-RULE INDEX)))
     (INSERT-AT INDICES INDEX
      (CL:IF TOUCHEDDEFAULT? TRUE-WRAPPER FALSE-WRAPPER)))))
  (CL:LET*
   ((P NULL)
    (ITER-001
     (APPLICABLE-RULES-OF-DESCRIPTION DESCRIPTION KWD-PROPAGATE-FORWARD
      TRUE)))
   (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NIL)) DO
    (CL:PROGN (CL:SETQ P (%%VALUE ITER-001))
     (CL:SETQ ITER-001 (%%REST ITER-001)))
    (CL:WHEN
     (CL:NOT
      (%WRAPPER-VALUE
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS P)
        SYM-PROPAGATE-LOGIC-BACKWARD-ONLY? FALSE-WRAPPER)))
     (CL:LET*
      ((CONSEQUENT
        (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS P))) 1)))
      (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE CONSEQUENT)))
       (CL:COND
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-NAMED-DESCRIPTION)
         (CL:PROGN
          (CL:LET* ((SUPER (VALUE-OF CONSEQUENT)))
           (CL:WHEN (DEFAULT-TRUE? P) (CL:SETQ TOUCHEDDEFAULT? TRUE))
           (CL:WHEN (CL:NOT (CL:EQ (CLASS? DESCRIPTION) (CLASS? SUPER)))
            (INSERT-AT RULES P
             (CL:IF TOUCHEDDEFAULT? TRUE-WRAPPER FALSE-WRAPPER)))
           (CL:WHEN (CL:NOT (MEMBER? BEENTHERE SUPER))
            (HELP-COLLECT-FORWARD-RULES SUPER RULES INDICES TOUCHEDDEFAULT?
             BEENTHERE)))))
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-DESCRIPTION)
         (CL:PROGN
          (INSERT-AT RULES P
           (CL:IF TOUCHEDDEFAULT? TRUE-WRAPPER FALSE-WRAPPER))))
        (CL:T)))))))
  (CONS RULES (CONS INDICES NIL)))

;;; (DEFUN (COLLECT-FORWARD-CHAINING-RULES (KEY-VALUE-LIST OF PROPOSITION BOOLEAN-WRAPPER) (KEY-VALUE-LIST OF FORWARD-CHAINING-INDEX BOOLEAN-WRAPPER)) ...)

(CL:DEFUN COLLECT-FORWARD-CHAINING-RULES (DESCRIPTION)
  (CL:LET*
   ((MEMO-TABLE-000
     (%SURROGATE-VALUE
      SGT-PROPAGATE-LOGIC-COLLECT-FORWARD-CHAINING-RULES-MEMO-TABLE-000))
    (MEMOIZED-ENTRY-000 NULL) (MEMOIZED-VALUE-000 NULL))
   (CL:WHEN (CL:EQ MEMO-TABLE-000 NULL)
    (INITIALIZE-MEMOIZATION-TABLE
     SGT-PROPAGATE-LOGIC-COLLECT-FORWARD-CHAINING-RULES-MEMO-TABLE-000
     "(:MAX-VALUES 500 :TIMESTAMPS (:META-KB-UPDATE))")
    (CL:SETQ MEMO-TABLE-000
     (%SURROGATE-VALUE
      SGT-PROPAGATE-LOGIC-COLLECT-FORWARD-CHAINING-RULES-MEMO-TABLE-000)))
   (CL:WHEN *MEMOIZATION-ENABLED?*
    (CL:SETQ MEMOIZED-ENTRY-000
     (LOOKUP-MRU-MEMOIZED-VALUE MEMO-TABLE-000 DESCRIPTION *CONTEXT*
      MEMOIZED-NULL-VALUE NULL -1))
    (CL:SETQ MEMOIZED-VALUE-000 (%%VALUE MEMOIZED-ENTRY-000)))
   (CL:COND
    ((CL:NOT (CL:EQ MEMOIZED-VALUE-000 NULL))
     (CL:WHEN (CL:EQ MEMOIZED-VALUE-000 MEMOIZED-NULL-VALUE)
      (CL:SETQ MEMOIZED-VALUE-000 NULL)))
    (CL:T
     (CL:SETQ MEMOIZED-VALUE-000
      (HELP-COLLECT-FORWARD-RULES DESCRIPTION NULL NULL FALSE NULL))
     (CL:WHEN *MEMOIZATION-ENABLED?*
      (CL:SETF (%%VALUE MEMOIZED-ENTRY-000)
       (CL:IF (CL:EQ MEMOIZED-VALUE-000 NULL) MEMOIZED-NULL-VALUE
        MEMOIZED-VALUE-000)))))
   (CL:LET* ((RULES MEMOIZED-VALUE-000))
    (CL:VALUES (%%VALUE RULES) (%%VALUE (%%REST RULES))))))

;;; (DEFUN (HAS-FORWARD-CHAINING-RULES? BOOLEAN) ...)

(CL:DEFUN HAS-FORWARD-CHAINING-RULES? (DESCRIPTION PROPOSITION)
  (CL:WHEN (FALSE? PROPOSITION)
   (CL:SETQ DESCRIPTION (GET-INFERABLE-COMPLEMENT-DESCRIPTION DESCRIPTION))
   (CL:WHEN (CL:EQ DESCRIPTION NULL)
    (CL:RETURN-FROM HAS-FORWARD-CHAINING-RULES? FALSE)))
  (CL:LET* ((RULES NULL) (INDICES NULL))
   (CL:MULTIPLE-VALUE-SETQ (RULES INDICES)
    (COLLECT-FORWARD-CHAINING-RULES DESCRIPTION))
   (CL:OR (NON-EMPTY? RULES) (NON-EMPTY? INDICES))))

;;; (DEFUN POST-TO-FORWARD-CHAINING-QUEUE ...)

(CL:DEFUN POST-TO-FORWARD-CHAINING-QUEUE (SELF WORLD)
  (CL:WHEN
   (CL:AND (UNKNOWN? SELF)
    (CL:NOT (CL:EQ (%KIND SELF) KWD-PROPAGATE-FUNCTION)))
   (CL:RETURN-FROM POST-TO-FORWARD-CHAINING-QUEUE))
  (CL:LET*
   ((FORWARDCHAININGQUEUE
     (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS WORLD)
      SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE NULL)))
   (CL:LET* ((TEST-VALUE-000 (%KIND SELF)))
    (CL:COND
     ((CL:OR (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-ISA)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-PREDICATE)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-FUNCTION))
      (CL:LET* ((FOUND?-000 FALSE))
       (CL:LET*
        ((ARG NULL) (VECTOR-000 (%ARGUMENTS SELF)) (INDEX-000 0)
         (LENGTH-000 (LENGTH VECTOR-000)))
        (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
        (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
         (CL:PROGN
          (CL:SETQ ARG
           (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
            INDEX-000))
          (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
         (CL:WHEN
          (CL:AND (ISA? ARG SGT-PROPAGATE-LOGIC-SKOLEM)
           (CL:EQ (%DEFINING-PROPOSITION ARG) NULL)
           (CL:NOT
            (%WRAPPER-VALUE
             (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS ARG)
              SYM-PROPAGATE-LOGIC-HYPOTHESIZED-INSTANCE? FALSE-WRAPPER))))
          (CL:SETQ FOUND?-000 TRUE) (CL:RETURN))))
       (CL:WHEN FOUND?-000
        (CL:WHEN (CL:NOT (CL:EQ *COLLECTFORWARDPROPOSITIONS* NULL))
         (CL:RETURN-FROM POST-TO-FORWARD-CHAINING-QUEUE))))
      (CL:LET* ((DESCRIPTION (GET-DESCRIPTION (%OPERATOR SELF))))
       (CL:WHEN
        (CL:AND (CL:NOT (CL:EQ DESCRIPTION NULL))
         (HAS-FORWARD-CHAINING-RULES? DESCRIPTION SELF))
        (PUSH FORWARDCHAININGQUEUE SELF))))
     (CL:T)))))

;;; (DEFSPECIAL *COLLECTFORWARDPROPOSITIONS* ...)

(CL:DEFVAR *COLLECTFORWARDPROPOSITIONS* NULL
  "Collect goes-true propositions produced by forward
chaining.")

;;; (DEFUN APPLY-RULE-CONSEQUENT-TO-VECTOR ...)

(CL:DEFUN APPLY-RULE-CONSEQUENT-TO-VECTOR (CONSEQUENT ARGUMENTS RULE TRIGGERDESCRIPTION TRIGGERPROPOSITION TOUCHEDDEFAULT?)
  (CL:LET* ((*COLLECTFORWARDPROPOSITIONS* NIL))
   (CL:DECLARE (CL:SPECIAL *COLLECTFORWARDPROPOSITIONS*))
   (CL:LET*
    ((TOUCHEDDEFAULTKNOWLEDGE?
      (CL:OR (DEFAULT-TRUE? TRIGGERPROPOSITION) (DEFAULT-TRUE? RULE)
       TOUCHEDDEFAULT?)))
    (CL:LET* ((*QUERYITERATOR* NULL))
     (CL:DECLARE (CL:SPECIAL *QUERYITERATOR*))
     (INHERIT-DESCRIPTION ARGUMENTS CONSEQUENT TOUCHEDDEFAULTKNOWLEDGE?)))
   (CL:LET* ((P NULL) (ITER-000 *COLLECTFORWARDPROPOSITIONS*))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
     (CL:PROGN (CL:SETQ P (%%VALUE ITER-000))
      (CL:SETQ ITER-000 (%%REST ITER-000)))
     (CL:WHEN (TRACE-KEYWORD? KWD-PROPAGATE-PROPAGATE)
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) "MODUS PONENS:" EOL)
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) "   ANTECEDENT:  (")
      (CL:IF (NAMED-DESCRIPTION? TRIGGERDESCRIPTION)
       (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT)
        (DESCRIPTION-NAME TRIGGERDESCRIPTION))
       (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) "not-"
        (DESCRIPTION-NAME
         (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS TRIGGERDESCRIPTION)
          SYM-PROPAGATE-LOGIC-COMPLEMENT-DESCRIPTION NULL))))
      (CL:LET*
       ((ARG NULL) (VECTOR-000 ARGUMENTS) (INDEX-000 0)
        (LENGTH-000 (LENGTH VECTOR-000)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
       (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
        (CL:PROGN
         (CL:SETQ ARG
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
           INDEX-000))
         (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
        (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) " ")
        (PRINT-FORMULA ARG 15)))
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) ")" EOL)
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) "   RULE:  ")
      (PRINT-FORMULA RULE 10)
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) EOL)
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) "   CONSEQUENT: ")
      (PRINT-FORMULA P 15)
      (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) EOL))
     (RECORD-FORWARD-GOAL RULE ARGUMENTS P)
     (RECORD-FORWARD-JUSTIFICATION P RULE (CONS-LIST TRIGGERPROPOSITION))))))

;;; (DEFUN APPLY-FORWARD-RULES-TO-VECTOR ...)

(CL:DEFUN APPLY-FORWARD-RULES-TO-VECTOR (TRIGGERDESCRIPTION ARGUMENTS TRIGGERPROPOSITION)
  (CL:LET* ((EVALUATEDARGS ARGUMENTS))
   (CL:LET*
    ((I NULL-INTEGER) (ARG NULL) (VECTOR-000 ARGUMENTS) (INDEX-000 0)
     (LENGTH-000 (LENGTH VECTOR-000)) (ITER-000 0))
    (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000 ITER-000 I))
    (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
     (CL:PROGN
      (CL:SETQ ARG
       (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000)) INDEX-000))
      (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
     (CL:PROGN (CL:SETQ I ITER-000) (CL:SETQ ITER-000 (CL:1+ ITER-000)))
     (CL:COND ((DELETED? ARG) (CL:RETURN-FROM APPLY-FORWARD-RULES-TO-VECTOR))
      ((CL:NOT (EQL? ARG (VALUE-OF ARG)))
       (CL:WHEN (CL:EQ ARGUMENTS EVALUATEDARGS)
        (CL:SETQ EVALUATEDARGS (COPY ARGUMENTS)))
       (CL:LET
        ((SELF (%THE-ARRAY EVALUATEDARGS)) (VALUE (VALUE-OF ARG))
         (POSITION I))
        (CL:DECLARE (CL:TYPE CL:SIMPLE-VECTOR SELF)
         (CL:TYPE CL:FIXNUM POSITION))
        (CL:SETF (CL:AREF SELF POSITION) VALUE))))))
   (CL:SETQ ARGUMENTS EVALUATEDARGS))
  (CL:LET* ((RULES NULL) (INDICES NULL))
   (CL:MULTIPLE-VALUE-SETQ (RULES INDICES)
    (COLLECT-FORWARD-CHAINING-RULES TRIGGERDESCRIPTION))
   (CL:LET* ((TOUCHEDDEFAULT? NULL) (R NULL) (ITER-001 (%THE-KV-LIST RULES)))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NULL)) DO
     (CL:SETQ R (%KEY ITER-001))
     (CL:PROGN (CL:SETQ TOUCHEDDEFAULT? (%VALUE ITER-001))
      (CL:SETQ ITER-001 (%REST ITER-001)))
     (CL:WHEN (APPLICABLE-FORWARD-RULE? R ARGUMENTS)
      (APPLY-RULE-CONSEQUENT-TO-VECTOR
       (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS R))) 1)
       ARGUMENTS R TRIGGERDESCRIPTION TRIGGERPROPOSITION
       (COERCE-WRAPPED-BOOLEAN-TO-BOOLEAN TOUCHEDDEFAULT?)))))
   (CL:LET*
    ((TOUCHEDDEFAULT? NULL) (FWDINDEX NULL) (ITER-002 (%THE-KV-LIST INDICES)))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-002 NULL)) DO
     (CL:SETQ FWDINDEX (%KEY ITER-002))
     (CL:PROGN (CL:SETQ TOUCHEDDEFAULT? (%VALUE ITER-002))
      (CL:SETQ ITER-002 (%REST ITER-002)))
     (CL:TAGBODY
      (CL:LET* ((INPUTARGS NIL))
       (CL:WHEN
        (CL:NOT
         (CL:= (LENGTH ARGUMENTS)
          (LENGTH (%ARGUMENTS (%FORWARD-GOAL FWDINDEX)))))
        (CL:GO :CONTINUE))
       (CL:LET*
        ((GOALARG NULL) (ARG NULL) (VECTOR-001 ARGUMENTS) (INDEX-001 0)
         (LENGTH-001 (LENGTH VECTOR-001))
         (VECTOR-002 (%ARGUMENTS (%FORWARD-GOAL FWDINDEX))) (INDEX-002 0)
         (LENGTH-002 (LENGTH VECTOR-002)))
        (CL:DECLARE
         (CL:TYPE CL:FIXNUM INDEX-001 LENGTH-001 INDEX-002 LENGTH-002))
        (CL:LOOP WHILE
         (CL:AND (CL:< INDEX-001 LENGTH-001) (CL:< INDEX-002 LENGTH-002)) DO
         (CL:PROGN
          (CL:SETQ ARG
           (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-001))
            INDEX-001))
          (CL:SETQ INDEX-001 (CL:1+ INDEX-001)))
         (CL:PROGN
          (CL:SETQ GOALARG
           (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-002))
            INDEX-002))
          (CL:SETQ INDEX-002 (CL:1+ INDEX-002)))
         (CL:COND
          ((VARIABLE? GOALARG) (CL:SETQ INPUTARGS (CONS ARG INPUTARGS)))
          ((CL:NOT (EQL? ARG (VALUE-OF GOALARG))) (CL:SETQ INPUTARGS NULL)
           (CL:RETURN)))))
       (CL:WHEN (CL:EQ INPUTARGS NULL) (CL:GO :CONTINUE))
       (CL:LET*
        ((ARG NULL) (IT (ALLOCATE-ITERATOR (%INPUT-BINDINGS FWDINDEX)))
         (ITER-003 (REVERSE INPUTARGS)))
        (CL:LOOP WHILE (CL:AND (NEXT? IT) (CL:NOT (CL:EQ ITER-003 NIL))) DO
         (CL:PROGN (CL:SETQ ARG (%%VALUE ITER-003))
          (CL:SETQ ITER-003 (%%REST ITER-003)))
         (CL:PROGN) (VALUE-SETTER IT ARG))))
      (CL:LET* ((OUTPUTBINDINGS NULL) (TRUTHVALUES NULL))
       (CL:MULTIPLE-VALUE-SETQ (OUTPUTBINDINGS TRUTHVALUES)
        (APPLY-CACHED-RETRIEVE (%IO-VARIABLES FWDINDEX) (%QUERY-BODY FWDINDEX)
         (%INPUT-BINDINGS FWDINDEX)
         (CONS-LIST KWD-PROPAGATE-SINGLETONS? TRUE-WRAPPER)
         (%CACHE-ID FWDINDEX)))
       (CL:LET*
        ((TV NULL) (OB NULL) (ITER-004 OUTPUTBINDINGS) (ITER-005 TRUTHVALUES))
        (CL:LOOP WHILE
         (CL:AND (CL:NOT (CL:EQ ITER-004 NIL)) (CL:NOT (CL:EQ ITER-005 NIL)))
         DO
         (CL:PROGN (CL:SETQ OB (%%VALUE ITER-004))
          (CL:SETQ ITER-004 (%%REST ITER-004)))
         (CL:PROGN (CL:SETQ TV (%%VALUE ITER-005))
          (CL:SETQ ITER-005 (%%REST ITER-005)))
         (APPLY-RULE-CONSEQUENT-TO-VECTOR (%CONSEQUENT FWDINDEX)
          (COPY-LIST-TO-ARGUMENTS-VECTOR (LISTIFY OB)) (%MASTER-RULE FWDINDEX)
          TRIGGERDESCRIPTION TRIGGERPROPOSITION
          (CL:OR (COERCE-WRAPPED-BOOLEAN-TO-BOOLEAN TOUCHEDDEFAULT?)
           (CL:EQ TV DEFAULT-TRUE-TRUTH-VALUE))))))
      :CONTINUE)))))

;;; (DEFUN (APPLICABLE-FORWARD-RULE? BOOLEAN) ...)

(CL:DEFUN APPLICABLE-FORWARD-RULE? (RULE ARGUMENTS)
  (CL:LET*
   ((ANTECEDENTDESCRIPTION
     (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS RULE))) 0)))
   (CL:AND (TRUE? RULE)
    (CL:= (LENGTH ARGUMENTS) (LENGTH (%IO-VARIABLES ANTECEDENTDESCRIPTION))))))

;;; (DEFMETHOD REACT-TO-INFERENCE-UPDATE ...)

(CL:DEFMETHOD REACT-TO-INFERENCE-UPDATE ((SELF SKOLEM))
  (CL:LET* ((DERIVEDVALUE (VALUE-OF SELF)) (ASSERTEDVALUE NULL))
   (CL:LET* ((*MODULE* (%BASE-MODULE *CONTEXT*)) (*CONTEXT* *MODULE*))
    (CL:DECLARE (CL:SPECIAL *MODULE* *CONTEXT*))
    (CL:SETQ ASSERTEDVALUE (VALUE-OF SELF)))
   (CL:WHEN
    (CL:AND (CL:NOT (CL:EQ ASSERTEDVALUE NULL))
     (CL:NOT (EQL? ASSERTEDVALUE DERIVEDVALUE)))
    (EQUATE-VALUES (VALUE-OF ASSERTEDVALUE) DERIVEDVALUE))
   (CL:LET*
    ((PROP NULL)
     (ITER-000
      (ALLOCATE-ITERATOR (UNFILTERED-DEPENDENT-PROPOSITIONS SELF NULL))))
    (CL:LOOP WHILE (NEXT? ITER-000) DO (CL:SETQ PROP (%VALUE ITER-000))
     (POST-FOR-EVALUATION PROP)))
   (CL:WHEN *FILLINGCONSTRAINTPROPAGATIONQUEUES?*
    (CL:LET* ((DEFININGPROPOSITION (%DEFINING-PROPOSITION SELF)))
     (CL:WHEN (CL:NOT (CL:EQ DEFININGPROPOSITION NULL))
      (POST-TO-FORWARD-CHAINING-QUEUE DEFININGPROPOSITION *CONTEXT*))))))

;;; (DEFMETHOD REACT-TO-INFERENCE-UPDATE ...)

(CL:DEFMETHOD REACT-TO-INFERENCE-UPDATE ((SELF PROPOSITION))
  (CL:LET*
   ((PARENTTRUTHVALUE
     (ACCESS-IN-CONTEXT (%TRUTH-VALUE SELF) (%HOME-CONTEXT SELF) FALSE)))
   (CL:WHEN
    (CL:AND
     (CL:NOT
      (CL:EQ PARENTTRUTHVALUE
       (ACCESS-IN-CONTEXT (%TRUTH-VALUE SELF) (%HOME-CONTEXT SELF) FALSE)))
     (CL:NOT (CL:EQ PARENTTRUTHVALUE NULL)))
    (SIGNAL-TRUTH-VALUE-CLASH SELF))
   (POST-FOR-EVALUATION SELF)
   (CL:LET*
    ((PROP NULL)
     (ITER-000 (ALLOCATE-ITERATOR (%DEPENDENT-PROPOSITIONS SELF))))
    (CL:LOOP WHILE (NEXT? ITER-000) DO (CL:SETQ PROP (%VALUE ITER-000))
     (POST-FOR-EVALUATION PROP)))
   (CL:WHEN *FILLINGCONSTRAINTPROPAGATIONQUEUES?*
    (POST-TO-FORWARD-CHAINING-QUEUE SELF *CONTEXT*))))

;;; (DEFUN REACT-TO-KB-UPDATE ...)

(CL:DEFUN REACT-TO-KB-UPDATE (CONTEXT OBJECT)
  (CL:LET* ((SKOLEM? FALSE) (PROPOSITION? FALSE) (METAPROPOSITION? FALSE))
   (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE OBJECT)))
    (CL:COND
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-SKOLEM)
      (CL:PROGN (CL:SETQ SKOLEM? TRUE)))
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-PROPOSITION)
      (CL:PROGN (CL:SETQ PROPOSITION? TRUE)))
     (CL:T
      (CL:LET* ((STREAM-000 (NEW-OUTPUT-STRING-STREAM)))
       (%%PRINT-STREAM (%NATIVE-STREAM STREAM-000) "`" TEST-VALUE-000
        "' is not a valid case option")
       (CL:ERROR (NEW-STELLA-EXCEPTION (THE-STRING-READER STREAM-000)))))))
   (CL:WHEN (CL:AND PROPOSITION? (META-PROPOSITION? OBJECT))
    (CL:SETQ METAPROPOSITION? TRUE))
   (BUMP-MEMOIZATION-TIMESTAMP KWD-PROPAGATE-KB-UPDATE)
   (CL:WHEN METAPROPOSITION?
    (BUMP-MEMOIZATION-TIMESTAMP KWD-PROPAGATE-META-KB-UPDATE))
   (CL:COND (*INVISIBLEASSERTION?* (CL:RETURN-FROM REACT-TO-KB-UPDATE))
    ((DESCRIPTION-MODE?)
     (CL:LET* ((*FILLINGCONSTRAINTPROPAGATIONQUEUES?* FALSE))
      (CL:DECLARE (CL:SPECIAL *FILLINGCONSTRAINTPROPAGATIONQUEUES?*))
      (CL:COND (PROPOSITION? (REACT-TO-INFERENCE-UPDATE OBJECT))
       (SKOLEM? (REACT-TO-INFERENCE-UPDATE OBJECT))))
     (CL:RETURN-FROM REACT-TO-KB-UPDATE))
    (*FILLINGCONSTRAINTPROPAGATIONQUEUES?*
     (CL:COND (PROPOSITION? (REACT-TO-INFERENCE-UPDATE OBJECT))
      (SKOLEM? (REACT-TO-INFERENCE-UPDATE OBJECT)))
     (CL:RETURN-FROM REACT-TO-KB-UPDATE)))
   (CL:LET*
    ((WORLD
      (CL:IF METAPROPOSITION?
       (LOOKUP-INFERENCE-CACHE CONTEXT KWD-PROPAGATE-META)
       (LOOKUP-CONSTRAINT-PROPAGATION-WORLD CONTEXT))))
    (CL:WHEN (CL:NOT (CL:EQ WORLD NULL))
     (CL:LET* ((*CONTEXT* WORLD) (*MODULE* (%BASE-MODULE *CONTEXT*)))
      (CL:DECLARE (CL:SPECIAL *CONTEXT* *MODULE*))
      (INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES WORLD)
      (CL:LET* ((*FILLINGCONSTRAINTPROPAGATIONQUEUES?* TRUE))
       (CL:DECLARE (CL:SPECIAL *FILLINGCONSTRAINTPROPAGATIONQUEUES?*))
       (CL:COND (PROPOSITION? (REACT-TO-INFERENCE-UPDATE OBJECT))
        (SKOLEM? (REACT-TO-INFERENCE-UPDATE OBJECT)))
       (EXECUTE-CONSTRAINT-PROPAGATION-QUEUES)))))
   (CL:WHEN (ISA? CONTEXT SGT-PROPAGATE-STELLA-MODULE)
    (CL:LET*
     ((CHILD NULL) (ITER-000 (%THE-CONS-LIST (%CHILD-CONTEXTS CONTEXT))))
     (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
      (CL:PROGN (CL:SETQ CHILD (%%VALUE ITER-000))
       (CL:SETQ ITER-000 (%%REST ITER-000)))
      (CL:WHEN (ISA? CHILD SGT-PROPAGATE-STELLA-MODULE)
       (REACT-TO-KB-UPDATE CHILD OBJECT)))))))

;;; (DEFUN INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES ...)

(CL:DEFUN INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES (WORLD)
  (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS WORLD)
   SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE (NEW-LIST) NULL)
  (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS WORLD)
   SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE (NEW-LIST) NULL))

;;; (DEFUN EXECUTE-CONSTRAINT-PROPAGATION-QUEUES ...)

(CL:DEFUN EXECUTE-CONSTRAINT-PROPAGATION-QUEUES ()
  (CL:LET* ((*INHIBITOBJECTFINALIZATION?* TRUE))
   (CL:DECLARE (CL:SPECIAL *INHIBITOBJECTFINALIZATION?*))
   (CL:LOOP
    (CL:LET*
     ((QUEUE
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
        SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE NULL)))
     (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
      SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE (NEW-LIST) NULL)
     (CL:LET* ((P NULL) (ITER-000 (%THE-CONS-LIST QUEUE)))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
       (CL:PROGN (CL:SETQ P (%%VALUE ITER-000))
        (CL:SETQ ITER-000 (%%REST ITER-000)))
       (EVALUATE-PROPOSITION P))))
    (CL:LET*
     ((QUEUE
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
        SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE NULL)))
     (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
      SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE (NEW-LIST) NULL)
     (CL:LET* ((PROPOSITION NULL) (ITER-001 (%THE-CONS-LIST QUEUE)))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NIL)) DO
       (CL:PROGN (CL:SETQ PROPOSITION (%%VALUE ITER-001))
        (CL:SETQ ITER-001 (%%REST ITER-001)))
       (CL:PROGN)
       (CL:LET* ((DESCRIPTION (GET-DESCRIPTION (%OPERATOR PROPOSITION))))
        (CL:WHEN (FALSE? PROPOSITION)
         (CL:SETQ DESCRIPTION
          (GET-INFERABLE-COMPLEMENT-DESCRIPTION DESCRIPTION)))
        (CL:WHEN (CL:NOT (CL:EQ DESCRIPTION NULL))
         (APPLY-FORWARD-RULES-TO-VECTOR DESCRIPTION (%ARGUMENTS PROPOSITION)
          PROPOSITION))))))
    (CL:WHEN
     (CL:AND
      (EMPTY?
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
        SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE NULL))
      (EMPTY?
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS *CONTEXT*)
        SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE NULL)))
     (CL:RETURN-FROM EXECUTE-CONSTRAINT-PROPAGATION-QUEUES)))))

;;; (DEFUN EVALUATE-NEW-PROPOSITION ...)

(CL:DEFUN EVALUATE-NEW-PROPOSITION (SELF)
  (CL:WHEN (DESCRIPTION-MODE?) (CL:RETURN-FROM EVALUATE-NEW-PROPOSITION))
  (CL:LET* ((WORLD (LOOKUP-CONSTRAINT-PROPAGATION-WORLD *CONTEXT*)))
   (CL:WHEN (CL:NOT (CL:EQ WORLD NULL))
    (CL:LET* ((*CONTEXT* WORLD)) (CL:DECLARE (CL:SPECIAL *CONTEXT*))
     (EVALUATE-PROPOSITION SELF)
     (CL:WHEN (CL:NOT (UNKNOWN? SELF))
      (CL:WHEN
       (CL:AND (CL:NOT (CL:EQ *TRACED-KEYWORDS* NULL))
        (MEMB? *TRACED-KEYWORDS* KWD-PROPAGATE-PROPAGATE))
       (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT)
        "evaluate-new-proposition:  " SELF EOL))
      (CL:LET*
       ((PROP NULL)
        (ITER-000 (ALLOCATE-ITERATOR (%DEPENDENT-PROPOSITIONS SELF))))
       (CL:LOOP WHILE (NEXT? ITER-000) DO (CL:SETQ PROP (%VALUE ITER-000))
        (EVALUATE-PROPOSITION PROP))))))))

;;; (DEFUN EVALUATE-AND-PROPOSITION ...)

(CL:DEFUN EVALUATE-AND-PROPOSITION (SELF)
  (CL:LET* ((ARGUMENTS (%ARGUMENTS SELF)))
   (CL:COND
    ((TRUE? SELF)
     (CL:LET*
      ((ARG NULL) (VECTOR-000 ARGUMENTS) (INDEX-000 0)
       (LENGTH-000 (LENGTH VECTOR-000)))
      (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
      (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
       (CL:PROGN
        (CL:SETQ ARG
         (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
          INDEX-000))
        (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
       (ASSIGN-TRUTH-VALUE ARG
        (ACCESS-IN-CONTEXT (%TRUTH-VALUE SELF) (%HOME-CONTEXT SELF) FALSE)))))
    ((FALSE? SELF)
     (CL:LET* ((ONLYUNBOUNDARG NULL) (TOUCHEDDEFAULT? (DEFAULT-FALSE? SELF)))
      (CL:LET*
       ((ARG NULL) (VECTOR-001 ARGUMENTS) (INDEX-001 0)
        (LENGTH-001 (LENGTH VECTOR-001)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-001 LENGTH-001))
       (CL:LOOP WHILE (CL:< INDEX-001 LENGTH-001) DO
        (CL:PROGN
         (CL:SETQ ARG
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-001))
           INDEX-001))
         (CL:SETQ INDEX-001 (CL:1+ INDEX-001)))
        (CL:COND ((FALSE? ARG) (CL:RETURN-FROM EVALUATE-AND-PROPOSITION))
         ((UNKNOWN? ARG)
          (CL:WHEN (CL:NOT (CL:EQ ONLYUNBOUNDARG NULL))
           (CL:RETURN-FROM EVALUATE-AND-PROPOSITION))
          (CL:SETQ ONLYUNBOUNDARG ARG))
         ((DEFAULT-TRUE? ARG) (CL:SETQ TOUCHEDDEFAULT? TRUE)))))
      (CL:IF (CL:NOT (CL:EQ ONLYUNBOUNDARG NULL))
       (ASSIGN-TRUTH-VALUE ONLYUNBOUNDARG
        (DEFAULTIFY-TRUTH-VALUE FALSE TOUCHEDDEFAULT?))
       (ASSIGN-TRUTH-VALUE SELF
        (DEFAULTIFY-TRUTH-VALUE TRUE TOUCHEDDEFAULT?)))))
    (CL:T
     (CL:LET* ((UNBOUNDARG? FALSE) (TOUCHEDDEFAULT? FALSE))
      (CL:LET*
       ((ARG NULL) (VECTOR-002 ARGUMENTS) (INDEX-002 0)
        (LENGTH-002 (LENGTH VECTOR-002)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-002 LENGTH-002))
       (CL:LOOP WHILE (CL:< INDEX-002 LENGTH-002) DO
        (CL:PROGN
         (CL:SETQ ARG
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-002))
           INDEX-002))
         (CL:SETQ INDEX-002 (CL:1+ INDEX-002)))
        (CL:COND
         ((FALSE? ARG)
          (CL:IF (DEFAULT-FALSE? ARG)
           (ASSIGN-TRUTH-VALUE SELF DEFAULT-FALSE-TRUTH-VALUE)
           (CL:PROGN (ASSIGN-TRUTH-VALUE SELF FALSE-WRAPPER)
            (CL:RETURN-FROM EVALUATE-AND-PROPOSITION))))
         ((UNKNOWN? ARG) (CL:SETQ UNBOUNDARG? TRUE))
         ((DEFAULT-TRUE? ARG) (CL:SETQ TOUCHEDDEFAULT? TRUE)))))
      (CL:WHEN (CL:AND (CL:NOT UNBOUNDARG?) (CL:NOT (DEFAULT-FALSE? SELF)))
       (ASSIGN-TRUTH-VALUE SELF
        (DEFAULTIFY-TRUTH-VALUE TRUE TOUCHEDDEFAULT?))))))))

;;; (DEFUN EVALUATE-OR-PROPOSITION ...)

(CL:DEFUN EVALUATE-OR-PROPOSITION (SELF)
  (CL:LET* ((ARGUMENTS (%ARGUMENTS SELF)))
   (CL:COND
    ((FALSE? SELF)
     (CL:LET*
      ((ARG NULL) (VECTOR-000 ARGUMENTS) (INDEX-000 0)
       (LENGTH-000 (LENGTH VECTOR-000)))
      (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
      (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
       (CL:PROGN
        (CL:SETQ ARG
         (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
          INDEX-000))
        (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
       (ASSIGN-TRUTH-VALUE ARG
        (ACCESS-IN-CONTEXT (%TRUTH-VALUE SELF) (%HOME-CONTEXT SELF) FALSE)))))
    ((TRUE? SELF)
     (CL:LET* ((ONLYUNBOUNDARG NULL) (TOUCHEDDEFAULT? (DEFAULT-TRUE? SELF)))
      (CL:LET*
       ((ARG NULL) (VECTOR-001 ARGUMENTS) (INDEX-001 0)
        (LENGTH-001 (LENGTH VECTOR-001)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-001 LENGTH-001))
       (CL:LOOP WHILE (CL:< INDEX-001 LENGTH-001) DO
        (CL:PROGN
         (CL:SETQ ARG
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-001))
           INDEX-001))
         (CL:SETQ INDEX-001 (CL:1+ INDEX-001)))
        (CL:COND ((TRUE? ARG) (CL:RETURN-FROM EVALUATE-OR-PROPOSITION))
         ((UNKNOWN? ARG)
          (CL:WHEN (CL:NOT (CL:EQ ONLYUNBOUNDARG NULL))
           (CL:RETURN-FROM EVALUATE-OR-PROPOSITION))
          (CL:SETQ ONLYUNBOUNDARG ARG))
         ((DEFAULT-FALSE? ARG) (CL:SETQ TOUCHEDDEFAULT? TRUE)))))
      (CL:IF (CL:NOT (CL:EQ ONLYUNBOUNDARG NULL))
       (ASSIGN-TRUTH-VALUE ONLYUNBOUNDARG
        (DEFAULTIFY-TRUTH-VALUE TRUE TOUCHEDDEFAULT?))
       (ASSIGN-TRUTH-VALUE SELF
        (DEFAULTIFY-TRUTH-VALUE FALSE TOUCHEDDEFAULT?)))))
    (CL:T
     (CL:LET* ((UNBOUNDARG? FALSE) (TOUCHEDDEFAULT? FALSE))
      (CL:LET*
       ((ARG NULL) (VECTOR-002 ARGUMENTS) (INDEX-002 0)
        (LENGTH-002 (LENGTH VECTOR-002)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-002 LENGTH-002))
       (CL:LOOP WHILE (CL:< INDEX-002 LENGTH-002) DO
        (CL:PROGN
         (CL:SETQ ARG
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-002))
           INDEX-002))
         (CL:SETQ INDEX-002 (CL:1+ INDEX-002)))
        (CL:COND
         ((TRUE? ARG)
          (CL:IF (DEFAULT-TRUE? ARG)
           (ASSIGN-TRUTH-VALUE SELF DEFAULT-TRUE-TRUTH-VALUE)
           (CL:PROGN (ASSIGN-TRUTH-VALUE SELF TRUE-WRAPPER)
            (CL:RETURN-FROM EVALUATE-OR-PROPOSITION))))
         ((UNKNOWN? ARG) (CL:SETQ UNBOUNDARG? TRUE))
         ((DEFAULT-FALSE? ARG) (CL:SETQ TOUCHEDDEFAULT? TRUE)))))
      (CL:WHEN (CL:AND (CL:NOT UNBOUNDARG?) (CL:NOT (DEFAULT-TRUE? SELF)))
       (ASSIGN-TRUTH-VALUE SELF
        (DEFAULTIFY-TRUTH-VALUE FALSE TOUCHEDDEFAULT?))))))))

;;; (DEFUN EVALUATE-NOT-PROPOSITION ...)

(CL:DEFUN EVALUATE-NOT-PROPOSITION (SELF)
  (CL:LET*
   ((ARGUMENT
     (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS SELF))) 0)))
   (CL:WHEN (CL:OR (TRUE? ARGUMENT) (FALSE? ARGUMENT))
    (ASSIGN-TRUTH-VALUE SELF
     (INVERT-TRUTH-VALUE
      (ACCESS-IN-CONTEXT (%TRUTH-VALUE ARGUMENT) (%HOME-CONTEXT ARGUMENT)
       FALSE))))
   (CL:WHEN (CL:OR (TRUE? SELF) (FALSE? SELF))
    (ASSIGN-TRUTH-VALUE ARGUMENT
     (INVERT-TRUTH-VALUE
      (ACCESS-IN-CONTEXT (%TRUTH-VALUE SELF) (%HOME-CONTEXT SELF) FALSE))))))

;;; (DEFUN EVALUATE-EQUIVALENT-PROPOSITION ...)

(CL:DEFUN EVALUATE-EQUIVALENT-PROPOSITION (SELF)
  (CL:LET*
   ((ARGUMENTS (%ARGUMENTS SELF))
    (FIRSTARG
     (INNERMOST-OF
      (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY ARGUMENTS)) 0)))
    (SECONDARG
     (INNERMOST-OF
      (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY ARGUMENTS)) 1))))
   (CL:COND
    ((TRUE? SELF)
     (CL:WHEN (DEFAULT-TRUE? SELF)
      (CL:WARN
       "INTERNAL ERROR: DON'T KNOW YET HOW TO EQUATE THINGS BY DEFAULT."))
     (EQUATE-VALUES FIRSTARG SECONDARG))
    ((EQL? FIRSTARG SECONDARG) (ASSIGN-TRUTH-VALUE SELF TRUE-WRAPPER))
    ((CL:AND (CL:NOT (SKOLEM? FIRSTARG)) (CL:NOT (SKOLEM? SECONDARG)))
     (ASSIGN-TRUTH-VALUE SELF FALSE-WRAPPER)))))

;;; (DEFUN EQUATE-EQUIVALENT-FUNCTION-PROPOSITIONS ...)

(CL:DEFUN EQUATE-EQUIVALENT-FUNCTION-PROPOSITIONS (SELF)
  (CL:LET*
   ((RELATIONREF (%OPERATOR SELF)) (NOFARGS (LENGTH (%ARGUMENTS SELF)))
    (CANDIDATEPROPOSITIONS (NEW-LIST)))
   (CL:DECLARE (CL:TYPE CL:FIXNUM NOFARGS))
   (CL:LET*
    ((I NULL-INTEGER) (ARG NULL) (VECTOR-000 (%ARGUMENTS SELF)) (INDEX-000 0)
     (LENGTH-000 (LENGTH VECTOR-000)) (ITER-000 1))
    (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000 ITER-000 I))
    (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
     (CL:PROGN
      (CL:SETQ ARG
       (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000)) INDEX-000))
      (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
     (CL:PROGN (CL:SETQ I ITER-000) (CL:SETQ ITER-000 (CL:1+ ITER-000)))
     (CL:WHEN (CL:NOT (CL:= I NOFARGS))
      (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE ARG)))
       (CL:COND
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-SKOLEM)
         (CL:PROGN
          (CL:WHEN
           (CL:AND
            (CL:NOT
             (CL:EQ
              (ACCESS-IN-CONTEXT (%VARIABLE-VALUE ARG) (%HOME-CONTEXT ARG)
               FALSE)
              NULL))
            (ISA?
             (ACCESS-IN-CONTEXT (%VARIABLE-VALUE ARG) (%HOME-CONTEXT ARG)
              FALSE)
             SGT-PROPAGATE-LOGIC-LOGIC-OBJECT))
           (CL:LET*
            ((PROP NULL)
             (ITER-001
              (ALLOCATE-ITERATOR
               (UNFILTERED-DEPENDENT-PROPOSITIONS
                (ACCESS-IN-CONTEXT (%VARIABLE-VALUE ARG) (%HOME-CONTEXT ARG)
                 FALSE)
                RELATIONREF))))
            (CL:LOOP WHILE (NEXT? ITER-001) DO
             (CL:SETQ PROP (%VALUE ITER-001))
             (CL:WHEN
              (CL:AND (CL:EQ (%OPERATOR PROP) RELATIONREF)
               (CL:NOT (CL:EQ PROP SELF)))
              (INSERT-NEW CANDIDATEPROPOSITIONS PROP)))))
          (CL:LET* ((INVARG NULL) (ITER-002 (VARIABLE-VALUE-INVERSE ARG)))
           (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-002 NIL)) DO
            (CL:PROGN (CL:SETQ INVARG (%%VALUE ITER-002))
             (CL:SETQ ITER-002 (%%REST ITER-002)))
            (CL:LET*
             ((PROP NULL)
              (ITER-003
               (ALLOCATE-ITERATOR
                (UNFILTERED-DEPENDENT-PROPOSITIONS INVARG RELATIONREF))))
             (CL:LOOP WHILE (NEXT? ITER-003) DO
              (CL:SETQ PROP (%VALUE ITER-003))
              (CL:WHEN
               (CL:AND (CL:EQ (%OPERATOR PROP) RELATIONREF)
                (CL:NOT (CL:EQ PROP SELF)))
               (INSERT-NEW CANDIDATEPROPOSITIONS PROP))))))))
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-LOGIC-OBJECT)
         (CL:PROGN
          (CL:LET* ((INVARG NULL) (ITER-004 (VARIABLE-VALUE-INVERSE ARG)))
           (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-004 NIL)) DO
            (CL:PROGN (CL:SETQ INVARG (%%VALUE ITER-004))
             (CL:SETQ ITER-004 (%%REST ITER-004)))
            (CL:LET*
             ((PROP NULL)
              (ITER-005
               (ALLOCATE-ITERATOR
                (UNFILTERED-DEPENDENT-PROPOSITIONS INVARG RELATIONREF))))
             (CL:LOOP WHILE (NEXT? ITER-005) DO
              (CL:SETQ PROP (%VALUE ITER-005))
              (CL:WHEN
               (CL:AND (CL:EQ (%OPERATOR PROP) RELATIONREF)
                (CL:NOT (CL:EQ PROP SELF)))
               (INSERT-NEW CANDIDATEPROPOSITIONS PROP))))))))
        (CL:T))))))
   (CL:LET* ((PROP NULL) (ITER-006 (%THE-CONS-LIST CANDIDATEPROPOSITIONS)))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-006 NIL)) DO
     (CL:PROGN (CL:SETQ PROP (%%VALUE ITER-006))
      (CL:SETQ ITER-006 (%%REST ITER-006)))
     (CL:LET* ((ALWAYS?-000 TRUE))
      (CL:LET*
       ((I NULL-INTEGER) (ARG2 NULL) (ARG1 NULL)
        (VECTOR-001 (%ARGUMENTS SELF)) (INDEX-001 0)
        (LENGTH-001 (LENGTH VECTOR-001)) (VECTOR-002 (%ARGUMENTS PROP))
        (INDEX-002 0) (LENGTH-002 (LENGTH VECTOR-002)) (ITER-007 1))
       (CL:DECLARE
        (CL:TYPE CL:FIXNUM INDEX-001 LENGTH-001 INDEX-002 LENGTH-002 ITER-007
         I))
       (CL:LOOP WHILE
        (CL:AND (CL:< INDEX-001 LENGTH-001) (CL:< INDEX-002 LENGTH-002)) DO
        (CL:PROGN
         (CL:SETQ ARG1
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-001))
           INDEX-001))
         (CL:SETQ INDEX-001 (CL:1+ INDEX-001)))
        (CL:PROGN
         (CL:SETQ ARG2
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-002))
           INDEX-002))
         (CL:SETQ INDEX-002 (CL:1+ INDEX-002)))
        (CL:PROGN (CL:SETQ I ITER-007) (CL:SETQ ITER-007 (CL:1+ ITER-007)))
        (CL:WHEN (CL:NOT (CL:= I NOFARGS))
         (CL:WHEN (CL:NOT (EQL? (INNERMOST-OF ARG1) (INNERMOST-OF ARG2)))
          (CL:SETQ ALWAYS?-000 FALSE) (CL:RETURN)))))
      (CL:WHEN ALWAYS?-000
       (EQUATE-VALUES (INNERMOST-OF (LAST (%ARGUMENTS SELF)))
        (INNERMOST-OF (LAST (%ARGUMENTS PROP))))))))
   (FREE CANDIDATEPROPOSITIONS)))

;;; (DEFUN EVALUATE-FUNCTION-PROPOSITION ...)

(CL:DEFUN EVALUATE-FUNCTION-PROPOSITION (SELF)
  (CL:WHEN (ISA? (%OPERATOR SELF) SGT-PROPAGATE-STELLA-SYMBOL)
   (CL:RETURN-FROM EVALUATE-FUNCTION-PROPOSITION))
  (EQUATE-EQUIVALENT-FUNCTION-PROPOSITIONS SELF)
  (CL:LET*
   ((DESCRIPTION (GET-DESCRIPTION (%OPERATOR SELF)))
    (CODE (LOOKUP-CONSTRAINT DESCRIPTION)) (BOUNDARGUMENTS NIL)
    (COMPUTEDVALUE NULL) (STOREDVALUE NULL) (MISSINGVALUEINDEX -1))
   (CL:DECLARE (CL:TYPE CL:FIXNUM MISSINGVALUEINDEX))
   (CL:WHEN (CL:NOT (CL:EQ CODE NULL))
    (CL:LET*
     ((I NULL-INTEGER) (ARG NULL) (VECTOR-000 (%ARGUMENTS SELF)) (INDEX-000 0)
      (LENGTH-000 (LENGTH VECTOR-000)) (ITER-000 0) (COLLECT-000 NULL))
     (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000 ITER-000 I))
     (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
      (CL:PROGN
       (CL:SETQ ARG
        (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000)) INDEX-000))
       (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
      (CL:PROGN (CL:SETQ I ITER-000) (CL:SETQ ITER-000 (CL:1+ ITER-000)))
      (CL:WHEN (ISA? (VALUE-OF ARG) SGT-PROPAGATE-LOGIC-SKOLEM)
       (CL:IF (CL:= MISSINGVALUEINDEX -1)
        (CL:PROGN (CL:SETQ MISSINGVALUEINDEX I) (CL:SETQ ARG NULL))
        (CL:RETURN-FROM EVALUATE-FUNCTION-PROPOSITION)))
      (CL:IF (CL:EQ COLLECT-000 NULL)
       (CL:PROGN (CL:SETQ COLLECT-000 (CONS (VALUE-OF ARG) NIL))
        (CL:IF (CL:EQ BOUNDARGUMENTS NIL) (CL:SETQ BOUNDARGUMENTS COLLECT-000)
         (ADD-CONS-TO-END-OF-CONS-LIST BOUNDARGUMENTS COLLECT-000)))
       (CL:PROGN (CL:SETF (%%REST COLLECT-000) (CONS (VALUE-OF ARG) NIL))
        (CL:SETQ COLLECT-000 (%%REST COLLECT-000))))))
    (CL:WHEN (CL:= MISSINGVALUEINDEX -1)
     (CL:RETURN-FROM EVALUATE-FUNCTION-PROPOSITION))
    (CL:SETQ COMPUTEDVALUE
     (APPLY CODE (CONS (WRAP-INTEGER MISSINGVALUEINDEX) BOUNDARGUMENTS)))
    (CL:SETQ STOREDVALUE
     (VALUE-OF
      (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS SELF)))
       MISSINGVALUEINDEX)))
    (CL:WHEN
     (CL:AND (CL:NOT (EQL? COMPUTEDVALUE STOREDVALUE))
      (CL:NOT (CL:EQ COMPUTEDVALUE NULL)))
     (EQUATE-VALUES COMPUTEDVALUE STOREDVALUE))
    (CL:RETURN-FROM EVALUATE-FUNCTION-PROPOSITION))
   (CL:SETQ CODE (LOOKUP-COMPUTATION DESCRIPTION))
   (CL:WHEN (CL:NOT (CL:EQ CODE NULL))
    (CL:LET*
     ((ARG NULL) (ITER-001 (BUT-LAST (%ARGUMENTS SELF))) (COLLECT-001 NULL))
     (CL:LOOP WHILE (NEXT? ITER-001) DO (CL:SETQ ARG (%VALUE ITER-001))
      (CL:WHEN (ISA? (VALUE-OF ARG) SGT-PROPAGATE-LOGIC-SKOLEM)
       (CL:RETURN-FROM EVALUATE-FUNCTION-PROPOSITION))
      (CL:IF (CL:EQ COLLECT-001 NULL)
       (CL:PROGN (CL:SETQ COLLECT-001 (CONS (VALUE-OF ARG) NIL))
        (CL:IF (CL:EQ BOUNDARGUMENTS NIL) (CL:SETQ BOUNDARGUMENTS COLLECT-001)
         (ADD-CONS-TO-END-OF-CONS-LIST BOUNDARGUMENTS COLLECT-001)))
       (CL:PROGN (CL:SETF (%%REST COLLECT-001) (CONS (VALUE-OF ARG) NIL))
        (CL:SETQ COLLECT-001 (%%REST COLLECT-001))))))
    (CL:SETQ COMPUTEDVALUE (APPLY CODE BOUNDARGUMENTS))
    (CL:SETQ STOREDVALUE
     (VALUE-OF
      (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS SELF)))
       (CL:THE CL:FIXNUM (CL:1- (LENGTH (%ARGUMENTS SELF)))))))
    (CL:WHEN
     (CL:AND (CL:NOT (EQL? COMPUTEDVALUE STOREDVALUE))
      (CL:NOT (CL:EQ COMPUTEDVALUE NULL)))
     (EQUATE-VALUES COMPUTEDVALUE STOREDVALUE)))))

;;; (DEFUN EVALUATE-PREDICATE-PROPOSITION ...)

(CL:DEFUN EVALUATE-PREDICATE-PROPOSITION (SELF)
  (CL:LET*
   ((DESCRIPTION (GET-DESCRIPTION (%OPERATOR SELF)))
    (CODE (LOOKUP-CONSTRAINT DESCRIPTION)) (BOUNDARGUMENTS NIL)
    (SUCCESS? FALSE))
   (CL:WHEN (CL:NOT (CL:EQ CODE NULL))
    (CL:LET*
     ((ARG NULL) (VECTOR-000 (%ARGUMENTS SELF)) (INDEX-000 0)
      (LENGTH-000 (LENGTH VECTOR-000)) (COLLECT-000 NULL))
     (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
     (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
      (CL:PROGN
       (CL:SETQ ARG
        (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000)) INDEX-000))
       (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
      (CL:WHEN (ISA? (VALUE-OF ARG) SGT-PROPAGATE-LOGIC-SKOLEM)
       (CL:RETURN-FROM EVALUATE-PREDICATE-PROPOSITION))
      (CL:IF (CL:EQ COLLECT-000 NULL)
       (CL:PROGN (CL:SETQ COLLECT-000 (CONS ARG NIL))
        (CL:IF (CL:EQ BOUNDARGUMENTS NIL) (CL:SETQ BOUNDARGUMENTS COLLECT-000)
         (ADD-CONS-TO-END-OF-CONS-LIST BOUNDARGUMENTS COLLECT-000)))
       (CL:PROGN (CL:SETF (%%REST COLLECT-000) (CONS ARG NIL))
        (CL:SETQ COLLECT-000 (%%REST COLLECT-000))))))
    (CL:SETQ SUCCESS?
     (COERCE-WRAPPED-BOOLEAN-TO-BOOLEAN
      (APPLY CODE (CONS (WRAP-INTEGER -1) BOUNDARGUMENTS))))
    (ASSIGN-TRUTH-VALUE SELF (CL:IF SUCCESS? TRUE-WRAPPER FALSE-WRAPPER))
    (CL:RETURN-FROM EVALUATE-PREDICATE-PROPOSITION))
   (CL:SETQ CODE (LOOKUP-COMPUTATION DESCRIPTION))
   (CL:WHEN (CL:NOT (CL:EQ CODE NULL))
    (CL:LET*
     ((ARG NULL) (VECTOR-001 (%ARGUMENTS SELF)) (INDEX-001 0)
      (LENGTH-001 (LENGTH VECTOR-001)) (COLLECT-001 NULL))
     (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-001 LENGTH-001))
     (CL:LOOP WHILE (CL:< INDEX-001 LENGTH-001) DO
      (CL:PROGN
       (CL:SETQ ARG
        (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-001)) INDEX-001))
       (CL:SETQ INDEX-001 (CL:1+ INDEX-001)))
      (CL:WHEN (ISA? (VALUE-OF ARG) SGT-PROPAGATE-LOGIC-SKOLEM)
       (CL:RETURN-FROM EVALUATE-PREDICATE-PROPOSITION))
      (CL:IF (CL:EQ COLLECT-001 NULL)
       (CL:PROGN (CL:SETQ COLLECT-001 (CONS ARG NIL))
        (CL:IF (CL:EQ BOUNDARGUMENTS NIL) (CL:SETQ BOUNDARGUMENTS COLLECT-001)
         (ADD-CONS-TO-END-OF-CONS-LIST BOUNDARGUMENTS COLLECT-001)))
       (CL:PROGN (CL:SETF (%%REST COLLECT-001) (CONS ARG NIL))
        (CL:SETQ COLLECT-001 (%%REST COLLECT-001))))))
    (CL:SETQ SUCCESS?
     (COERCE-WRAPPED-BOOLEAN-TO-BOOLEAN (APPLY CODE BOUNDARGUMENTS)))
    (ASSIGN-TRUTH-VALUE SELF (CL:IF SUCCESS? TRUE-WRAPPER FALSE-WRAPPER)))))

;;; (DEFUN EVALUATE-PROPOSITION ...)

(CL:DEFUN EVALUATE-PROPOSITION (SELF)
  (EVALUATION-STATE-SETTER SELF KWD-PROPAGATE-EVALUATED)
  (CL:LET* ((TEST-VALUE-000 (%KIND SELF)))
   (CL:COND
    ((CL:EQ TEST-VALUE-000 KWD-PROPAGATE-AND) (EVALUATE-AND-PROPOSITION SELF))
    ((CL:EQ TEST-VALUE-000 KWD-PROPAGATE-OR) (EVALUATE-OR-PROPOSITION SELF))
    ((CL:EQ TEST-VALUE-000 KWD-PROPAGATE-NOT) (EVALUATE-NOT-PROPOSITION SELF))
    ((CL:EQ TEST-VALUE-000 KWD-PROPAGATE-EQUIVALENT)
     (EVALUATE-EQUIVALENT-PROPOSITION SELF))
    ((CL:EQ TEST-VALUE-000 KWD-PROPAGATE-ISA))
    ((CL:OR (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-FUNCTION)
      (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-PREDICATE))
     (CL:LET*
      ((DESCRIPTION (GET-DESCRIPTION (%OPERATOR SELF)))
       (EVALUATOR (LOOKUP-EVALUATOR DESCRIPTION)))
      (CL:COND ((CL:NOT (CL:EQ EVALUATOR NULL)) (CL:FUNCALL EVALUATOR SELF))
       ((CL:EQ (%KIND SELF) KWD-PROPAGATE-PREDICATE)
        (EVALUATE-PREDICATE-PROPOSITION SELF))
       (CL:T (EVALUATE-FUNCTION-PROPOSITION SELF)))))
    (CL:T))))

;;; (DEFMETHOD MARK-AS-INCOHERENT ...)

(CL:DEFMETHOD MARK-AS-INCOHERENT ((SELF LOGIC-OBJECT)))

;;; (DEFMETHOD MARK-AS-INCOHERENT ...)

(CL:DEFMETHOD MARK-AS-INCOHERENT ((SELF SKOLEM))
  (CL:WHEN (CL:NOT (CL:EQ (%DEFINING-PROPOSITION SELF) NULL))
   (CL:LET*
    ((FIRSTARG
      (CL:AREF
       (CL:THE CL:SIMPLE-VECTOR
        (%THE-ARRAY (%ARGUMENTS (%DEFINING-PROPOSITION SELF))))
       0)))
    (CL:COND
     ((SUBTYPE-OF? (SAFE-PRIMARY-TYPE FIRSTARG)
       SGT-PROPAGATE-LOGIC-LOGIC-OBJECT)
      (CL:PROGN (MARK-AS-INCOHERENT FIRSTARG) NULL))
     (CL:T)))))

;;; (DEFGLOBAL *JUST-IN-TIME-FORWARD-INFERENCE?* ...)

(CL:DEFVAR *JUST-IN-TIME-FORWARD-INFERENCE?* FALSE
  "If TRUE, ensures that forward propagation has
been applied to each instance 'touched' during a query.")

(CL:DEFMETHOD ELABORATED-IN-WORLDS ((SELF LOGIC-OBJECT))
  (CL:LET*
   ((ANSWER
     (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
      SYM-PROPAGATE-LOGIC-ELABORATED-IN-WORLDS NULL)))
   (CL:IF (CL:EQ ANSWER NULL) NIL-LIST ANSWER)))

;;; (DEFUN RECURSIVELY-REACT-TO-INFERENCE-UPDATE ...)

(CL:DEFUN RECURSIVELY-REACT-TO-INFERENCE-UPDATE (SELF)
  (REACT-TO-INFERENCE-UPDATE SELF)
  (CL:LET*
   ((ARG NULL) (VECTOR-000 (%ARGUMENTS SELF)) (INDEX-000 0)
    (LENGTH-000 (LENGTH VECTOR-000)))
   (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
   (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
    (CL:PROGN
     (CL:SETQ ARG
      (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000)) INDEX-000))
     (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
    (CL:WHEN (ISA? ARG SGT-PROPAGATE-LOGIC-PROPOSITION)
     (RECURSIVELY-REACT-TO-INFERENCE-UPDATE ARG)))))

;;; (DEFUN ELABORATE-META-INSTANCE ...)

(CL:DEFUN ELABORATE-META-INSTANCE (SELF)
  (CL:LET*
   ((WORLDSTATE (GET-WORLD-STATE *CONTEXT*))
    (METAWORLD (GET-INFERENCE-CACHE WORLDSTATE KWD-PROPAGATE-META)))
   (CL:LET* ((*CONTEXT* METAWORLD)) (CL:DECLARE (CL:SPECIAL *CONTEXT*))
    (ELABORATE-INSTANCE SELF))))

;;; (DEFUN ELABORATE-INSTANCE ...)

(CL:DEFUN ELABORATE-INSTANCE (SELF)
  (CL:WHEN *FILLINGCONSTRAINTPROPAGATIONQUEUES?*
   (CL:RETURN-FROM ELABORATE-INSTANCE))
  (CL:WHEN *JUST-IN-TIME-FORWARD-INFERENCE?*
   (CL:COND
    ((SUBTYPE-OF? (SAFE-PRIMARY-TYPE SELF) SGT-PROPAGATE-LOGIC-LOGIC-OBJECT)
     (CL:PROGN
      (CL:WHEN (MEMBER? (ELABORATED-IN-WORLDS SELF) *CONTEXT*)
       (CL:RETURN-FROM ELABORATE-INSTANCE))
      (CL:WHEN
       (CL:AND (CL:NOT (CL:EQ *TRACED-KEYWORDS* NULL))
        (MEMB? *TRACED-KEYWORDS* KWD-PROPAGATE-ELABORATE))
       (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-OUTPUT) " ELABORATING:  " SELF
        " in world " *CONTEXT* EOL))
      (CL:IF (CL:EQ (ELABORATED-IN-WORLDS SELF) NIL-LIST)
       (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
        SYM-PROPAGATE-LOGIC-ELABORATED-IN-WORLDS (LIST) NULL)
       (REMOVE-DELETED-MEMBERS (ELABORATED-IN-WORLDS SELF)))
      (PUSH (ELABORATED-IN-WORLDS SELF) *CONTEXT*)
      (CL:LET*
       ((*EVALUATIONMODE* KWD-PROPAGATE-EXTENSIONAL-ASSERTION)
        (*INHIBITOBJECTFINALIZATION?* TRUE) (*QUERYITERATOR* NULL))
       (CL:DECLARE
        (CL:SPECIAL *EVALUATIONMODE* *INHIBITOBJECTFINALIZATION?*
         *QUERYITERATOR*))
       (CL:HANDLER-CASE
        (CL:PROGN (INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES *CONTEXT*)
         (CL:LET* ((*FILLINGCONSTRAINTPROPAGATIONQUEUES?* TRUE))
          (CL:DECLARE (CL:SPECIAL *FILLINGCONSTRAINTPROPAGATIONQUEUES?*))
          (EVALUATE-REACHABLE-INEQUALITIES SELF (LIST))
          (CL:LET*
           ((PROP NULL)
            (ITER-000
             (%THE-CONS-LIST (ALL-FACTS-OF-INSTANCE SELF TRUE FALSE))))
           (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
            (CL:PROGN (CL:SETQ PROP (%%VALUE ITER-000))
             (CL:SETQ ITER-000 (%%REST ITER-000)))
            (RECURSIVELY-REACT-TO-INFERENCE-UPDATE PROP)))
          (CL:LET* ((*DEFERINGDEFAULTFORWARDINFERENCES?* TRUE))
           (CL:DECLARE (CL:SPECIAL *DEFERINGDEFAULTFORWARDINFERENCES?*))
           (EXECUTE-CONSTRAINT-PROPAGATION-QUEUES))
          (CL:LET*
           ((P NULL) (ITER-001 (DEFERRED-DEFAULT-PROPOSITIONS *CONTEXT*)))
           (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NIL)) DO
            (CL:PROGN (CL:SETQ P (%%VALUE ITER-001))
             (CL:SETQ ITER-001 (%%REST ITER-001)))
            (ASSIGN-TRUTH-VALUE P DEFAULT-TRUE-TRUTH-VALUE)))
          (DEFERRED-DEFAULT-PROPOSITIONS-SETTER *CONTEXT* NIL))
         (EXECUTE-CONSTRAINT-PROPAGATION-QUEUES))
        (CLASH (E)
         (%%PRINT-STREAM (%NATIVE-STREAM STANDARD-ERROR) (EXCEPTION-MESSAGE E)
          EOL))))))
    (CL:T))))

;;; (DEFUN EVALUATE-REACHABLE-INEQUALITIES ...)

(CL:DEFUN EVALUATE-REACHABLE-INEQUALITIES (SELF VISITEDLIST)
  (CL:LET*
   ((P NULL)
    (ITER-000
     (ALL-TRUE-DEPENDENT-PROPOSITIONS SELF
      SGT-PROPAGATE-PL-KERNEL-KB-INEQUALITY TRUE)))
   (CL:LOOP WHILE (NEXT? ITER-000) DO (CL:SETQ P (%VALUE ITER-000))
    (CL:WHEN (CL:EQ (EVALUATION-STATE P) NULL) (POST-FOR-EVALUATION P))
    (PUSH VISITEDLIST SELF)
    (CL:LET*
     ((ARG NULL) (VECTOR-000 (%ARGUMENTS P)) (INDEX-000 0)
      (LENGTH-000 (LENGTH VECTOR-000)))
     (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
     (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
      (CL:PROGN
       (CL:SETQ ARG
        (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000)) INDEX-000))
       (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
      (CL:WHEN
       (CL:AND (CL:NOT (MEMBER? VISITEDLIST ARG))
        (ISA? ARG SGT-PROPAGATE-LOGIC-LOGIC-OBJECT))
       (EVALUATE-REACHABLE-INEQUALITIES ARG VISITEDLIST)))))))

;;; (DEFUN ELABORATE-SURROGATES-IN-PROPOSITION ...)

(CL:DEFUN ELABORATE-SURROGATES-IN-PROPOSITION (PROPOSITION)
  (CL:WHEN *JUST-IN-TIME-FORWARD-INFERENCE?*
   (CL:LET* ((TEST-VALUE-000 (%KIND PROPOSITION)))
    (CL:COND
     ((CL:OR (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-PREDICATE)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-FUNCTION)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-ISA))
      (ELABORATE-META-INSTANCE (GET-DESCRIPTION (%OPERATOR PROPOSITION)))
      (CL:LET*
       ((ARG NULL) (VECTOR-000 (%ARGUMENTS PROPOSITION)) (INDEX-000 0)
        (LENGTH-000 (LENGTH VECTOR-000)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
       (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
        (CL:PROGN
         (CL:SETQ ARG
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
           INDEX-000))
         (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
        (CL:WHEN (ISA? ARG SGT-PROPAGATE-STELLA-SURROGATE)
         (CL:LET* ((VALUE (VALUE-OF ARG)))
          (CL:WHEN (CL:NOT (CL:EQ VALUE NULL))
           (ELABORATE-INSTANCE VALUE)))))))
     ((CL:OR (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-AND)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-OR)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-NOT)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-EXISTS)
       (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-FORALL))
      (CL:LET*
       ((P NULL) (VECTOR-001 (%ARGUMENTS PROPOSITION)) (INDEX-001 0)
        (LENGTH-001 (LENGTH VECTOR-001)))
       (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-001 LENGTH-001))
       (CL:LOOP WHILE (CL:< INDEX-001 LENGTH-001) DO
        (CL:PROGN
         (CL:SETQ P
          (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-001))
           INDEX-001))
         (CL:SETQ INDEX-001 (CL:1+ INDEX-001)))
        (ELABORATE-SURROGATES-IN-PROPOSITION P))))
     (CL:T)))))

;;; (DEFUN COLLECT-FUNCTION-PROPOSITION-FACTS ...)

(CL:DEFUN COLLECT-FUNCTION-PROPOSITION-FACTS (SELF FACTS BEENTHERE INCLUDEUNKNOWN?)
  (INSERT BEENTHERE SELF)
  (CL:LET*
   ((OUTPUTSKOLEM
     (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY (%ARGUMENTS SELF)))
      (CL:THE CL:FIXNUM (CL:1- (LENGTH (%ARGUMENTS SELF))))))
    (VALUE (VALUE-OF OUTPUTSKOLEM))
    (HASASSERTEDVALUE? (CL:NOT (EQL? VALUE OUTPUTSKOLEM))))
   (CL:WHEN HASASSERTEDVALUE?
    (CL:WHEN (CL:NOT (MEMBER? FACTS SELF)) (INSERT FACTS SELF)))
   (CL:WHEN
    (CL:OR (CL:NOT HASASSERTEDVALUE?)
     (CL:AND (NATIVE-VALUE? VALUE)
      (CL:NOT (ISA? VALUE SGT-PROPAGATE-STELLA-LITERAL-WRAPPER))))
    (HELP-COLLECT-FACTS OUTPUTSKOLEM FACTS BEENTHERE INCLUDEUNKNOWN?))))

;;; (DEFUN HELP-COLLECT-FACTS ...)

(CL:DEFUN HELP-COLLECT-FACTS (SELF FACTS BEENTHERE INCLUDEUNKNOWN?)
  (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE SELF)))
   (CL:COND
    ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-LOGIC-OBJECT)
     (CL:PROGN
      (CL:LET*
       ((P NULL)
        (ITER-000
         (ALLOCATE-ITERATOR (UNFILTERED-DEPENDENT-PROPOSITIONS SELF NULL))))
       (CL:LOOP WHILE (NEXT? ITER-000) DO (CL:SETQ P (%VALUE ITER-000))
        (CL:WHEN (CL:NOT (MEMBER? BEENTHERE P))
         (CL:IF (CL:EQ (%KIND P) KWD-PROPAGATE-FUNCTION)
          (COLLECT-FUNCTION-PROPOSITION-FACTS P FACTS BEENTHERE
           INCLUDEUNKNOWN?)
          (HELP-COLLECT-FACTS P FACTS BEENTHERE INCLUDEUNKNOWN?)))))
      (CL:LET* ((EQUIVALENT NULL) (ITER-001 (VARIABLE-VALUE-INVERSE SELF)))
       (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NIL)) DO
        (CL:PROGN (CL:SETQ EQUIVALENT (%%VALUE ITER-001))
         (CL:SETQ ITER-001 (%%REST ITER-001)))
        (HELP-COLLECT-FACTS EQUIVALENT FACTS BEENTHERE INCLUDEUNKNOWN?)))))
    ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-LOGIC-PROPOSITION)
     (CL:PROGN (INSERT BEENTHERE SELF)
      (CL:WHEN (CL:OR (CL:NOT (UNKNOWN? SELF)) INCLUDEUNKNOWN?)
       (INSERT FACTS SELF))
      (CL:LET*
       ((P NULL)
        (ITER-002 (ALLOCATE-ITERATOR (%DEPENDENT-PROPOSITIONS SELF))))
       (CL:LOOP WHILE (NEXT? ITER-002) DO (CL:SETQ P (%VALUE ITER-002))
        (CL:WHEN (CL:NOT (MEMBER? BEENTHERE P))
         (HELP-COLLECT-FACTS P FACTS BEENTHERE INCLUDEUNKNOWN?))))))
    ((SUBTYPE-OF-SURROGATE? TEST-VALUE-000)
     (CL:PROGN
      (CL:WHEN (CL:NOT (CL:EQ (%SURROGATE-VALUE SELF) NULL))
       (HELP-COLLECT-FACTS (%SURROGATE-VALUE SELF) FACTS BEENTHERE
        INCLUDEUNKNOWN?))))
    ((SUBTYPE-OF-CLASS? TEST-VALUE-000)
     (CL:PROGN
      (CL:WHEN
       (CL:NOT
        (CL:EQ
         (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
          SYM-PROPAGATE-LOGIC-DESCRIPTION NULL)
         NULL))
       (HELP-COLLECT-FACTS
        (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
         SYM-PROPAGATE-LOGIC-DESCRIPTION NULL)
        FACTS BEENTHERE INCLUDEUNKNOWN?))))
    ((SUBTYPE-OF? TEST-VALUE-000 SGT-PROPAGATE-STELLA-SLOT)
     (CL:PROGN
      (CL:WHEN
       (CL:NOT
        (CL:EQ
         (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
          SYM-PROPAGATE-LOGIC-DESCRIPTION NULL)
         NULL))
       (HELP-COLLECT-FACTS
        (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
         SYM-PROPAGATE-LOGIC-DESCRIPTION NULL)
        FACTS BEENTHERE INCLUDEUNKNOWN?))))
    (CL:T))))

;;; (DEFUN (ALL-FACTS-OF-INSTANCE (LIST OF PROPOSITION)) ...)

(CL:DEFUN ALL-FACTS-OF-INSTANCE (SELF INCLUDEUNKNOWNFACTS? ELABORATE?)
  "Return a list of all definite (TRUE or FALSE) propositions
attached to `self'."
  (CL:LET* ((FACTS (NEW-LIST)) (BEENTHERE (NEW-LIST)))
   (CL:LET*
    ((*CONTEXT* (GET-QUERY-CONTEXT)) (*MODULE* (%BASE-MODULE *CONTEXT*)))
    (CL:DECLARE (CL:SPECIAL *CONTEXT* *MODULE*))
    (CL:WHEN ELABORATE? (ELABORATE-INSTANCE SELF))
    (HELP-COLLECT-FACTS SELF FACTS BEENTHERE INCLUDEUNKNOWNFACTS?))
   FACTS))

;;; (DEFUN (ALL-FACTS-OF (CONS OF PROPOSITION)) ...)

(CL:DEFUN %ALL-FACTS-OF (INSTANCEREF)
  "Return a cons list of all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'.  This includes propositions
asserted to be true by default, but it does not include propositions
that are found to be TRUE only by running the query engine.  Facts
inferred to be TRUE by the forward chainer will be included.
Hence, the returned list of facts may be longer in a context where the
forward chainer has been run then in one where it has not (see
`run-forward-rules')."
  (CL:LET* ((INSTANCE (COERCE-TO-INSTANCE INSTANCEREF NULL)))
   (CL:IF (CL:NOT (CL:EQ INSTANCE NULL))
    (%THE-CONS-LIST (ALL-FACTS-OF-INSTANCE INSTANCE FALSE TRUE)) NULL)))

(CL:DEFMACRO ALL-FACTS-OF (CL:&WHOLE EXPRESSION CL:&REST IGNORE)
  "Return a cons list of all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'.  This includes propositions
asserted to be true by default, but it does not include propositions
that are found to be TRUE only by running the query engine.  Facts
inferred to be TRUE by the forward chainer will be included.
Hence, the returned list of facts may be longer in a context where the
forward chainer has been run then in one where it has not (see
`run-forward-rules')."
  (CL:DECLARE (CL:IGNORE IGNORE))
  (CL:LET ((*IGNORETRANSLATIONERRORS?* FALSE))
   (CL-INCREMENTALLY-TRANSLATE EXPRESSION)))

(CL:SETF (CL:MACRO-FUNCTION (CL:QUOTE /LOGIC/ALL-FACTS-OF)) (CL:MACRO-FUNCTION (CL:QUOTE ALL-FACTS-OF)))

;;; (DEFUN (CALL-ALL-FACTS-OF (LIST OF PROPOSITION)) ...)

(CL:DEFUN %CALL-ALL-FACTS-OF (INSTANCEREF)
  "Return a list of all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'."
  (CL:LET* ((INSTANCE (COERCE-TO-INSTANCE INSTANCEREF NULL)))
   (CL:IF (CL:NOT (CL:EQ INSTANCE NULL))
    (ALL-FACTS-OF-INSTANCE INSTANCE FALSE TRUE) NULL)))

(CL:DEFMACRO CALL-ALL-FACTS-OF (CL:&WHOLE EXPRESSION CL:&REST IGNORE)
  "Return a list of all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'."
  (CL:DECLARE (CL:IGNORE IGNORE))
  (CL:LET ((*IGNORETRANSLATIONERRORS?* FALSE))
   (CL-INCREMENTALLY-TRANSLATE EXPRESSION)))

(CL:SETF (CL:MACRO-FUNCTION (CL:QUOTE /LOGIC/CALL-ALL-FACTS-OF)) (CL:MACRO-FUNCTION (CL:QUOTE CALL-ALL-FACTS-OF)))

;;; (DEFUN RETRACT-FACTS-OF-INSTANCE ...)

(CL:DEFUN RETRACT-FACTS-OF-INSTANCE (SELF)
  "Retract all definite (TRUE or FALSE) propositions attached to `self'."
  (CL:WHEN (CL:EQ SELF NULL) (CL:RETURN-FROM RETRACT-FACTS-OF-INSTANCE))
  (CL:LET*
   ((P NULL)
    (ITER-000 (%THE-CONS-LIST (ALL-FACTS-OF-INSTANCE SELF FALSE TRUE))))
   (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
    (CL:PROGN (CL:SETQ P (%%VALUE ITER-000))
     (CL:SETQ ITER-000 (%%REST ITER-000)))
    (CL:TAGBODY
     (CL:COND
      ((CL:EQ (%KIND P) KWD-PROPAGATE-FUNCTION)
       (CL:LET* ((SKOLEM (LAST (%ARGUMENTS P))))
        (CL:WHEN (CL:EQ SKOLEM SELF) (CL:GO :CONTINUE))
        (CL:WHEN
         (CL:AND (CL:NOT (CL:EQ SKOLEM NULL))
          (CL:NOT
           (CL:EQ
            (ACCESS-IN-CONTEXT (%VARIABLE-VALUE SKOLEM) (%HOME-CONTEXT SKOLEM)
             FALSE)
            NULL)))
         (UNASSERT-FACT P))
        (RETRACT-FACTS-OF-INSTANCE SKOLEM)))
      (CL:T (UNASSERT-FACT P)))
     :CONTINUE))))

;;; (DEFUN RETRACT-FACTS-OF ...)

(CL:DEFUN %RETRACT-FACTS-OF (INSTANCEREF)
  "Retract all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'."
  (CL:LET* ((INSTANCE (COERCE-TO-INSTANCE INSTANCEREF NULL)))
   (CL:WHEN (CL:NOT (CL:EQ INSTANCE NULL))
    (RETRACT-FACTS-OF-INSTANCE INSTANCE))))

(CL:DEFMACRO RETRACT-FACTS-OF (CL:&WHOLE EXPRESSION CL:&REST IGNORE)
  "Retract all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'."
  (CL:DECLARE (CL:IGNORE IGNORE))
  (CL:LET ((*IGNORETRANSLATIONERRORS?* FALSE))
   (CL-INCREMENTALLY-TRANSLATE EXPRESSION)))

(CL:SETF (CL:MACRO-FUNCTION (CL:QUOTE /LOGIC/RETRACT-FACTS-OF)) (CL:MACRO-FUNCTION (CL:QUOTE RETRACT-FACTS-OF)))

;;; (DEFUN (EXPLODE-VARIABLIZED-STRING LIST) ...)

(CL:DEFUN EXPLODE-VARIABLIZED-STRING (STRING)
  (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING STRING))
  #+MCL
  (CL:CHECK-TYPE STRING CL:SIMPLE-STRING)
  (CL:LET*
   ((LIST (NEW-LIST)) (ENDOFSTRING (CL:THE CL:FIXNUM (CL:LENGTH STRING)))
    (BEGIN 0) (END NULL-INTEGER) (SUBSTRING STELLA::NULL-STRING)
    (SYMBOL NULL))
   (CL:DECLARE (CL:TYPE CL:FIXNUM ENDOFSTRING BEGIN END)
    (CL:TYPE CL:SIMPLE-STRING SUBSTRING))
   (CL:LOOP
    (CL:WHEN (CL:>= BEGIN ENDOFSTRING)
     (CL:RETURN-FROM EXPLODE-VARIABLIZED-STRING LIST))
    (CL:SETQ END (POSITION STRING #\? BEGIN))
    (CL:COND
     ((NULL? END) (CL:SETQ SUBSTRING (SUBSEQUENCE STRING BEGIN ENDOFSTRING))
      (INSERT-LAST LIST (WRAP-STRING SUBSTRING))
      (CL:RETURN-FROM EXPLODE-VARIABLIZED-STRING LIST))
     ((CL:> END BEGIN) (CL:SETQ SUBSTRING (SUBSEQUENCE STRING BEGIN END))
      (INSERT-LAST LIST (WRAP-STRING SUBSTRING))))
    (CL:SETQ BEGIN END) (CL:SETQ END (POSITION STRING #\  BEGIN))
    (CL:WHEN (NULL? END) (CL:SETQ END ENDOFSTRING))
    (CL:SETQ SUBSTRING (SUBSEQUENCE STRING BEGIN END))
    (CL:SETQ SYMBOL (INTERN-SYMBOL (STRING-UPCASE SUBSTRING)))
    (INSERT-LAST LIST SYMBOL) (CL:SETQ BEGIN END))))

;;; (DEFUN (FACT-TO-SENTENCE STRING) ...)

(CL:DEFUN FACT-TO-SENTENCE (SELF PERIOD?)
  (CL:LET* ((TEST-VALUE-000 (%KIND SELF)))
   (CL:COND
    ((CL:OR (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-PREDICATE)
      (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-FUNCTION)
      (CL:EQ TEST-VALUE-000 KWD-PROPAGATE-ISA))
     (CL:LET*
      ((RELATION (%SURROGATE-VALUE (%OPERATOR SELF)))
       (PHRASE
        (ACCESS-BINARY-VALUE RELATION SGT-PROPAGATE-PL-KERNEL-KB-PHRASE))
       (EXPLODEDSTRING
        (CL:IF (CL:NOT (CL:EQ PHRASE NULL))
         (EXPLODE-VARIABLIZED-STRING (%WRAPPER-VALUE PHRASE)) NULL)))
      (CL:WHEN (CL:NOT (CL:EQ EXPLODEDSTRING NULL))
       (CL:LET* ((TABLE (NEW-KEY-VALUE-LIST)) (SENTENCE ""))
        (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING SENTENCE))
        (CL:LET*
         ((ARG NULL) (VBL NULL)
          (ITER-000 (%THE-CONS-LIST (%IO-VARIABLE-NAMES RELATION)))
          (VECTOR-000 (%ARGUMENTS SELF)) (INDEX-000 0)
          (LENGTH-000 (LENGTH VECTOR-000)))
         (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
         (CL:LOOP WHILE
          (CL:AND (CL:NOT (CL:EQ ITER-000 NIL)) (CL:< INDEX-000 LENGTH-000))
          DO
          (CL:PROGN (CL:SETQ VBL (%%VALUE ITER-000))
           (CL:SETQ ITER-000 (%%REST ITER-000)))
          (CL:PROGN
           (CL:SETQ ARG
            (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
             INDEX-000))
           (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
          (INSERT-AT TABLE VBL ARG)))
        (CL:LET* ((ITEM NULL) (ITER-001 (%THE-CONS-LIST EXPLODEDSTRING)))
         (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NIL)) DO
          (CL:PROGN (CL:SETQ ITEM (%%VALUE ITER-001))
           (CL:SETQ ITER-001 (%%REST ITER-001)))
          (CL:LET* ((VALUE NULL) (STRINGVALUE STELLA::NULL-STRING))
           (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING STRINGVALUE))
           (CL:LET* ((TEST-VALUE-002 (SAFE-PRIMARY-TYPE ITEM)))
            (CL:COND
             ((SUBTYPE-OF-SYMBOL? TEST-VALUE-002)
              (CL:PROGN (CL:SETQ VALUE (LOOKUP TABLE ITEM))
               (CL:WHEN (CL:NOT (CL:EQ VALUE NULL))
                (CL:LET* ((TEST-VALUE-001 (SAFE-PRIMARY-TYPE VALUE)))
                 (CL:COND
                  ((SUBTYPE-OF? TEST-VALUE-001 SGT-PROPAGATE-LOGIC-SKOLEM)
                   (CL:PROGN
                    (CL:LET*
                     ((SURROGATE (%SURROGATE-VALUE-INVERSE VALUE))
                      (DEFININGPROPOSITION (%DEFINING-PROPOSITION VALUE)))
                     (CL:COND
                      ((CL:NOT (CL:EQ DEFININGPROPOSITION NULL))
                       (CL:SETQ STRINGVALUE
                        (FACT-TO-SENTENCE DEFININGPROPOSITION FALSE)))
                      ((CL:NOT (CL:EQ SURROGATE NULL))
                       (CL:SETQ STRINGVALUE (%SYMBOL-NAME SURROGATE)))
                      (CL:T
                       (CL:SETQ STRINGVALUE
                        (%SYMBOL-NAME (%SKOLEM-NAME VALUE))))))))
                  ((SUBTYPE-OF? TEST-VALUE-001
                    SGT-PROPAGATE-LOGIC-LOGIC-OBJECT)
                   (CL:PROGN
                    (CL:LET* ((SURROGATE (%SURROGATE-VALUE-INVERSE VALUE)))
                     (CL:WHEN (CL:NOT (CL:EQ SURROGATE NULL))
                      (CL:SETQ STRINGVALUE (%SYMBOL-NAME SURROGATE))))))
                  ((SUBTYPE-OF-STRING? TEST-VALUE-001)
                   (CL:PROGN
                    (CL:SETQ STRINGVALUE
                     (CONCATENATE "'" (%WRAPPER-VALUE VALUE) "'"))))
                  ((SUBTYPE-OF-INTEGER? TEST-VALUE-001)
                   (CL:PROGN
                    (CL:SETQ STRINGVALUE
                     (INTEGER-TO-STRING (%WRAPPER-VALUE VALUE)))))
                  ((SUBTYPE-OF-FLOAT? TEST-VALUE-001)
                   (CL:PROGN
                    (CL:SETQ STRINGVALUE
                     (FLOAT-TO-STRING (%WRAPPER-VALUE VALUE)))))
                  (CL:T))))
               (CL:IF (CL:NOT (CL:EQ STRINGVALUE STELLA::NULL-STRING))
                (CL:SETQ SENTENCE (CONCATENATE SENTENCE STRINGVALUE))
                (CL:SETQ SENTENCE
                 (CONCATENATE SENTENCE (%SYMBOL-NAME ITEM))))))
             ((SUBTYPE-OF-STRING? TEST-VALUE-002)
              (CL:PROGN
               (CL:SETQ SENTENCE
                (CONCATENATE SENTENCE (%WRAPPER-VALUE ITEM)))))
             (CL:T
              (CL:LET* ((STREAM-000 (NEW-OUTPUT-STRING-STREAM)))
               (%%PRINT-STREAM (%NATIVE-STREAM STREAM-000) "`" TEST-VALUE-002
                "' is not a valid case option")
               (CL:ERROR
                (NEW-STELLA-EXCEPTION (THE-STRING-READER STREAM-000))))))))))
        (CL:WHEN PERIOD? (CL:SETQ SENTENCE (CONCATENATE SENTENCE ".")))
        (CL:RETURN-FROM FACT-TO-SENTENCE SENTENCE)))
      STELLA::NULL-STRING))
    ((CL:EQ TEST-VALUE-000 KWD-PROPAGATE-IMPLIES) "rules not yet implemented")
    (CL:T "???"))))
(CL:DECLAIM
 (CL:FTYPE (CL:FUNCTION (CL:T CL:T) CL:SIMPLE-STRING) FACT-TO-SENTENCE))

;;; (DEFUN (ALL-SENTENCES-OF (CONS OF STRING-WRAPPER)) ...)

(CL:DEFUN %ALL-SENTENCES-OF (INSTANCEREF)
  "Return a list of sentences describing facts about 'instanceRef'."
  (CL:LET*
   ((INSTANCE (COERCE-TO-INSTANCE INSTANCEREF NULL))
    (SENTENCE STELLA::NULL-STRING))
   (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING SENTENCE))
   (CL:IF (CL:NOT (CL:EQ INSTANCE NULL))
    (CL:LET* ((LIST NIL))
     (CL:LET*
      ((FACT NULL)
       (ITER-000 (%THE-CONS-LIST (ALL-FACTS-OF-INSTANCE INSTANCE FALSE TRUE)))
       (COLLECT-000 NULL))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
       (CL:PROGN (CL:SETQ FACT (%%VALUE ITER-000))
        (CL:SETQ ITER-000 (%%REST ITER-000)))
       (CL:TAGBODY
        (CL:PROGN (CL:SETQ SENTENCE (FACT-TO-SENTENCE FACT TRUE))
         (CL:WHEN (CL:EQ SENTENCE STELLA::NULL-STRING) (CL:GO :CONTINUE)))
        (CL:IF (CL:EQ COLLECT-000 NULL)
         (CL:PROGN (CL:SETQ COLLECT-000 (CONS (WRAP-STRING SENTENCE) NIL))
          (CL:IF (CL:EQ LIST NIL) (CL:SETQ LIST COLLECT-000)
           (ADD-CONS-TO-END-OF-CONS-LIST LIST COLLECT-000)))
         (CL:PROGN
          (CL:SETF (%%REST COLLECT-000) (CONS (WRAP-STRING SENTENCE) NIL))
          (CL:SETQ COLLECT-000 (%%REST COLLECT-000))))
        :CONTINUE)))
     LIST)
    NULL)))

(CL:DEFMACRO ALL-SENTENCES-OF (CL:&WHOLE EXPRESSION CL:&REST IGNORE)
  "Return a list of sentences describing facts about 'instanceRef'."
  (CL:DECLARE (CL:IGNORE IGNORE))
  (CL:LET ((*IGNORETRANSLATIONERRORS?* FALSE))
   (CL-INCREMENTALLY-TRANSLATE EXPRESSION)))

(CL:SETF (CL:MACRO-FUNCTION (CL:QUOTE /LOGIC/ALL-SENTENCES-OF)) (CL:MACRO-FUNCTION (CL:QUOTE ALL-SENTENCES-OF)))

;;; (DEFUN (CALL-ALL-FACTS-OF-N (LIST OF PROPOSITION)) ...)

(CL:DEFUN CALL-ALL-FACTS-OF-N (N SEEDREFS)
  (CL:DECLARE (CL:TYPE CL:FIXNUM N))
  #+MCL
  (CL:CHECK-TYPE N CL:FIXNUM)
  (CL:LET*
   ((VISITED (NEW-HASH-TABLE)) (FACTS (NEW-HASH-TABLE)) (FACTLIST (NEW-LIST))
    (SEEDS (NEW-LIST)) (NEWSEEDS (NEW-LIST)) (INSTANCE NULL))
   (CL:LET* ((SEEDREF NULL) (ITER-000 SEEDREFS))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
     (CL:PROGN (CL:SETQ SEEDREF (%%VALUE ITER-000))
      (CL:SETQ ITER-000 (%%REST ITER-000)))
     (CL:SETQ INSTANCE (COERCE-TO-INSTANCE SEEDREF NULL))
     (CL:WHEN (CL:NOT (CL:EQ INSTANCE NULL)) (INSERT-NEW SEEDS INSTANCE))))
   (CL:LET*
    ((I NULL-INTEGER) (ITER-001 0) (UPPER-BOUND-000 (CL:IF (DEFINED? N) N 0))
     (UNBOUNDED?-000 (NULL? UPPER-BOUND-000)))
    (CL:DECLARE (CL:TYPE CL:FIXNUM ITER-001 UPPER-BOUND-000 I))
    (CL:LOOP WHILE (CL:OR UNBOUNDED?-000 (CL:<= ITER-001 UPPER-BOUND-000)) DO
     (CL:PROGN (CL:SETQ I ITER-001) (CL:SETQ ITER-001 (CL:1+ ITER-001)))
     (CL:LET* ((SEED NULL) (ITER-002 (%THE-CONS-LIST SEEDS)))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-002 NIL)) DO
       (CL:PROGN (CL:SETQ SEED (%%VALUE ITER-002))
        (CL:SETQ ITER-002 (%%REST ITER-002)))
       (INSERT-AT VISITED SEED SEED)))
     (CL:LET* ((SEED NULL) (ITER-003 (%THE-CONS-LIST SEEDS)))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-003 NIL)) DO
       (CL:PROGN (CL:SETQ SEED (%%VALUE ITER-003))
        (CL:SETQ ITER-003 (%%REST ITER-003)))
       (CL:LET*
        ((PROP NULL)
         (ITER-004 (%THE-CONS-LIST (ALL-FACTS-OF-INSTANCE SEED FALSE TRUE))))
        (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-004 NIL)) DO
         (CL:PROGN (CL:SETQ PROP (%%VALUE ITER-004))
          (CL:SETQ ITER-004 (%%REST ITER-004)))
         (CL:TAGBODY
          (CL:WHEN (CL:NOT (CL:EQ (LOOKUP FACTS PROP) NULL))
           (CL:GO :CONTINUE))
          (INSERT-AT FACTS PROP PROP) (INSERT FACTLIST PROP)
          (CL:LET*
           ((ARG NULL) (VECTOR-000 (%ARGUMENTS PROP)) (INDEX-000 0)
            (LENGTH-000 (LENGTH VECTOR-000)))
           (CL:DECLARE (CL:TYPE CL:FIXNUM INDEX-000 LENGTH-000))
           (CL:LOOP WHILE (CL:< INDEX-000 LENGTH-000) DO
            (CL:PROGN
             (CL:SETQ ARG
              (CL:AREF (CL:THE CL:SIMPLE-VECTOR (%THE-ARRAY VECTOR-000))
               INDEX-000))
             (CL:SETQ INDEX-000 (CL:1+ INDEX-000)))
            (CL:WHEN (CL:EQ (LOOKUP VISITED ARG) NULL)
             (INSERT-NEW NEWSEEDS ARG))))
          :CONTINUE)))))
     (CL:SETQ SEEDS NEWSEEDS) (CL:SETQ NEWSEEDS (NEW-LIST))))
   (REVERSE FACTLIST)))

;;; (DEFUN (FILTERED-ALL-FACTS-OF-N (LIST OF PROPOSITION)) ...)

(CL:DEFUN FILTERED-ALL-FACTS-OF-N (N SEEDREFS FILTER)
  (CL:DECLARE (CL:TYPE CL:FIXNUM N))
  #+MCL
  (CL:CHECK-TYPE N CL:FIXNUM)
  (CL:LET*
   ((VISITED (NEW-HASH-TABLE)) (FACTS (NEW-HASH-TABLE)) (FACTLIST (NEW-LIST))
    (SEEDS (NEW-LIST)) (NEWSEEDS (NEW-LIST)) (INSTANCE NULL))
   (CL:LET* ((SEEDREF NULL) (ITER-000 SEEDREFS))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
     (CL:PROGN (CL:SETQ SEEDREF (%%VALUE ITER-000))
      (CL:SETQ ITER-000 (%%REST ITER-000)))
     (CL:SETQ INSTANCE (COERCE-TO-INSTANCE SEEDREF NULL))
     (CL:WHEN (CL:NOT (CL:EQ INSTANCE NULL)) (INSERT-NEW SEEDS INSTANCE))))
   (CL:LET*
    ((I NULL-INTEGER) (ITER-001 0) (UPPER-BOUND-000 (CL:IF (DEFINED? N) N 0))
     (UNBOUNDED?-000 (NULL? UPPER-BOUND-000)))
    (CL:DECLARE (CL:TYPE CL:FIXNUM ITER-001 UPPER-BOUND-000 I))
    (CL:LOOP WHILE (CL:OR UNBOUNDED?-000 (CL:<= ITER-001 UPPER-BOUND-000)) DO
     (CL:PROGN (CL:SETQ I ITER-001) (CL:SETQ ITER-001 (CL:1+ ITER-001)))
     (CL:LET* ((SEED NULL) (ITER-002 (%THE-CONS-LIST SEEDS)))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-002 NIL)) DO
       (CL:PROGN (CL:SETQ SEED (%%VALUE ITER-002))
        (CL:SETQ ITER-002 (%%REST ITER-002)))
       (INSERT-AT VISITED SEED SEED)))
     (CL:LET* ((SEED NULL) (ITER-003 (%THE-CONS-LIST SEEDS)))
      (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-003 NIL)) DO
       (CL:PROGN (CL:SETQ SEED (%%VALUE ITER-003))
        (CL:SETQ ITER-003 (%%REST ITER-003)))
       (CL:LET*
        ((PROP NULL)
         (ITER-004 (%THE-CONS-LIST (ALL-FACTS-OF-INSTANCE SEED FALSE TRUE))))
        (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-004 NIL)) DO
         (CL:PROGN (CL:SETQ PROP (%%VALUE ITER-004))
          (CL:SETQ ITER-004 (%%REST ITER-004)))
         (CL:TAGBODY
          (CL:WHEN (CL:NOT (CL:EQ (LOOKUP FACTS PROP) NULL))
           (CL:GO :CONTINUE))
          (CL:LET*
           ((FILTEROUTPUT (CL:FUNCALL FILTER PROP))
            (FACTPROPS (%%VALUE FILTEROUTPUT))
            (PROPSEEDS (%%VALUE (%%REST FILTEROUTPUT))))
           (CL:LET* ((FACTPROP NULL) (ITER-005 FACTPROPS))
            (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-005 NIL)) DO
             (CL:PROGN (CL:SETQ FACTPROP (%%VALUE ITER-005))
              (CL:SETQ ITER-005 (%%REST ITER-005)))
             (INSERT-AT FACTS FACTPROP FACTPROP) (INSERT FACTLIST FACTPROP)))
           (CL:LET* ((ARG NULL) (ITER-006 PROPSEEDS))
            (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-006 NIL)) DO
             (CL:PROGN (CL:SETQ ARG (%%VALUE ITER-006))
              (CL:SETQ ITER-006 (%%REST ITER-006)))
             (CL:WHEN (CL:EQ (LOOKUP VISITED ARG) NULL)
              (INSERT-NEW NEWSEEDS ARG)))))
          :CONTINUE)))))
     (CL:SETQ SEEDS NEWSEEDS) (CL:SETQ NEWSEEDS (NEW-LIST))))
   (REVERSE FACTLIST)))

;;; (DEFUN (ALL-FACTS-OF-N (CONS OF PROPOSITION)) ...)

(CL:DEFUN %ALL-FACTS-OF-N (N INSTANCEREFS)
  "This is a generalization of `all-facts-of' (which see).  With `n' = 0
and only one instance this command behaves just like `all-facts-of'.
Otherwise, returns a cons list of all definite (TRUE or FALSE) propositions
that reference any of the instances listed in `instanceRefs', plus if `n' >= 1
all propositions that reference any instances that are arguments of
propositions collected in the previous step, plus if `n' >= 2... and so on.
That is, if we only consider binary propositions, this can be viewed as
growing a graph with instances as its nodes and predicates as its arcs
starting from the set of seed `instanceRefs' to depth `n'-1.
Caution: with a fully connected KB and large enough `n' this could return
the whole knowledge base.

The returned propositions include those asserted to be true or false by
default, but it does not include propositions that are found to be
true only by running the query engine.  Facts inferred to be true by
the forward chainer will be included.  Hence, the returned list of
facts may be longer in a context where the forward chainer has been
run then in one where it has not (see `run-forward-rules')."
  (CL:DECLARE (CL:TYPE CL:FIXNUM N))
  #+MCL
  (CL:CHECK-TYPE N CL:FIXNUM)
  (%THE-CONS-LIST (CALL-ALL-FACTS-OF-N N INSTANCEREFS)))

(CL:DEFUN ALL-FACTS-OF-N-EVALUATOR-WRAPPER (ARGUMENTS)
  (%ALL-FACTS-OF-N (%WRAPPER-VALUE (%%VALUE ARGUMENTS)) (%%REST ARGUMENTS)))

(CL:DEFMACRO ALL-FACTS-OF-N (CL:&WHOLE EXPRESSION CL:&REST IGNORE)
  "This is a generalization of `all-facts-of' (which see).  With `n' = 0
and only one instance this command behaves just like `all-facts-of'.
Otherwise, returns a cons list of all definite (TRUE or FALSE) propositions
that reference any of the instances listed in `instanceRefs', plus if `n' >= 1
all propositions that reference any instances that are arguments of
propositions collected in the previous step, plus if `n' >= 2... and so on.
That is, if we only consider binary propositions, this can be viewed as
growing a graph with instances as its nodes and predicates as its arcs
starting from the set of seed `instanceRefs' to depth `n'-1.
Caution: with a fully connected KB and large enough `n' this could return
the whole knowledge base.

The returned propositions include those asserted to be true or false by
default, but it does not include propositions that are found to be
true only by running the query engine.  Facts inferred to be true by
the forward chainer will be included.  Hence, the returned list of
facts may be longer in a context where the forward chainer has been
run then in one where it has not (see `run-forward-rules')."
  (CL:DECLARE (CL:IGNORE IGNORE))
  (CL:LET ((*IGNORETRANSLATIONERRORS?* FALSE))
   (CL-INCREMENTALLY-TRANSLATE EXPRESSION)))

(CL:SETF (CL:MACRO-FUNCTION (CL:QUOTE /LOGIC/ALL-FACTS-OF-N)) (CL:MACRO-FUNCTION (CL:QUOTE ALL-FACTS-OF-N)))

(CL:DEFUN HELP-STARTUP-PROPAGATE1 ()
  (CL:PROGN
   (CL:SETQ SYM-PROPAGATE-LOGIC-DEFERRED-DEFAULT-PROPOSITIONS-INTERNAL
    (INTERN-RIGID-SYMBOL-WRT-MODULE "DEFERRED-DEFAULT-PROPOSITIONS-INTERNAL"
     NULL 0))
   (CL:SETQ SYM-PROPAGATE-LOGIC-EVALUATION-STATE-TABLE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "EVALUATION-STATE-TABLE" NULL 0))
   (CL:SETQ KWD-PROPAGATE-POSTED
    (INTERN-RIGID-SYMBOL-WRT-MODULE "POSTED" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-EVALUATION-QUEUE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "EVALUATION-QUEUE" NULL 0))
   (CL:SETQ KWD-PROPAGATE-FORWARD
    (INTERN-RIGID-SYMBOL-WRT-MODULE "FORWARD" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-BACKWARD-ONLY?
    (INTERN-RIGID-SYMBOL-WRT-MODULE "BACKWARD-ONLY?" NULL 0))
   (CL:SETQ SGT-PROPAGATE-LOGIC-NAMED-DESCRIPTION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "NAMED-DESCRIPTION" NULL 1))
   (CL:SETQ SGT-PROPAGATE-LOGIC-DESCRIPTION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "DESCRIPTION" NULL 1))
   (CL:SETQ SGT-PROPAGATE-LOGIC-COLLECT-FORWARD-CHAINING-RULES-MEMO-TABLE-000
    (INTERN-RIGID-SYMBOL-WRT-MODULE
     "COLLECT-FORWARD-CHAINING-RULES-MEMO-TABLE-000" NULL 1))
   (CL:SETQ KWD-PROPAGATE-FUNCTION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "FUNCTION" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-FORWARD-CHAINING-QUEUE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "FORWARD-CHAINING-QUEUE" NULL 0))
   (CL:SETQ KWD-PROPAGATE-ISA (INTERN-RIGID-SYMBOL-WRT-MODULE "ISA" NULL 2))
   (CL:SETQ KWD-PROPAGATE-PREDICATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "PREDICATE" NULL 2))
   (CL:SETQ SGT-PROPAGATE-LOGIC-SKOLEM
    (INTERN-RIGID-SYMBOL-WRT-MODULE "SKOLEM" NULL 1))
   (CL:SETQ SYM-PROPAGATE-LOGIC-HYPOTHESIZED-INSTANCE?
    (INTERN-RIGID-SYMBOL-WRT-MODULE "HYPOTHESIZED-INSTANCE?" NULL 0))
   (CL:SETQ KWD-PROPAGATE-PROPAGATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "PROPAGATE" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-COMPLEMENT-DESCRIPTION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "COMPLEMENT-DESCRIPTION" NULL 0))
   (CL:SETQ KWD-PROPAGATE-SINGLETONS?
    (INTERN-RIGID-SYMBOL-WRT-MODULE "SINGLETONS?" NULL 2))
   (CL:SETQ SGT-PROPAGATE-LOGIC-PROPOSITION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "PROPOSITION" NULL 1))
   (CL:SETQ KWD-PROPAGATE-KB-UPDATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "KB-UPDATE" NULL 2))
   (CL:SETQ KWD-PROPAGATE-META-KB-UPDATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "META-KB-UPDATE" NULL 2))
   (CL:SETQ KWD-PROPAGATE-META (INTERN-RIGID-SYMBOL-WRT-MODULE "META" NULL 2))
   (CL:SETQ SGT-PROPAGATE-STELLA-MODULE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "MODULE"
     (GET-STELLA-MODULE "/STELLA" TRUE) 1))
   (CL:SETQ SGT-PROPAGATE-LOGIC-LOGIC-OBJECT
    (INTERN-RIGID-SYMBOL-WRT-MODULE "LOGIC-OBJECT" NULL 1))
   (CL:SETQ SGT-PROPAGATE-STELLA-SYMBOL
    (INTERN-RIGID-SYMBOL-WRT-MODULE "SYMBOL"
     (GET-STELLA-MODULE "/STELLA" TRUE) 1))
   (CL:SETQ KWD-PROPAGATE-EVALUATED
    (INTERN-RIGID-SYMBOL-WRT-MODULE "EVALUATED" NULL 2))
   (CL:SETQ KWD-PROPAGATE-AND (INTERN-RIGID-SYMBOL-WRT-MODULE "AND" NULL 2))
   (CL:SETQ KWD-PROPAGATE-OR (INTERN-RIGID-SYMBOL-WRT-MODULE "OR" NULL 2))
   (CL:SETQ KWD-PROPAGATE-NOT (INTERN-RIGID-SYMBOL-WRT-MODULE "NOT" NULL 2))
   (CL:SETQ KWD-PROPAGATE-EQUIVALENT
    (INTERN-RIGID-SYMBOL-WRT-MODULE "EQUIVALENT" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-ELABORATED-IN-WORLDS
    (INTERN-RIGID-SYMBOL-WRT-MODULE "ELABORATED-IN-WORLDS" NULL 0))
   (CL:SETQ KWD-PROPAGATE-ELABORATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "ELABORATE" NULL 2))
   (CL:SETQ KWD-PROPAGATE-EXTENSIONAL-ASSERTION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "EXTENSIONAL-ASSERTION" NULL 2))
   (CL:SETQ SGT-PROPAGATE-PL-KERNEL-KB-INEQUALITY
    (INTERN-RIGID-SYMBOL-WRT-MODULE "INEQUALITY"
     (GET-STELLA-MODULE "/PL-KERNEL-KB" TRUE) 1))
   (CL:SETQ SGT-PROPAGATE-STELLA-SURROGATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "SURROGATE"
     (GET-STELLA-MODULE "/STELLA" TRUE) 1))
   (CL:SETQ KWD-PROPAGATE-EXISTS
    (INTERN-RIGID-SYMBOL-WRT-MODULE "EXISTS" NULL 2))
   (CL:SETQ KWD-PROPAGATE-FORALL
    (INTERN-RIGID-SYMBOL-WRT-MODULE "FORALL" NULL 2))
   (CL:SETQ SGT-PROPAGATE-STELLA-LITERAL-WRAPPER
    (INTERN-RIGID-SYMBOL-WRT-MODULE "LITERAL-WRAPPER"
     (GET-STELLA-MODULE "/STELLA" TRUE) 1))
   (CL:SETQ SYM-PROPAGATE-LOGIC-DESCRIPTION
    (INTERN-RIGID-SYMBOL-WRT-MODULE "DESCRIPTION" NULL 0))
   (CL:SETQ SGT-PROPAGATE-STELLA-SLOT
    (INTERN-RIGID-SYMBOL-WRT-MODULE "SLOT" (GET-STELLA-MODULE "/STELLA" TRUE)
     1))
   (CL:SETQ SYM-PROPAGATE-LOGIC-ALL-FACTS-OF
    (INTERN-RIGID-SYMBOL-WRT-MODULE "ALL-FACTS-OF" NULL 0))
   (CL:SETQ KWD-PROPAGATE-COMMON-LISP
    (INTERN-RIGID-SYMBOL-WRT-MODULE "COMMON-LISP" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-CALL-ALL-FACTS-OF
    (INTERN-RIGID-SYMBOL-WRT-MODULE "CALL-ALL-FACTS-OF" NULL 0))
   (CL:SETQ SYM-PROPAGATE-LOGIC-RETRACT-FACTS-OF
    (INTERN-RIGID-SYMBOL-WRT-MODULE "RETRACT-FACTS-OF" NULL 0))
   (CL:SETQ SGT-PROPAGATE-PL-KERNEL-KB-PHRASE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "PHRASE"
     (GET-STELLA-MODULE "/PL-KERNEL-KB" TRUE) 1))
   (CL:SETQ KWD-PROPAGATE-IMPLIES
    (INTERN-RIGID-SYMBOL-WRT-MODULE "IMPLIES" NULL 2))
   (CL:SETQ SYM-PROPAGATE-LOGIC-ALL-SENTENCES-OF
    (INTERN-RIGID-SYMBOL-WRT-MODULE "ALL-SENTENCES-OF" NULL 0))
   (CL:SETQ SYM-PROPAGATE-LOGIC-ALL-FACTS-OF-N
    (INTERN-RIGID-SYMBOL-WRT-MODULE "ALL-FACTS-OF-N" NULL 0))
   (CL:SETQ SYM-PROPAGATE-LOGIC-STARTUP-PROPAGATE
    (INTERN-RIGID-SYMBOL-WRT-MODULE "STARTUP-PROPAGATE" NULL 0))
   (CL:SETQ SYM-PROPAGATE-STELLA-METHOD-STARTUP-CLASSNAME
    (INTERN-RIGID-SYMBOL-WRT-MODULE "METHOD-STARTUP-CLASSNAME"
     (GET-STELLA-MODULE "/STELLA" TRUE) 0))))

(CL:DEFUN HELP-STARTUP-PROPAGATE2 ()
  (CL:PROGN
   (DEFINE-EXTERNAL-SLOT-FROM-STRINGIFIED-SOURCE
    "(DEFSLOT CONTEXT DEFERRED-DEFAULT-PROPOSITIONS-INTERNAL :TYPE (CONS OF PROPOSITION) :PUBLIC? TRUE :DOCUMENTATION \"List of propositions infered to be true or false by
default before we enter default propagation mode.\" :DEFAULT NULL :ALLOCATION :DYNAMIC)")
   (DEFINE-FUNCTION-OBJECT "DEFERRED-DEFAULT-PROPOSITIONS"
    "(DEFUN (DEFERRED-DEFAULT-PROPOSITIONS (CONS OF PROPOSITION)) ((SELF CONTEXT)))"
    (CL:FUNCTION DEFERRED-DEFAULT-PROPOSITIONS) NULL)
   (DEFINE-FUNCTION-OBJECT "DEFERRED-DEFAULT-PROPOSITIONS-SETTER"
    "(DEFUN DEFERRED-DEFAULT-PROPOSITIONS-SETTER ((SELF CONTEXT) (NEWVALUE (CONS OF PROPOSITION))))"
    (CL:FUNCTION DEFERRED-DEFAULT-PROPOSITIONS-SETTER) NULL)
   (DEFINE-EXTERNAL-SLOT-FROM-STRINGIFIED-SOURCE
    "(DEFSLOT CONTEXT EVALUATION-QUEUE :TYPE (LIST OF PROPOSITION) :PUBLIC? TRUE :DOCUMENTATION \"List of propositions waiting for evaluation in 
*context*.\" :DEFAULT NULL :ALLOCATION :DYNAMIC)")
   (DEFINE-EXTERNAL-SLOT-FROM-STRINGIFIED-SOURCE
    "(DEFSLOT CONTEXT EVALUATION-STATE-TABLE :TYPE (HASH-TABLE OF PROPOSITION KEYWORD) :PUBLIC? TRUE :DOCUMENTATION \"Records which propositions have been evaluated at
least one in the world 'self'.\" :ALLOCATION :DYNAMIC)")
   (DEFINE-FUNCTION-OBJECT "EVALUATION-STATE"
    "(DEFUN (EVALUATION-STATE KEYWORD) ((PROPOSITION PROPOSITION)))"
    (CL:FUNCTION EVALUATION-STATE) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATION-STATE-SETTER"
    "(DEFUN EVALUATION-STATE-SETTER ((PROPOSITION PROPOSITION) (STATE KEYWORD)))"
    (CL:FUNCTION EVALUATION-STATE-SETTER) NULL)
   (DEFINE-FUNCTION-OBJECT "POST-FOR-EVALUATION"
    "(DEFUN POST-FOR-EVALUATION ((SELF PROPOSITION)) :PUBLIC? TRUE)"
    (CL:FUNCTION POST-FOR-EVALUATION) NULL)
   (DEFINE-EXTERNAL-SLOT-FROM-STRINGIFIED-SOURCE
    "(DEFSLOT CONTEXT FORWARD-CHAINING-QUEUE :TYPE (LIST OF PROPOSITION) :PUBLIC? TRUE :DOCUMENTATION \"Queue of recently asserted or inferred propositions
having applicable forward rules.  Rules are triggered for each queue entry.\" :DEFAULT NULL :ALLOCATION :DYNAMIC)")
   (DEFINE-FUNCTION-OBJECT "HELP-COLLECT-FORWARD-RULES"
    "(DEFUN (HELP-COLLECT-FORWARD-RULES (CONS OF KEY-VALUE-LIST)) ((DESCRIPTION DESCRIPTION) (RULES KEY-VALUE-LIST) (INDICES KEY-VALUE-LIST) (TOUCHEDDEFAULT? BOOLEAN) (BEENTHERE LIST)))"
    (CL:FUNCTION HELP-COLLECT-FORWARD-RULES) NULL)
   (DEFINE-FUNCTION-OBJECT "COLLECT-FORWARD-CHAINING-RULES"
    "(DEFUN (COLLECT-FORWARD-CHAINING-RULES (KEY-VALUE-LIST OF PROPOSITION BOOLEAN-WRAPPER) (KEY-VALUE-LIST OF FORWARD-CHAINING-INDEX BOOLEAN-WRAPPER)) ((DESCRIPTION DESCRIPTION)))"
    (CL:FUNCTION COLLECT-FORWARD-CHAINING-RULES) NULL)
   (DEFINE-FUNCTION-OBJECT "HAS-FORWARD-CHAINING-RULES?"
    "(DEFUN (HAS-FORWARD-CHAINING-RULES? BOOLEAN) ((DESCRIPTION DESCRIPTION) (PROPOSITION PROPOSITION)))"
    (CL:FUNCTION HAS-FORWARD-CHAINING-RULES?) NULL)
   (DEFINE-FUNCTION-OBJECT "POST-TO-FORWARD-CHAINING-QUEUE"
    "(DEFUN POST-TO-FORWARD-CHAINING-QUEUE ((SELF PROPOSITION) (WORLD WORLD)))"
    (CL:FUNCTION POST-TO-FORWARD-CHAINING-QUEUE) NULL)
   (DEFINE-FUNCTION-OBJECT "APPLY-RULE-CONSEQUENT-TO-VECTOR"
    "(DEFUN APPLY-RULE-CONSEQUENT-TO-VECTOR ((CONSEQUENT DESCRIPTION) (ARGUMENTS ARGUMENTS-VECTOR) (RULE PROPOSITION) (TRIGGERDESCRIPTION DESCRIPTION) (TRIGGERPROPOSITION PROPOSITION) (TOUCHEDDEFAULT? BOOLEAN)))"
    (CL:FUNCTION APPLY-RULE-CONSEQUENT-TO-VECTOR) NULL)
   (DEFINE-FUNCTION-OBJECT "APPLY-FORWARD-RULES-TO-VECTOR"
    "(DEFUN APPLY-FORWARD-RULES-TO-VECTOR ((TRIGGERDESCRIPTION DESCRIPTION) (ARGUMENTS ARGUMENTS-VECTOR) (TRIGGERPROPOSITION PROPOSITION)))"
    (CL:FUNCTION APPLY-FORWARD-RULES-TO-VECTOR) NULL)
   (DEFINE-FUNCTION-OBJECT "APPLICABLE-FORWARD-RULE?"
    "(DEFUN (APPLICABLE-FORWARD-RULE? BOOLEAN) ((RULE PROPOSITION) (ARGUMENTS ARGUMENTS-VECTOR)))"
    (CL:FUNCTION APPLICABLE-FORWARD-RULE?) NULL)
   (DEFINE-METHOD-OBJECT
    "(DEFMETHOD REACT-TO-INFERENCE-UPDATE ((SELF SKOLEM)))"
    (CL:FUNCTION REACT-TO-INFERENCE-UPDATE) NULL)
   (DEFINE-METHOD-OBJECT
    "(DEFMETHOD REACT-TO-INFERENCE-UPDATE ((SELF PROPOSITION)))"
    (CL:FUNCTION REACT-TO-INFERENCE-UPDATE) NULL)
   (DEFINE-FUNCTION-OBJECT "REACT-TO-KB-UPDATE"
    "(DEFUN REACT-TO-KB-UPDATE ((CONTEXT CONTEXT) (OBJECT OBJECT)))"
    (CL:FUNCTION REACT-TO-KB-UPDATE) NULL)
   (DEFINE-FUNCTION-OBJECT "INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES"
    "(DEFUN INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES ((WORLD WORLD)))"
    (CL:FUNCTION INITIALIZE-CONSTRAINT-PROPAGATION-QUEUES) NULL)
   (DEFINE-FUNCTION-OBJECT "EXECUTE-CONSTRAINT-PROPAGATION-QUEUES"
    "(DEFUN EXECUTE-CONSTRAINT-PROPAGATION-QUEUES ())"
    (CL:FUNCTION EXECUTE-CONSTRAINT-PROPAGATION-QUEUES) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-NEW-PROPOSITION"
    "(DEFUN EVALUATE-NEW-PROPOSITION ((SELF PROPOSITION)))"
    (CL:FUNCTION EVALUATE-NEW-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-AND-PROPOSITION"
    "(DEFUN EVALUATE-AND-PROPOSITION ((SELF PROPOSITION)))"
    (CL:FUNCTION EVALUATE-AND-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-OR-PROPOSITION"
    "(DEFUN EVALUATE-OR-PROPOSITION ((SELF PROPOSITION)))"
    (CL:FUNCTION EVALUATE-OR-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-NOT-PROPOSITION"
    "(DEFUN EVALUATE-NOT-PROPOSITION ((SELF PROPOSITION)))"
    (CL:FUNCTION EVALUATE-NOT-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-EQUIVALENT-PROPOSITION"
    "(DEFUN EVALUATE-EQUIVALENT-PROPOSITION ((SELF PROPOSITION)))"
    (CL:FUNCTION EVALUATE-EQUIVALENT-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EQUATE-EQUIVALENT-FUNCTION-PROPOSITIONS"
    "(DEFUN EQUATE-EQUIVALENT-FUNCTION-PROPOSITIONS ((SELF PROPOSITION)))"
    (CL:FUNCTION EQUATE-EQUIVALENT-FUNCTION-PROPOSITIONS) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-FUNCTION-PROPOSITION"
    "(DEFUN EVALUATE-FUNCTION-PROPOSITION ((SELF PROPOSITION)) :PUBLIC? TRUE)"
    (CL:FUNCTION EVALUATE-FUNCTION-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-PREDICATE-PROPOSITION"
    "(DEFUN EVALUATE-PREDICATE-PROPOSITION ((SELF PROPOSITION)) :PUBLIC? TRUE)"
    (CL:FUNCTION EVALUATE-PREDICATE-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-PROPOSITION"
    "(DEFUN EVALUATE-PROPOSITION ((SELF PROPOSITION)) :PUBLIC? TRUE)"
    (CL:FUNCTION EVALUATE-PROPOSITION) NULL)
   (DEFINE-METHOD-OBJECT
    "(DEFMETHOD MARK-AS-INCOHERENT ((SELF LOGIC-OBJECT)))"
    (CL:FUNCTION MARK-AS-INCOHERENT) NULL)
   (DEFINE-METHOD-OBJECT "(DEFMETHOD MARK-AS-INCOHERENT ((SELF SKOLEM)))"
    (CL:FUNCTION MARK-AS-INCOHERENT) NULL)
   (DEFINE-EXTERNAL-SLOT-FROM-STRINGIFIED-SOURCE
    "(DEFSLOT LOGIC-OBJECT ELABORATED-IN-WORLDS :TYPE (LIST OF WORLD) :DOCUMENTATION \"List of worlds in which 'elaborate-instance' has
been applied to 'self'.\" :ALLOCATION :DYNAMIC)")
   (DEFINE-FUNCTION-OBJECT "RECURSIVELY-REACT-TO-INFERENCE-UPDATE"
    "(DEFUN RECURSIVELY-REACT-TO-INFERENCE-UPDATE ((SELF PROPOSITION)))"
    (CL:FUNCTION RECURSIVELY-REACT-TO-INFERENCE-UPDATE) NULL)
   (DEFINE-FUNCTION-OBJECT "ELABORATE-META-INSTANCE"
    "(DEFUN ELABORATE-META-INSTANCE ((SELF OBJECT)))"
    (CL:FUNCTION ELABORATE-META-INSTANCE) NULL)
   (DEFINE-FUNCTION-OBJECT "ELABORATE-INSTANCE"
    "(DEFUN ELABORATE-INSTANCE ((SELF OBJECT)))"
    (CL:FUNCTION ELABORATE-INSTANCE) NULL)
   (DEFINE-FUNCTION-OBJECT "EVALUATE-REACHABLE-INEQUALITIES"
    "(DEFUN EVALUATE-REACHABLE-INEQUALITIES ((SELF LOGIC-OBJECT) (VISITEDLIST LIST)))"
    (CL:FUNCTION EVALUATE-REACHABLE-INEQUALITIES) NULL)
   (DEFINE-FUNCTION-OBJECT "ELABORATE-SURROGATES-IN-PROPOSITION"
    "(DEFUN ELABORATE-SURROGATES-IN-PROPOSITION ((PROPOSITION PROPOSITION)))"
    (CL:FUNCTION ELABORATE-SURROGATES-IN-PROPOSITION) NULL)
   (DEFINE-FUNCTION-OBJECT "COLLECT-FUNCTION-PROPOSITION-FACTS"
    "(DEFUN COLLECT-FUNCTION-PROPOSITION-FACTS ((SELF PROPOSITION) (FACTS (LIST OF PROPOSITION)) (BEENTHERE (LIST OF PROPOSITION)) (INCLUDEUNKNOWN? BOOLEAN)))"
    (CL:FUNCTION COLLECT-FUNCTION-PROPOSITION-FACTS) NULL)
   (DEFINE-FUNCTION-OBJECT "HELP-COLLECT-FACTS"
    "(DEFUN HELP-COLLECT-FACTS ((SELF OBJECT) (FACTS (LIST OF PROPOSITION)) (BEENTHERE (LIST OF PROPOSITION)) (INCLUDEUNKNOWN? BOOLEAN)))"
    (CL:FUNCTION HELP-COLLECT-FACTS) NULL)
   (DEFINE-FUNCTION-OBJECT "ALL-FACTS-OF-INSTANCE"
    "(DEFUN (ALL-FACTS-OF-INSTANCE (LIST OF PROPOSITION)) ((SELF OBJECT) (INCLUDEUNKNOWNFACTS? BOOLEAN) (ELABORATE? BOOLEAN)) :DOCUMENTATION \"Return a list of all definite (TRUE or FALSE) propositions
attached to `self'.\" :PUBLIC? TRUE)" (CL:FUNCTION ALL-FACTS-OF-INSTANCE)
    NULL)
   (DEFINE-FUNCTION-OBJECT "ALL-FACTS-OF"
    "(DEFUN (ALL-FACTS-OF (CONS OF PROPOSITION)) ((INSTANCEREF NAME)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Return a cons list of all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'.  This includes propositions
asserted to be true by default, but it does not include propositions
that are found to be TRUE only by running the query engine.  Facts
inferred to be TRUE by the forward chainer will be included.
Hence, the returned list of facts may be longer in a context where the
forward chainer has been run then in one where it has not (see
`run-forward-rules').\")" (CL:FUNCTION %ALL-FACTS-OF) NULL)
   (DEFINE-FUNCTION-OBJECT "CALL-ALL-FACTS-OF"
    "(DEFUN (CALL-ALL-FACTS-OF (LIST OF PROPOSITION)) ((INSTANCEREF OBJECT)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? TRUE :DOCUMENTATION \"Return a list of all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'.\")"
    (CL:FUNCTION %CALL-ALL-FACTS-OF) NULL)
   (DEFINE-FUNCTION-OBJECT "RETRACT-FACTS-OF-INSTANCE"
    "(DEFUN RETRACT-FACTS-OF-INSTANCE ((SELF LOGIC-OBJECT)) :DOCUMENTATION \"Retract all definite (TRUE or FALSE) propositions attached to `self'.\" :PUBLIC? TRUE)"
    (CL:FUNCTION RETRACT-FACTS-OF-INSTANCE) NULL)
   (DEFINE-FUNCTION-OBJECT "RETRACT-FACTS-OF"
    "(DEFUN RETRACT-FACTS-OF ((INSTANCEREF OBJECT)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Retract all definite (TRUE or FALSE) propositions
that reference the instance `instanceRef'.\")" (CL:FUNCTION %RETRACT-FACTS-OF)
    NULL)
   (DEFINE-FUNCTION-OBJECT "EXPLODE-VARIABLIZED-STRING"
    "(DEFUN (EXPLODE-VARIABLIZED-STRING LIST) ((STRING STRING)))"
    (CL:FUNCTION EXPLODE-VARIABLIZED-STRING) NULL)
   (DEFINE-FUNCTION-OBJECT "FACT-TO-SENTENCE"
    "(DEFUN (FACT-TO-SENTENCE STRING) ((SELF PROPOSITION) (PERIOD? BOOLEAN)))"
    (CL:FUNCTION FACT-TO-SENTENCE) NULL)
   (DEFINE-FUNCTION-OBJECT "ALL-SENTENCES-OF"
    "(DEFUN (ALL-SENTENCES-OF (CONS OF STRING-WRAPPER)) ((INSTANCEREF OBJECT)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? TRUE :DOCUMENTATION \"Return a list of sentences describing facts about 'instanceRef'.\")"
    (CL:FUNCTION %ALL-SENTENCES-OF) NULL)
   (DEFINE-FUNCTION-OBJECT "CALL-ALL-FACTS-OF-N"
    "(DEFUN (CALL-ALL-FACTS-OF-N (LIST OF PROPOSITION)) ((N INTEGER) (SEEDREFS CONS)))"
    (CL:FUNCTION CALL-ALL-FACTS-OF-N) NULL)
   (DEFINE-FUNCTION-OBJECT "FILTERED-ALL-FACTS-OF-N"
    "(DEFUN (FILTERED-ALL-FACTS-OF-N (LIST OF PROPOSITION)) ((N INTEGER) (SEEDREFS CONS) (FILTER FUNCTION-CODE)))"
    (CL:FUNCTION FILTERED-ALL-FACTS-OF-N) NULL)
   (DEFINE-FUNCTION-OBJECT "ALL-FACTS-OF-N"
    "(DEFUN (ALL-FACTS-OF-N (CONS OF PROPOSITION)) ((N INTEGER) |&REST| (INSTANCEREFS NAME)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"This is a generalization of `all-facts-of' (which see).  With `n' = 0
and only one instance this command behaves just like `all-facts-of'.
Otherwise, returns a cons list of all definite (TRUE or FALSE) propositions
that reference any of the instances listed in `instanceRefs', plus if `n' >= 1
all propositions that reference any instances that are arguments of
propositions collected in the previous step, plus if `n' >= 2... and so on.
That is, if we only consider binary propositions, this can be viewed as
growing a graph with instances as its nodes and predicates as its arcs
starting from the set of seed `instanceRefs' to depth `n'-1.
Caution: with a fully connected KB and large enough `n' this could return
the whole knowledge base.

The returned propositions include those asserted to be true or false by
default, but it does not include propositions that are found to be
true only by running the query engine.  Facts inferred to be true by
the forward chainer will be included.  Hence, the returned list of
facts may be longer in a context where the forward chainer has been
run then in one where it has not (see `run-forward-rules').\")"
    (CL:FUNCTION %ALL-FACTS-OF-N)
    (CL:FUNCTION ALL-FACTS-OF-N-EVALUATOR-WRAPPER))
   (DEFINE-FUNCTION-OBJECT "STARTUP-PROPAGATE"
    "(DEFUN STARTUP-PROPAGATE () :PUBLIC? TRUE)"
    (CL:FUNCTION STARTUP-PROPAGATE) NULL)
   (CL:LET*
    ((FUNCTION (LOOKUP-FUNCTION SYM-PROPAGATE-LOGIC-STARTUP-PROPAGATE)))
    (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS FUNCTION)
     SYM-PROPAGATE-STELLA-METHOD-STARTUP-CLASSNAME
     (WRAP-STRING "_Startup-Propagate") NULL-STRING-WRAPPER))))

(CL:DEFUN STARTUP-PROPAGATE ()
  (CL:LET*
   ((*MODULE* (GET-STELLA-MODULE "/LOGIC" (> *STARTUP-TIME-PHASE* 1)))
    (*CONTEXT* *MODULE*))
   (CL:DECLARE (CL:SPECIAL *MODULE* *CONTEXT*))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 2) (HELP-STARTUP-PROPAGATE1))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 4)
    (CL:SETQ *FILLINGCONSTRAINTPROPAGATIONQUEUES?* FALSE)
    (CL:SETQ *DEFERINGDEFAULTFORWARDINFERENCES?* FALSE)
    (CL:SETQ *JUST-IN-TIME-FORWARD-INFERENCE?* TRUE))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 6) (FINALIZE-CLASSES))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 7) (HELP-STARTUP-PROPAGATE2))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 8) (FINALIZE-SLOTS)
    (CLEANUP-UNFINALIZED-CLASSES))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 9)
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *FILLINGCONSTRAINTPROPAGATIONQUEUES?* BOOLEAN FALSE :DOCUMENTATION \"True if we are inside of 'react-to-kb-update'.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *DEFERINGDEFAULTFORWARDINFERENCES?* BOOLEAN FALSE :DOCUMENTATION \"True if we are propagating strict inferences, and
posting derived default propositions to temporary queues.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *CURRENTLYEXECUTINGFORWARDCHAININGQUEUE* (LIST OF PROPOSITION) NULL :DOCUMENTATION \"Points to an active queue of propositions that will trigger
forward inference.  New queue entries are posted to a separate queue
attached to *context*.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *COLLECTFORWARDPROPOSITIONS* (CONS OF PROPOSITION) NULL :DOCUMENTATION \"Collect goes-true propositions produced by forward
chaining.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFGLOBAL *JUST-IN-TIME-FORWARD-INFERENCE?* BOOLEAN TRUE :DOCUMENTATION \"If TRUE, ensures that forward propagation has
been applied to each instance 'touched' during a query.\")")
    (REGISTER-NATIVE-NAME SYM-PROPAGATE-LOGIC-ALL-FACTS-OF
     KWD-PROPAGATE-COMMON-LISP KWD-PROPAGATE-FUNCTION)
    (REGISTER-NATIVE-NAME SYM-PROPAGATE-LOGIC-CALL-ALL-FACTS-OF
     KWD-PROPAGATE-COMMON-LISP KWD-PROPAGATE-FUNCTION)
    (REGISTER-NATIVE-NAME SYM-PROPAGATE-LOGIC-RETRACT-FACTS-OF
     KWD-PROPAGATE-COMMON-LISP KWD-PROPAGATE-FUNCTION)
    (REGISTER-NATIVE-NAME SYM-PROPAGATE-LOGIC-ALL-SENTENCES-OF
     KWD-PROPAGATE-COMMON-LISP KWD-PROPAGATE-FUNCTION)
    (REGISTER-NATIVE-NAME SYM-PROPAGATE-LOGIC-ALL-FACTS-OF-N
     KWD-PROPAGATE-COMMON-LISP KWD-PROPAGATE-FUNCTION))))
